/*
 *  $Id: app.c 27646 2025-03-27 18:26:14Z yeti-dn $
 *  Copyright (C) 2003-2024 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "libgwyddion/gwymacros.h"
#include "libgwyddion/gwythreads.h"
#include "libprocess/gwyprocess.h"
#include "libgwymodule/gwymodule.h"
#include "libgwydgets/gwydgets.h"
#include "libgwydgets/gwygraphwindow.h"

#include "gwyapp.h"
#include "gwymoduleutils.h"
#include "gwyappinternal.h"
#include "sanity.h"

enum {
    ITEM_PIXELSQUARE,
    ITEM_REALSQUARE
};

enum {
    PARAM_PREVIEW_TYPE,
    PARAM_PREVIEW_LEVEL,
    PARAM_PREVIEW_IMAGE,
    PARAM_PREVIEW_UPDATE,
};

typedef enum {
    BRICK_PREVIEW_MEAN,
    BRICK_PREVIEW_MINIMUM,
    BRICK_PREVIEW_MAXIMUM,
    BRICK_PREVIEW_MINPOS,
    BRICK_PREVIEW_MAXPOS,
    BRICK_PREVIEW_RMS,
    BRICK_PREVIEW_IMAGE,
    BRICK_PREVIEW_SECTION,
} BrickPreviewType;

typedef enum {
    LAWN_PREVIEW_MEAN,
    LAWN_PREVIEW_MINIMUM,
    LAWN_PREVIEW_MAXIMUM,
    LAWN_PREVIEW_RMS,
    LAWN_PREVIEW_IMAGE,
} LawnPreviewType;

typedef struct {
    GwyBrick *brick;
    GwyDataField *preview;
    GwyDataField *old_preview;
    GwyContainer *data;
    gint id;
    GQuark quark;
    GwyParams *params;
    GtkWidget *dialog;
    gboolean initial_preview;
} BrickPreviewGUI;

static GtkWidget *gwy_app_main_window = NULL;

static GwyTool* current_tool = NULL;
static GQuark corner_item_quark = 0;

static gboolean   gwy_app_main_window_save_position     (void);
static void       gwy_app_main_window_restore_position  (void);
static gboolean   gwy_app_confirm_quit                  (void);
static gboolean   gwy_app_confirm_quit_dialog           (GSList *unsaved);
static gboolean   gwy_app_data_window_configured        (GwyDataWindow *window);
static GtkWidget* gwy_app_menu_data_popup_create        (GtkAccelGroup *accel_group);
static GtkWidget* gwy_app_menu_data_corner_create       (GtkAccelGroup *accel_group,
                                                         GwyAppPage pageno);
static void       gwy_app_data_window_change_square     (GtkWidget *item,
                                                         gpointer user_data);
static gboolean   gwy_app_data_corner_menu_popup_mouse  (GtkWidget *menu,
                                                         GdkEventButton *event,
                                                         GtkWidget *ebox);
static gboolean   gwy_app_data_popup_menu_popup_mouse   (GtkWidget *menu,
                                                         GdkEventButton *event,
                                                         GwyDataView *data_view);
static void       gwy_app_data_popup_menu_popup_key     (GtkWidget *menu,
                                                         GtkWidget *data_window);
static gboolean   gwy_app_data_key_pressed              (GtkWidget *data_window,
                                                         GdkEventKey *event,
                                                         gpointer user_data);
static gboolean   gwy_app_graph_window_configured       (GwyGraphWindow *window);
static gboolean   gwy_app_graph_popup_menu_popup_mouse  (GtkWidget *menu,
                                                         GdkEventButton *event,
                                                         GwyGraph *graph);
static void       gwy_app_graph_popup_menu_popup_key    (GtkWidget *menu,
                                                         GtkWidget *graph);
static void       gwy_app_gl_window_export              (GwyGLWindow *window);
static void       gwy_app_gl_window_set_defaults        (GwyGLWindow *window);
static gboolean   gwy_app_brick_window_configured       (GwyDataWindow *window);
static void       change_brick_preview                  (GwyDataWindow *data_window);
static GtkWidget* gwy_app_menu_brick_popup_create       (GtkAccelGroup *accel_group);
static gboolean   gwy_app_brick_popup_menu_popup_mouse  (GtkWidget *menu,
                                                         GdkEventButton *event,
                                                         GwyDataView *data_view);
static void       gwy_app_brick_popup_menu_popup_key    (GtkWidget *menu,
                                                         GtkWidget *data_window);
static gboolean   gwy_app_surface_window_configured     (GwyDataWindow *window);
static void       update_surface_preview                (GwyDataWindow *data_window);
static void       surface_density_map_toggled           (GwyDataWindow *data_window,
                                                         GtkToggleButton *toggle);
static GtkWidget* gwy_app_menu_surface_popup_create     (GtkAccelGroup *accel_group);
static gboolean   gwy_app_surface_popup_menu_popup_mouse(GtkWidget *menu,
                                                         GdkEventButton *event,
                                                         GwyDataView *data_view);
static void       gwy_app_surface_popup_menu_popup_key  (GtkWidget *menu,
                                                         GtkWidget *data_window);
static gboolean   gwy_app_lawn_window_configured        (GwyDataWindow *window);
static void       change_lawn_preview                   (GwyDataWindow *data_window);
static GtkWidget* gwy_app_menu_lawn_popup_create        (GtkAccelGroup *accel_group);
static gboolean   gwy_app_lawn_popup_menu_popup_mouse   (GtkWidget *menu,
                                                         GdkEventButton *event,
                                                         GwyDataView *data_view);
static void       gwy_app_lawn_popup_menu_popup_key     (GtkWidget *menu,
                                                         GtkWidget *data_window);
static void       gwy_app_save_gl_export                (GtkWidget *dialog,
                                                         gint response,
                                                         GwyGLWindow *glwindow);
static gboolean   gwy_app_gl_window_configured          (GwyGLWindow *window);
static void       gwy_app_gl_window_add_overlay_menu    (GwyGLWindow *glwindow);
static void       gwy_app_gl_window_update_chooser      (GwyGLWindow *glwindow);
static void       gwy_app_gl_window_set_data2           (GwyGLWindow *glwindow,
                                                         gint id,
                                                         gboolean mask);
static gboolean   gwy_app_gl_window_data2_filter        (GwyContainer *data2,
                                                         gint id2,
                                                         gpointer user_data);
static void       gwy_app_any_window_reset_zoom         (gpointer user_data);
static void       metadata_browser                      (gpointer pwhat);
static void       log_browser                           (gpointer pwhat);
static void       gwy_app_change_mask_color             (void);
static void       save_widget_screen_relative_size      (GtkWidget *widget,
                                                         GwyContainer *container,
                                                         const gchar *prefix,
                                                         gboolean absolute_too);
static void       restore_data_window_zoom              (GwyDataWindow *data_window);
static void       restore_other_window_size             (GtkWindow *window,
                                                         GwyContainer *container,
                                                         const gchar *prefix);
static gint64     monitor_config_checksum               (GtkWidget *widget);

/* Must match GwyGLViewLabel */
static const struct {
    const gchar *key;
    const gchar *default_text;
}
gl_labels[GWY_GL_VIEW_NLABELS] = {
    { "x",   "x: $x", },
    { "y",   "y: $y", },
    { "min", "$min",  },
    { "max", "$max",  },
};

/********************************************************************************************************************
 *                                                                                                                  *
 *     Stubs FIXME GTK3                                                                                             *
 *                                                                                                                  *
 ********************************************************************************************************************/

void
gwy_app_gl_material_editor(void)
{
}

/********************************************************************************************************************
 *                                                                                                                  *
 *     Main, toolbox                                                                                                *
 *                                                                                                                  *
 ********************************************************************************************************************/

/**
 * gwy_app_quit:
 *
 * Quits the application.
 *
 * This function may present a confirmation dialog to the user and it may let the application to continue running.  If
 * it quits the application, it performs some shutdown actions and then quits the Gtk+ main loop with gtk_main_quit().
 *
 * Returns: Always %TRUE to be usable as an event handler.  However, if the application is actually terminated, this
 *          function does not return.
 **/
gboolean
gwy_app_quit(void)
{
    gwy_debug("");
    if (!gwy_app_confirm_quit())
        return TRUE;

    gwy_app_data_browser_shut_down();
    gwy_app_main_window_save_position();
    g_clear_object(&current_tool);
    /* XXX: EXIT-CLEAN-UP */
    gtk_widget_destroy(gwy_app_main_window);
    //gwy_app_get_graph_thumbnail(NULL, -1, 0, 0);
    /* FIXME: sometimes fails with
     * Sensitivity group is finialized when it still contains widget lists.
     */
    g_object_unref(gwy_app_sensitivity_get_group());

    gtk_main_quit();
    return TRUE;
}

static gboolean
gwy_app_main_window_save_position(void)
{
    gwy_app_save_window_position(GTK_WINDOW(gwy_app_main_window), "/app/toolbox", TRUE, FALSE);
    return FALSE;
}

static void
gwy_app_main_window_restore_position(void)
{
    gwy_app_restore_window_position(GTK_WINDOW(gwy_app_main_window), "/app/toolbox", FALSE);
}

/**
 * gwy_app_add_main_accel_group:
 * @window: A window.
 *
 * Adds main (global) application accelerator group to a window.
 *
 * This includes accelerators for terminating Gwyddion, opening files, etc.
 **/
void
gwy_app_add_main_accel_group(GtkWindow *window)
{
    GtkWidget *main_window;
    GtkAccelGroup *accel_group;

    g_return_if_fail(GTK_IS_WINDOW(window));
    main_window = gwy_app_main_window_get();
    if (!main_window)
        return;

    g_return_if_fail(GTK_IS_WINDOW(main_window));

    accel_group = GTK_ACCEL_GROUP(g_object_get_data(G_OBJECT(main_window), "accel_group"));
    if (accel_group)
        gtk_window_add_accel_group(window, accel_group);
}

/**
 * gwy_app_main_window_get:
 *
 * Returns Gwyddion main application window (toolbox).
 *
 * Returns: The Gwyddion toolbox.
 **/
GtkWidget*
gwy_app_main_window_get(void)
{
    return gwy_app_main_window;
}

/**
 * gwy_app_main_window_set:
 * @window: A window, presumably the Gwyddion toolbox.
 *
 * Sets the main application window.
 *
 * This function is probably only useful in Gwyddion itself and should be ignored.
 *
 * It needs to be called exactly once at Gwyddion startup. It restores the toolbox position and makes
 * gwy_app_main_window_get() work.
 **/
void
gwy_app_main_window_set(GtkWidget *window)
{
    g_return_if_fail(GTK_IS_WINDOW(window));
    if (gwy_app_main_window && window != gwy_app_main_window) {
        g_critical("The main window was already set to %p.", gwy_app_main_window);
        return;
    }

    gwy_app_main_window = window;
    gwy_app_main_window_restore_position();
    g_signal_connect(window, "delete-event", G_CALLBACK(gwy_app_main_window_save_position), NULL);
    g_signal_connect(window, "show", G_CALLBACK(gwy_app_main_window_restore_position), NULL);
}

static gboolean
gwy_app_confirm_quit(void)
{
    GSList *unsaved = NULL;
    gboolean ok G_GNUC_UNUSED;

    if (!unsaved)
        return TRUE;
    ok = gwy_app_confirm_quit_dialog(unsaved);
    g_slist_free(unsaved);

    return TRUE;
}

static gboolean
gwy_app_confirm_quit_dialog(GSList *unsaved)
{
    GtkWidget *dialog;
    gchar *text;
    gint response;

    text = NULL;
    while (unsaved) {
        GwyDataWindow *data_window = GWY_DATA_WINDOW(unsaved->data);
        /* TODO: must use filename, not channel name, undo is per-file */
        const gchar *filename = gwy_data_window_get_data_name(data_window);

        text = g_strconcat(filename, "\n", text, NULL);
        unsaved = g_slist_next(unsaved);
    }
    dialog = gtk_message_dialog_new(GTK_WINDOW(gwy_app_main_window_get()),
                                    GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
                                    _("Some data are unsaved:\n%s\nReally quit?"),
                                    text);
    g_free(text);

    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_window_present(GTK_WINDOW(dialog));
    response = gtk_dialog_run(GTK_DIALOG(dialog));
    if (response != GTK_RESPONSE_NONE)
        gtk_widget_destroy(dialog);

    return response == GTK_RESPONSE_YES;
}

/********************************************************************************************************************
 *                                                                                                                  *
 *     Data window management                                                                                       *
 *                                                                                                                  *
 ********************************************************************************************************************/

/**
 * _gwy_app_data_view_set_current:
 * @data_view: Data view, can be %NULL.
 *
 * Updates application state upon switch to new data.
 **/
void
_gwy_app_data_view_set_current(GwyDataView *data_view)
{
    if (current_tool)
        gwy_tool_data_switched(current_tool, data_view);
}

void
_gwy_app_data_window_setup(GwyDataWindow *data_window)
{
    static GtkWidget *popup_menu = NULL;
    static GtkWidget *corner_menu = NULL;

    GwyDataView *data_view;
    GtkWidget *corner, *ebox, *main_window;
    GtkAccelGroup *accel_group;
    GwyContainer *settings;
    GwyLayerBasicRangeType range_type;

    if ((!popup_menu || !corner_menu) && (main_window = gwy_app_main_window_get())) {
        g_return_if_fail(GTK_IS_WINDOW(main_window));
        accel_group = GTK_ACCEL_GROUP(g_object_get_data(G_OBJECT(main_window), "accel_group"));

        if (accel_group && !popup_menu) {
            popup_menu = gwy_app_menu_data_popup_create(accel_group);
            gtk_widget_show_all(popup_menu);
        }
        if (accel_group && !corner_menu) {
            corner_menu = gwy_app_menu_data_corner_create(accel_group, GWY_PAGE_CHANNELS);
            gtk_widget_show_all(corner_menu);
        }
    }

    gwy_app_add_main_accel_group(GTK_WINDOW(data_window));
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
    corner = gtk_arrow_new(GTK_ARROW_RIGHT, GTK_SHADOW_ETCHED_OUT);
#pragma GCC diagnostic pop
    gwy_set_widget_padding(corner, 2, 2, 2, 2);

    ebox = gtk_event_box_new();
    gtk_container_add(GTK_CONTAINER(ebox), corner);
    gtk_widget_add_events(ebox, GDK_BUTTON_PRESS_MASK);
    gtk_widget_show_all(ebox);

    gwy_data_window_set_ul_corner_widget(data_window, ebox);
    gwy_help_add_to_window(GTK_WINDOW(data_window), "data-windows", NULL, GWY_HELP_DEFAULT);
    restore_data_window_zoom(data_window);

    data_view = GWY_DATA_VIEW(gwy_data_window_get_data_view(data_window));
    g_signal_connect_swapped(data_view, "button-press-event",
                             G_CALLBACK(gwy_app_data_popup_menu_popup_mouse), popup_menu);
    g_signal_connect_swapped(data_window, "popup-menu", G_CALLBACK(gwy_app_data_popup_menu_popup_key), popup_menu);
    g_signal_connect_swapped(ebox, "button-press-event", G_CALLBACK(gwy_app_data_corner_menu_popup_mouse), corner_menu);
    g_signal_connect(data_window, "key-press-event", G_CALLBACK(gwy_app_data_key_pressed), NULL);
    g_signal_connect(data_window, "configure-event", G_CALLBACK(gwy_app_data_window_configured), NULL);
    g_signal_connect(data_window, "size-allocate", G_CALLBACK(gwy_app_data_window_configured), NULL);

    settings = gwy_app_settings_get();
    if (gwy_container_gis_enum_by_name(settings, "/app/default-range-type", &range_type)) {
        GwyDataViewLayer *layer = gwy_data_view_get_base_layer(data_view);
        g_object_set(layer, "default-range-type", range_type, NULL);
    }
}

static gboolean
gwy_app_data_window_configured(GwyDataWindow *window)
{
    GwyDataView *view = GWY_DATA_VIEW(gwy_data_window_get_data_view(window));
    GwyContainer *container = gwy_data_view_get_data(view);
    const gchar *prefix = gwy_data_view_get_data_prefix(view);
    gchar *key;

    g_return_val_if_fail(container, FALSE);
    g_return_val_if_fail(prefix, FALSE);

    key = g_strconcat(prefix, "/view/scale", NULL);
    gwy_container_set_double_by_name(container, key, gwy_data_view_get_real_zoom(view));
    g_free(key);

    save_widget_screen_relative_size(GTK_WIDGET(view), container, prefix, FALSE);

    return FALSE;
}

static GtkWidget*
gwy_app_menu_data_popup_create(GtkAccelGroup *accel_group)
{
    static struct {
        const gchar *label;
        gpointer callback;
        gpointer cbdata;
        guint key;
        GdkModifierType mods;
    }
    const menu_items[] = {
        {
            NULL, gwy_app_run_process_func, "mask_remove",
            GDK_KEY_K, GDK_CONTROL_MASK
        },
        {
            N_("Mask _Color..."),  gwy_app_change_mask_color, NULL,
            0, 0
        },
        {
            NULL, gwy_app_run_process_func, "fix_zero",
            0, 0
        },
        {
            NULL, gwy_app_run_process_func, "presentation_remove",
            GDK_KEY_K, GDK_CONTROL_MASK | GDK_SHIFT_MASK
        },
        {
            NULL, gwy_app_run_process_func, "level",
            GDK_KEY_L, GDK_CONTROL_MASK
        },
        {
            N_("Zoom _1:1"), gwy_app_any_window_reset_zoom, NULL,
            0, 0
        },
        {
            N_("Metadata _Browser..."), metadata_browser, GUINT_TO_POINTER(GWY_APP_DATA_FIELD),
            GDK_KEY_B, GDK_CONTROL_MASK | GDK_SHIFT_MASK
        },
        {
            N_("View _Log..."), log_browser, GUINT_TO_POINTER(GWY_APP_DATA_FIELD),
            0, 0
        },
    };
    GwySensitivityGroup *sensgroup;
    GtkWidget *menu, *item;
    const gchar *name;
    guint i, mask;

    menu = gtk_menu_new();
    if (accel_group)
        gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    sensgroup = gwy_app_sensitivity_get_group();
    for (i = 0; i < G_N_ELEMENTS(menu_items); i++) {
        if (menu_items[i].callback == gwy_app_run_process_func
            && !gwy_process_func_get_run_types((gchar*)menu_items[i].cbdata)) {
            g_warning("Processing function <%s> for data view context menu is not available.",
                      (const gchar*)menu_items[i].cbdata);
            continue;
        }
        if (menu_items[i].callback == gwy_app_run_process_func) {
            name = _(gwy_process_func_get_menu_path(menu_items[i].cbdata));
            name = strrchr(name, '/');
            if (!name) {
                g_warning("Invalid translated menu path for <%s>", (const gchar*)menu_items[i].cbdata);
                continue;
            }
            item = gtk_menu_item_new_with_mnemonic(name + 1);
            mask = gwy_process_func_get_sensitivity_mask(menu_items[i].cbdata);
            gwy_sensitivity_group_add_widget(sensgroup, item, mask);
        }
        else {
            item = gtk_menu_item_new_with_mnemonic(_(menu_items[i].label));

            if (menu_items[i].callback == gwy_app_change_mask_color)
                gwy_sensitivity_group_add_widget(sensgroup, item, GWY_MENU_FLAG_DATA_MASK);
        }

        if (menu_items[i].key)
            gtk_widget_add_accelerator(item, "activate", accel_group, menu_items[i].key, menu_items[i].mods,
                                       GTK_ACCEL_VISIBLE | GTK_ACCEL_LOCKED);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
        g_signal_connect_swapped(item, "activate", G_CALLBACK(menu_items[i].callback), menu_items[i].cbdata);
    }

    return menu;
}

static gboolean
gwy_app_data_popup_menu_popup_mouse(GtkWidget *menu,
                                    GdkEventButton *event,
                                    GwyDataView *data_view)
{
    if (event->button != 3)
        return FALSE;

    gwy_app_data_browser_select_data_view(data_view);
    gtk_menu_popup_at_pointer(GTK_MENU(menu), (GdkEvent*)event);

    return TRUE;
}

static void
gwy_app_data_popup_menu_popup_key(GtkWidget *menu,
                                  GtkWidget *data_window)
{
    gtk_menu_popup_at_widget(GTK_MENU(menu), data_window, GDK_GRAVITY_NORTH_WEST, GDK_GRAVITY_NORTH_WEST, NULL);
}

/**
 * gwy_app_data_corner_menu_update:
 * @menu: Data window corner menu.
 * @data_view: The corresponding data view.
 *
 * Updates corner menu to reflect data window's state before we show it.
 **/
static void
gwy_app_data_corner_menu_update(GtkWidget *menu,
                                GwyDataView *data_view)
{
    gboolean realsquare = FALSE;
    GwyContainer *data;
    const gchar *key;
    gchar *s;
    GtkWidget *item;
    GList *l;
    gulong id;
    guint i;

    /* Square mode */
    data = gwy_data_view_get_data(data_view);
    key = gwy_data_view_get_data_prefix(data_view);
    s = g_strconcat(key, "/realsquare", NULL);
    gwy_container_gis_boolean_by_name(data, s, &realsquare);
    gwy_debug("view's realsquare: %d", realsquare);
    g_free(s);

    /* Update stuff */
    l = gtk_container_get_children(GTK_CONTAINER(menu));
    while (l) {
        item = GTK_WIDGET(l->data);
        i = GPOINTER_TO_UINT(g_object_get_qdata(G_OBJECT(item), corner_item_quark));
        switch (i) {
            case ITEM_PIXELSQUARE:
            if (!realsquare) {
                gwy_debug("setting Pixelwise active");
                id = g_signal_handler_find(item, G_SIGNAL_MATCH_FUNC,
                                           0, 0, NULL, gwy_app_data_window_change_square, NULL);
                g_signal_handler_block(item, id);
                gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), TRUE);
                g_signal_handler_unblock(item, id);
            }
            break;

            case ITEM_REALSQUARE:
            if (realsquare) {
                gwy_debug("setting Physical active");
                id = g_signal_handler_find(item, G_SIGNAL_MATCH_FUNC,
                                           0, 0, NULL, gwy_app_data_window_change_square, NULL);
                g_signal_handler_block(item, id);
                gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), TRUE);
                g_signal_handler_unblock(item, id);
            }
            break;

            default:
            break;
        }
        l = g_list_next(l);
    }
}

static gboolean
gwy_app_data_corner_menu_popup_mouse(GtkWidget *menu,
                                     GdkEventButton *event,
                                     GtkWidget *ebox)
{
    GtkWidget *window;
    GwyDataView *data_view;
    GwyAppPage pageno;

    if (event->button != 1)
        return FALSE;

    window = gtk_widget_get_ancestor(ebox, GWY_TYPE_DATA_WINDOW);
    g_return_val_if_fail(window, FALSE);
    data_view = GWY_DATA_VIEW(gwy_data_window_get_data_view(GWY_DATA_WINDOW(window)));
    pageno = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(window), "gwy-app-page"));

    if (pageno == GWY_PAGE_CHANNELS)
        gwy_app_data_browser_select_data_view(data_view);
    else if (pageno == GWY_PAGE_VOLUMES)
        gwy_app_data_browser_select_volume(data_view);
    else if (pageno == GWY_PAGE_XYZS)
        gwy_app_data_browser_select_xyz(data_view);
    else if (pageno == GWY_PAGE_CURVE_MAPS)
        gwy_app_data_browser_select_curve_map(data_view);
    else {
        g_return_val_if_reached(FALSE);
    }
    gwy_app_data_corner_menu_update(menu, data_view);
    gtk_menu_popup_at_pointer(GTK_MENU(menu), (GdkEvent*)event);

    return FALSE;
}

static GtkWidget*
gwy_app_menu_data_corner_create(GtkAccelGroup *accel_group, GwyAppPage pageno)
{
    GtkWidget *menu, *item;
    GtkRadioMenuItem *r;

    corner_item_quark = g_quark_from_static_string("id");

    menu = gtk_menu_new();
    if (accel_group)
        gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);

    item = gtk_radio_menu_item_new_with_mnemonic(NULL, _("Pi_xelwise Square"));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
    g_object_set_qdata(G_OBJECT(item), corner_item_quark, GUINT_TO_POINTER(ITEM_PIXELSQUARE));
    g_object_set_data(G_OBJECT(item), "gwy-app-page", GUINT_TO_POINTER(pageno));
    g_signal_connect(item, "activate", G_CALLBACK(gwy_app_data_window_change_square), GINT_TO_POINTER(FALSE));

    r = GTK_RADIO_MENU_ITEM(item);
    item = gtk_radio_menu_item_new_with_mnemonic_from_widget(r, _("_Physically Square"));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
    g_object_set_qdata(G_OBJECT(item), corner_item_quark, GUINT_TO_POINTER(ITEM_REALSQUARE));
    g_object_set_data(G_OBJECT(item), "gwy-app-page", GUINT_TO_POINTER(pageno));
    g_signal_connect(item, "activate", G_CALLBACK(gwy_app_data_window_change_square), GINT_TO_POINTER(TRUE));

    return menu;
}

static void
gwy_app_data_window_change_square(GtkWidget *item,
                                  gpointer user_data)
{
    gboolean realsquare = GPOINTER_TO_INT(user_data);
    GwyAppPage pageno = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(item), "gwy-app-page"));
    GwyDataView *data_view;
    GwyContainer *data;
    GwyAppWhat what;
    const gchar *key;
    gchar *s;

    if (!gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(item))) {
        gwy_debug("bogus update");
        return;
    }

    gwy_debug("new square mode: %s", realsquare ? "Physical" : "Pixelwise");
    if (pageno == GWY_PAGE_CHANNELS)
        what = GWY_APP_DATA_VIEW;
    else if (pageno == GWY_PAGE_VOLUMES)
        what = GWY_APP_VOLUME_VIEW;
    else if (pageno == GWY_PAGE_XYZS)
        what = GWY_APP_XYZ_VIEW;
    else if (pageno == GWY_PAGE_CURVE_MAPS)
        what = GWY_APP_CURVE_MAP_VIEW;
    else {
        g_return_if_reached();
    }
    gwy_app_data_browser_get_current(what, &data_view, 0);
    data = gwy_data_view_get_data(data_view);
    key = gwy_data_view_get_data_prefix(data_view);
    g_return_if_fail(key);
    s = g_strconcat(key, "/realsquare", NULL);
    if (realsquare)
        gwy_container_set_boolean_by_name(data, s, realsquare);
    else
        gwy_container_remove_by_name(data, s);
    g_free(s);
}

static gboolean
gwy_app_data_key_pressed(G_GNUC_UNUSED GtkWidget *data_window,
                         GdkEventKey *event,
                         G_GNUC_UNUSED gpointer user_data)
{
    if (event->keyval != GDK_KEY_F3 || (event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK)))
        return FALSE;

    if (current_tool) {
        if (!gwy_tool_is_visible(current_tool))
            gwy_tool_show(current_tool);
        else
            gwy_tool_hide(current_tool);
    }
    return TRUE;
}

/********************************************************************************************************************
 *                                                                                                                  *
 *     Graph window management                                                                                      *
 *                                                                                                                  *
 ********************************************************************************************************************/

void
_gwy_app_graph_window_setup(GwyGraphWindow *graph_window,
                            GwyContainer *container, GQuark prefix)
{
    static GtkWidget *popup_menu = NULL;

    GtkWidget *graph, *main_window;
    GtkAccelGroup *accel_group;

    if (!popup_menu && (main_window = gwy_app_main_window_get())) {
        g_return_if_fail(GTK_IS_WINDOW(main_window));
        accel_group = GTK_ACCEL_GROUP(g_object_get_data(G_OBJECT(main_window), "accel_group"));

        if (accel_group && !popup_menu) {
            GList *items;

            popup_menu = gwy_app_build_graph_menu(accel_group);
            items = gtk_container_get_children(GTK_CONTAINER(popup_menu));
            g_list_free(items);
            gtk_widget_show_all(popup_menu);
        }
    }

    gwy_app_add_main_accel_group(GTK_WINDOW(graph_window));
    gwy_help_add_to_window(GTK_WINDOW(graph_window), "graph-windows", NULL, GWY_HELP_DEFAULT);
    restore_other_window_size(GTK_WINDOW(graph_window), container, g_quark_to_string(prefix));

    graph = gwy_graph_window_get_graph(graph_window);
    g_signal_connect_swapped(graph, "button-press-event",
                             G_CALLBACK(gwy_app_graph_popup_menu_popup_mouse), popup_menu);
    g_signal_connect_swapped(gwy_graph_get_area(GWY_GRAPH(graph)), "button-press-event",
                             G_CALLBACK(gwy_app_graph_popup_menu_popup_mouse), popup_menu);
    /* FIXME: Graphs don't get keyboard events. */
    g_signal_connect_swapped(graph, "popup-menu", G_CALLBACK(gwy_app_graph_popup_menu_popup_key), popup_menu);

    g_object_set_data(G_OBJECT(graph_window), "gwy-app-container", container);
    g_object_set_data(G_OBJECT(graph_window), "gwy-app-prefix-quark", GUINT_TO_POINTER(prefix));
    g_signal_connect(graph_window, "configure-event", G_CALLBACK(gwy_app_graph_window_configured), NULL);
    g_signal_connect(graph_window, "size-allocate", G_CALLBACK(gwy_app_graph_window_configured), NULL);
}

static gboolean
gwy_app_graph_window_configured(GwyGraphWindow *window)
{
    GwyContainer *container;
    const gchar *prefix;
    GQuark qprefix;

    container = g_object_get_data(G_OBJECT(window), "gwy-app-container");
    qprefix = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(window), "gwy-app-prefix-quark"));
    g_return_val_if_fail(GWY_IS_CONTAINER(container) && qprefix, FALSE);
    prefix = g_quark_to_string(qprefix);

    save_widget_screen_relative_size(GTK_WIDGET(window), container, prefix, TRUE);

    return FALSE;
}

static gboolean
gwy_app_graph_popup_menu_popup_mouse(GtkWidget *menu,
                                     GdkEventButton *event,
                                     GwyGraph *graph)
{
    if (event->button != 3)
        return FALSE;

    if (GWY_IS_GRAPH_AREA(graph)) {
        graph = GWY_GRAPH(gtk_widget_get_ancestor(GTK_WIDGET(graph), GWY_TYPE_GRAPH));
        if (!graph)
            return FALSE;
    }
    gwy_app_data_browser_select_graph(graph);
    gtk_menu_popup_at_pointer(GTK_MENU(menu), (GdkEvent*)event);

    return TRUE;
}

static void
gwy_app_graph_popup_menu_popup_key(GtkWidget *menu,
                                   GtkWidget *graph)
{
    gtk_menu_popup_at_widget(GTK_MENU(menu), graph, GDK_GRAVITY_NORTH_WEST, GDK_GRAVITY_NORTH_WEST, NULL);
}

/********************************************************************************************************************
 *                                                                                                                  *
 *     OpenGL window                                                                                                *
 *                                                                                                                  *
 ********************************************************************************************************************/

void
_gwy_app_gl_window_setup(GwyGLWindow *glwindow)
{
    GwyGLView *glview;
    GwyContainer *settings;
    gboolean autocrop = FALSE, transparent = FALSE;
    GtkWidget *button, *check, *label;

    gwy_app_add_main_accel_group(GTK_WINDOW(glwindow));

    button = gtk_button_new_with_mnemonic(_("Set as Default"));
    gtk_widget_set_tooltip_text(button, _("Set the current view setup as the default"));
    gwy_gl_window_add_action_widget(glwindow, button);
    g_signal_connect_swapped(button, "clicked", G_CALLBACK(gwy_app_gl_window_set_defaults), glwindow);

    label = gtk_label_new(NULL);
    gtk_widget_set_size_request(label, 12, 0);
    gwy_gl_window_add_action_widget(glwindow, label);

    button = gwy_stock_like_button_new(gwy_sgettext("verb|Save"), GWY_ICON_GTK_SAVE);
    gtk_widget_set_tooltip_text(button, _("Save 3D view to an image"));
    gwy_gl_window_add_action_widget(glwindow, button);
    gwy_gl_window_add_small_toolbar_button(glwindow, GWY_ICON_GTK_SAVE, _("Save 3D view to an image"),
                                           G_CALLBACK(gwy_app_gl_window_export), glwindow);
    g_signal_connect_swapped(button, "clicked", G_CALLBACK(gwy_app_gl_window_export), glwindow);

    settings = gwy_app_settings_get();

    gwy_container_gis_boolean_by_name(settings, "/app/3d/autocrop", &autocrop);
    check = gtk_check_button_new_with_mnemonic(_("_Autocrop"));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), autocrop);
    gwy_gl_window_add_action_widget(glwindow, check);
    gtk_widget_set_tooltip_text(check, _("Remove white borders from exported image"));
    g_object_set_data(G_OBJECT(glwindow), "gwy-app-autocrop-check", check);

    gwy_container_gis_boolean_by_name(settings, "/app/3d/transparent", &transparent);
    check = gtk_check_button_new_with_mnemonic(_("_Transparent"));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), transparent);
    gwy_gl_window_add_action_widget(glwindow, check);
    gtk_widget_set_tooltip_text(check, _("Export PNG images with transparent background"));
    g_object_set_data(G_OBJECT(glwindow), "gwy-app-transparent-check", check);

    gwy_app_gl_window_add_overlay_menu(glwindow);
    gwy_help_add_to_window(GTK_WINDOW(glwindow), "opengl-3d-view", NULL, GWY_HELP_DEFAULT);

    glview = GWY_GL_VIEW(gwy_gl_window_get_gl_view(glwindow));
    restore_other_window_size(GTK_WINDOW(glwindow), gwy_gl_view_get_data(glview), gwy_gl_view_get_setup_prefix(glview));

    g_signal_connect(glwindow, "configure-event", G_CALLBACK(gwy_app_gl_window_configured), NULL);
    g_signal_connect(glwindow, "size-allocate", G_CALLBACK(gwy_app_gl_window_configured), NULL);
}

static gboolean
gwy_app_gl_window_configured(GwyGLWindow *window)
{
    GwyContainer *container;
    const gchar *prefix;
    GwyGLView *view;

    view = GWY_GL_VIEW(gwy_gl_window_get_gl_view(window));
    container = gwy_gl_view_get_data(view);
    prefix = gwy_gl_view_get_setup_prefix(view);
    g_return_val_if_fail(GWY_IS_CONTAINER(container) && prefix, FALSE);

    save_widget_screen_relative_size(GTK_WIDGET(window), container, prefix, TRUE);

    return FALSE;
}

/* a widget for the glwindow as overlay chooser */
static void
gwy_app_gl_window_add_overlay_menu(GwyGLWindow *glwindow)
{
    GtkWidget *menu, *lay;
    GwyGLView *view;
    GQuark data_quark, data2_ref = 0;
    gint* ids, *nids;
    gint activeid = -1;
    const gchar *key;
    gchar refkey[40];
    gboolean showmask = FALSE;
    GwyContainer *settings, *data;

    view = GWY_GL_VIEW(gwy_gl_window_get_gl_view(glwindow));
    lay = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 5);
    menu = gwy_data_chooser_new_channels();
    data = gwy_gl_view_get_data(view);
    data_quark = g_quark_from_string(gwy_gl_view_get_data_key(view));

    gwy_data_chooser_set_filter(GWY_DATA_CHOOSER(menu), gwy_app_gl_window_data2_filter, view, NULL);

    ids = gwy_app_data_browser_get_data_ids(data);
    /* FIXME: This looks suspicious because the data key is usually something like "/0/data" */
    g_strlcpy(refkey, gwy_gl_view_get_data_key(view), 4);
    g_strlcat(refkey, "3d/data2ref", sizeof(refkey));
    if (gwy_container_gis_string_by_name(data, refkey, &key))
        data2_ref = g_quark_from_string(key);

    for (nids = ids; *nids != -1; nids++) {
        if (gwy_app_get_data_key_for_id(*nids) == data2_ref) {
            activeid = *nids;
            break;
        }
        if (gwy_app_get_data_key_for_id(*nids) == data_quark && activeid == -1) {
            activeid = *nids;
        }
    }
    g_free(ids);

    gwy_data_chooser_set_active(GWY_DATA_CHOOSER(menu), data, activeid);
    gwy_app_gl_window_set_data2(glwindow, activeid, FALSE);

    g_signal_connect_swapped(menu, "changed", G_CALLBACK(gwy_app_gl_window_update_chooser), glwindow);

    gtk_box_pack_start(GTK_BOX(lay), menu, FALSE, FALSE, 0);
    g_object_set_data(G_OBJECT(lay), "c", menu);

    menu = gtk_check_button_new_with_mnemonic(_("_Show mask"));
    settings = gwy_app_settings_get();
    gwy_container_gis_boolean_by_name(settings, "/app/3d/show-mask", &showmask);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(menu), showmask);

    g_signal_connect_swapped(menu, "toggled", G_CALLBACK(gwy_app_gl_window_update_chooser), glwindow);
    gtk_box_pack_start(GTK_BOX(lay), menu, FALSE, FALSE, 0);
    g_object_set_data(G_OBJECT(lay), "m", menu);

    gwy_gl_window_set_overlay_chooser(glwindow, lay);
    /* XXX: Gross! It does not take initial state of the checkbox into account.*/
    if (showmask)
        gwy_app_gl_window_update_chooser(glwindow);
}

/* set overlay source to channel id for GwyGLView in glwindow.
 * Show mask if mask==True
 */
static void
gwy_app_gl_window_set_data2(GwyGLWindow *glwindow,
                            gint id,
                            gboolean mask)
{
    GwyGLView *view;
    GwyPixmapLayer *ovplay[2];
    GwyDataViewLayer* dvl;

    view = GWY_GL_VIEW(gwy_gl_window_get_gl_view(glwindow));

    ovplay[0] = gwy_layer_basic_new();
    dvl = GWY_DATA_VIEW_LAYER(ovplay[0]);
    /* FIXME: How are we directly assigning to a data view layer->data? */
    dvl->data = gwy_gl_view_get_data(view);
    g_object_ref(dvl->data);

    gwy_pixmap_layer_set_data_key(ovplay[0], g_quark_to_string(gwy_app_get_data_key_for_id(id)));
    gwy_layer_basic_set_gradient_key(GWY_LAYER_BASIC(ovplay[0]), gwy_gl_view_get_gradient_key(view));
    gwy_layer_basic_set_min_max_key(GWY_LAYER_BASIC(ovplay[0]),
                                    g_quark_to_string(gwy_app_get_data_base_key_for_id(id)));
    gwy_layer_basic_set_range_type_key(GWY_LAYER_BASIC(ovplay[0]),
                                       g_quark_to_string(gwy_app_get_data_range_type_key_for_id(id)));

    if (mask) {
        ovplay[1] = gwy_layer_mask_new();
        dvl = GWY_DATA_VIEW_LAYER(ovplay[1]);
        /* FIXME: How are we directly assigning to a data view layer->data? */
        dvl->data = gwy_gl_view_get_data(view);
        g_object_ref(dvl->data);
        gwy_pixmap_layer_set_data_key(ovplay[1], g_quark_to_string(gwy_app_get_mask_key_for_id(id)));
        gwy_layer_mask_set_color_key(GWY_LAYER_MASK(ovplay[1]), g_quark_to_string(gwy_app_get_mask_key_for_id(id)));
        gwy_gl_view_set_ovlay(view, ovplay, 2);
    }
    else {
        gwy_gl_view_set_ovlay(view, ovplay, 1);
    }
}

/* callback for the chooser created by gwy_app_gl_window_add_overlay_menu */
static void
gwy_app_gl_window_update_chooser(GwyGLWindow *glwindow)
{
    gint id;
    gboolean mask = FALSE;
    GtkWidget *temp;
    guchar refkey[40];
    const guchar *name;
    GQuark key;
    GwyGLView *view;
    GtkWidget *dataov_menu = gwy_gl_window_get_overlay_chooser(glwindow);

    temp = g_object_get_data(G_OBJECT(dataov_menu), "m");

    mask = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(temp));
    temp = g_object_get_data(G_OBJECT(dataov_menu), "c");

    gwy_data_chooser_get_active(GWY_DATA_CHOOSER(temp), &id);
    key = gwy_app_get_data_key_for_id(id);
    name = g_quark_to_string(key);

    view = GWY_GL_VIEW(gwy_gl_window_get_gl_view(glwindow));
    key = gwy_app_get_data_key_for_id(id);
    name = g_quark_to_string(key);
    /* FIXME: This looks suspicious because the data key is usually something like "/0/data" */
    g_strlcpy(refkey, gwy_gl_view_get_data_key(view), 4);
    g_strlcat(refkey, "3d/data2ref", sizeof(refkey));
    gwy_container_set_string_by_name(gwy_gl_view_get_data(view), refkey, g_strdup(name));

    gwy_app_gl_window_set_data2(glwindow, id, mask);
}

static gboolean
gwy_app_gl_window_data2_filter(GwyContainer *data2,
                               gint id2,
                               gpointer user_data)
{
    GwyGLView* view;
    GwyDataField *data_field1, *data_field2;
    GQuark quark2;
    GwyContainer* data1;

    quark2 = gwy_app_get_data_key_for_id(id2);
    data_field2 = gwy_container_get_object(data2, quark2);
    view = GWY_GL_VIEW(user_data);
    data1 = gwy_gl_view_get_data(view);
    if (data1 != data2)
        return FALSE;

    gwy_container_gis_object_by_name(data1, gwy_gl_view_get_data_key(view), &data_field1);

    return !gwy_data_field_check_compatibility(data_field2, data_field1, GWY_DATA_COMPATIBILITY_RES);
}

static GdkPixbuf*
autocrop_gl_export_pixbuf(GdkPixbuf *pixbuf)
{
    gint width, height, stride, bpp, i, j, top, left, right, bot;
    guchar *data, *row;

    width = gdk_pixbuf_get_width(pixbuf);
    height = gdk_pixbuf_get_height(pixbuf);
    stride = gdk_pixbuf_get_rowstride(pixbuf);
    bpp = gdk_pixbuf_get_n_channels(pixbuf);
    data = gdk_pixbuf_get_pixels(pixbuf);

    top = height-1;
    bot = 0;
    left = width-1;
    right = 0;

    if (bpp == 4) {
        for (i = 0; i < height; i++) {
            row = data + i*stride;
            for (j = 0; j < width; j++, row += bpp) {
                if (row[3]) {
                    top = MIN(top, i);
                    bot = MAX(bot, i);
                    left = MIN(left, j);
                    right = MAX(right, j);
                }
            }
        }
    }
    else {
        for (i = 0; i < height; i++) {
            row = data + i*stride;
            for (j = 0; j < bpp*width; j++) {
                if (row[j] != 0xff) {
                    top = MIN(top, i);
                    bot = MAX(bot, i);
                    left = MIN(left, j/bpp);
                    right = MAX(right, j/bpp);
                }
            }
        }
    }

    if (top > bot || left > right)
        top = bot = left = right = 0;

    return gdk_pixbuf_new_subpixbuf(pixbuf, left, top, right+1 - left, bot+1 - top);
}

static void
gwy_app_save_gl_export(GtkWidget *dialog,
                       gint response,
                       GwyGLWindow *glwindow)
{
    gchar *filename_sys, *filename_utf8, *s, *filetype = NULL;
    GdkPixbuf *pixbuf;
    GError *err = NULL;

    pixbuf = g_object_get_data(G_OBJECT(dialog), "pixbuf-to-save");
    if (response != GTK_RESPONSE_OK) {
        g_object_unref(pixbuf);
        if (response != GTK_RESPONSE_NONE)
            gtk_widget_destroy(dialog);
        return;
    }

    filename_sys = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
    gtk_widget_destroy(dialog);

    filename_utf8 = g_filename_to_utf8(filename_sys, -1, NULL, NULL, NULL);
    if ((s = strrchr(filename_utf8, '.'))) {
        filetype = g_ascii_strdown(s+1, -1);
        if (gwy_stramong(filetype, "jpg", "jpeg", NULL))
            gwy_assign_string(&filetype, "jpeg");
        else if (gwy_stramong(filetype, "tif", "tiff", NULL))
            gwy_assign_string(&filetype, "tiff");
    }
    if (!gdk_pixbuf_save(pixbuf, filename_sys, filetype ? filetype : "png", &err, NULL)) {
        dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
                                        _("Saving of 3D view to `%s' failed"), filename_utf8);
        gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog), "%s", err->message);
        if (gtk_dialog_run(GTK_DIALOG(dialog)) != GTK_RESPONSE_NONE)
            gtk_widget_destroy(dialog);
        g_clear_error(&err);
    }
    g_free(filetype);
    g_free(filename_sys);
    g_object_unref(pixbuf);
    g_free(g_object_get_data(G_OBJECT(glwindow), "gwy-app-export-filename"));
    g_object_set_data(G_OBJECT(glwindow), "gwy-app-export-filename", filename_utf8);
}

static GdkPixbuf*
gwy_app_gl_window_get_pixbuf(GwyGLWindow *glwindow,
                             gboolean autocrop, gboolean transparent)
{
    GdkPixbuf *pixbuf, *cropped_pixbuf;
    GtkWidget *glview;

    glview = gwy_gl_window_get_gl_view(GWY_GL_WINDOW(glwindow));

    if (transparent)
        pixbuf = gwy_gl_view_get_pixbuf_with_alpha(GWY_GL_VIEW(glview));
    else
        pixbuf = gwy_gl_view_get_pixbuf(GWY_GL_VIEW(glview));

    if (!autocrop)
        return pixbuf;

    cropped_pixbuf = autocrop_gl_export_pixbuf(pixbuf);
    g_object_unref(pixbuf);
    return cropped_pixbuf;
}

static void
gwy_app_gl_window_export(GwyGLWindow *glwindow)
{
    GwyContainer *data, *settings;
    GtkWidget *dialog, *glview, *check;
    const gchar *filename_utf8;
    gchar *filename_sys;
    GdkPixbuf *pixbuf;
    gboolean need_free_utf = FALSE, autocrop, transparent;

    glview = gwy_gl_window_get_gl_view(glwindow);
    data = gwy_gl_view_get_data(GWY_GL_VIEW(glview));

    check = g_object_get_data(G_OBJECT(glwindow), "gwy-app-autocrop-check");
    autocrop = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(check));
    check = g_object_get_data(G_OBJECT(glwindow), "gwy-app-transparent-check");
    transparent = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(check));
    settings = gwy_app_settings_get();
    gwy_container_set_boolean_by_name(settings, "/app/3d/autocrop", autocrop);
    gwy_container_set_boolean_by_name(settings, "/app/3d/transparent", transparent);
    pixbuf = gwy_app_gl_window_get_pixbuf(glwindow, autocrop, transparent);

    filename_utf8 = g_object_get_data(G_OBJECT(glwindow), "gwy-app-export-filename");
    if (!filename_utf8) {
        if (gwy_container_gis_string_by_name(data, "/filename", &filename_utf8)) {
            /* FIXME: this is ugly, invent a better filename */
            filename_utf8 = g_strconcat(filename_utf8, ".png", NULL);
            need_free_utf = TRUE;
        }
        else
            filename_utf8 = "3d.png";
    }
    filename_sys = g_filename_from_utf8(filename_utf8, -1, NULL, NULL, NULL);
    if (need_free_utf)
        g_free((gpointer)filename_utf8);

    dialog = gtk_file_chooser_dialog_new(_("Export 3D View"), GTK_WINDOW(glwindow),
                                         GTK_FILE_CHOOSER_ACTION_SAVE, NULL, NULL);
    gwy_add_cancel_button_to_dialog(GTK_DIALOG(dialog));
    gwy_add_ok_button_to_dialog(GTK_DIALOG(dialog));
    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
    gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), gwy_app_get_current_directory());
    gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(dialog), filename_sys);
    gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(dialog), TRUE);
    g_free(filename_sys);

    g_object_set_data(G_OBJECT(dialog), "pixbuf-to-save", pixbuf);
    g_signal_connect(dialog, "response", G_CALLBACK(gwy_app_save_gl_export), glwindow);
    gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
    gtk_widget_show_all(dialog);
}

static void
save_or_restore_object_properties(GObject *object,
                                  GwyContainer *settings,
                                  gboolean restore, GString *key)
{
    GParamSpec **pspecs;
    gdouble dblvalue;
    gboolean boolvalue;
    gint enumvalue;
    const gchar *name, *strvalue;
    gchar *newstrvalue;
    guint len, i, nspecs;
    GType valtype;

    pspecs = g_object_class_list_properties(G_OBJECT_GET_CLASS(object),
                                            &nspecs);
    if (!g_str_has_suffix(key->str, "/"))
        g_string_append_c(key, '/');
    len = key->len;

    for (i = 0; i < nspecs; i++) {
        if (!(pspecs[i]->flags & G_PARAM_WRITABLE) || (pspecs[i]->flags & (G_PARAM_CONSTRUCT_ONLY | G_PARAM_PRIVATE)))
            continue;

        valtype = pspecs[i]->value_type;
        name = pspecs[i]->name;
        g_string_truncate(key, len);
        g_string_append(key, name);

        if (valtype == G_TYPE_BOOLEAN) {
            if (restore) {
                if (gwy_container_gis_boolean_by_name(settings, key->str, &boolvalue))
                    g_object_set(object, name, boolvalue, NULL);
            }
            else {
                g_object_get(object, name, &boolvalue, NULL);
                gwy_container_set_boolean_by_name(settings, key->str, boolvalue);
            }
        }
        else if (valtype == G_TYPE_DOUBLE) {
            if (restore) {
                if (gwy_container_gis_double_by_name(settings, key->str, &dblvalue))
                    g_object_set(object, name, dblvalue, NULL);
            }
            else {
                g_object_get(object, name, &dblvalue, NULL);
                gwy_container_set_double_by_name(settings, key->str, dblvalue);
            }
        }
        else if (valtype == G_TYPE_STRING) {
            if (restore) {
                if (gwy_container_gis_string_by_name(settings, key->str, &strvalue))
                    g_object_set(object, name, strvalue, NULL);
            }
            else {
                g_object_get(object, name, &newstrvalue, NULL);
                gwy_container_set_string_by_name(settings, key->str, newstrvalue);
            }
        }
        else if (g_type_is_a(valtype, G_TYPE_ENUM)) {
            if (restore) {
                if (gwy_container_gis_enum_by_name(settings, key->str, &enumvalue))
                    g_object_set(object, name, enumvalue, NULL);
            }
            else {
                g_object_get(object, name, &enumvalue, NULL);
                gwy_container_set_enum_by_name(settings, key->str, enumvalue);
            }
        }
        else {
            g_warning("Unhandled property %s", name);
        }
    }

    g_free(pspecs);
}

static void
gwy_app_gl_window_set_defaults(GwyGLWindow *window)
{
    GwyGLView *view;
    GwyGLSetup *setup;
    GtkToggleButton *toggle;
    GwyContainer *settings, *data;
    const gchar *prefix;
    GString *str;
    guint i;

    view = (GwyGLView*)gwy_gl_window_get_gl_view(window);
    setup = gwy_gl_view_get_setup(view);
    g_return_if_fail(GWY_IS_GL_SETUP(setup));
    data = gwy_gl_view_get_data(view);

    settings = gwy_app_settings_get();
    str = g_string_new(NULL);

    g_string_assign(str, "/app/3d/");
    save_or_restore_object_properties(G_OBJECT(setup), settings, FALSE, str);

    GtkWidget *dataov_menu = gwy_gl_window_get_overlay_chooser(window);
    toggle = GTK_TOGGLE_BUTTON(g_object_get_data(G_OBJECT(dataov_menu), "m"));
    gwy_container_set_boolean_by_name(settings, "/app/3d/show-mask", gtk_toggle_button_get_active(toggle));

    prefix = gwy_gl_view_get_setup_prefix(view);
    for (i = 0; i < G_N_ELEMENTS(gl_labels); i++) {
        GwyGLLabel *label = NULL;

        g_string_printf(str, "%s/%s", prefix, gl_labels[i].key);
        gwy_container_gis_object_by_name(data, str->str, &label);
        if (!label)
            continue;

        g_string_printf(str, "/app/3d/labels/%s/", gl_labels[i].key);
        save_or_restore_object_properties(G_OBJECT(label), settings, FALSE, str);
    }

    g_string_free(str, TRUE);
}

gboolean
_gwy_app_gl_view_init_setup(GwyContainer *container,
                            const gchar *setup_prefix)
{
    GwyContainer *settings;
    GwyGLSetup *setup;
    GString *str;
    guint i;

    g_return_val_if_fail(GWY_IS_CONTAINER(container), FALSE);
    g_return_val_if_fail(setup_prefix, FALSE);

    str = g_string_new(setup_prefix);
    g_string_append(str, "/setup");
    if (gwy_container_gis_object_by_name(container, str->str, &setup)
        && GWY_IS_GL_SETUP(setup)) {
        g_string_free(str, TRUE);
        return FALSE;
    }

    setup = gwy_gl_setup_new();
    gwy_container_set_object_by_name(container, str->str, setup);

    settings = gwy_app_settings_get();
    g_string_assign(str, "/app/3d/");
    save_or_restore_object_properties(G_OBJECT(setup), settings, TRUE, str);
    g_object_unref(setup);

    for (i = 0; i < G_N_ELEMENTS(gl_labels); i++) {
        GwyGLLabel *label = NULL;
        g_string_printf(str, "%s/%s", setup_prefix, gl_labels[i].key);
        gwy_container_gis_object_by_name(container, str->str, &label);
        if (label)
            continue;

        label = gwy_gl_label_new(gl_labels[i].default_text);
        gwy_container_set_object_by_name(container, str->str, label);

        g_string_printf(str, "/app/3d/labels/%s/", gl_labels[i].key);
        save_or_restore_object_properties(G_OBJECT(label), settings, TRUE, str);
        g_object_unref(label);
    }

    g_string_free(str, TRUE);

    return TRUE;
}

/********************************************************************************************************************
 *                                                                                                                  *
 *     Spectra                                                                                                      *
 *                                                                                                                  *
 ********************************************************************************************************************/

void
_gwy_app_spectra_set_current(GwySpectra *spectra)
{
    if (current_tool)
        gwy_tool_spectra_switched(current_tool, spectra);
}

/********************************************************************************************************************
 *                                                                                                                  *
 *     Bricks                                                                                                       *
 *                                                                                                                  *
 ********************************************************************************************************************/

void
_gwy_app_brick_window_setup(GwyDataWindow *data_window)
{
    static GtkWidget *popup_menu = NULL;

    GtkAccelGroup *accel_group;
    GtkWidget *data_view, *main_window, *vbox, *hbox, *button, *label;

    if (!popup_menu && (main_window = gwy_app_main_window_get())) {
        g_return_if_fail(GTK_IS_WINDOW(main_window));
        accel_group = GTK_ACCEL_GROUP(g_object_get_data(G_OBJECT(main_window), "accel_group"));

        if (accel_group && !popup_menu) {
            popup_menu = gwy_app_menu_brick_popup_create(accel_group);
            gtk_widget_show_all(popup_menu);
        }
    }

    gwy_app_add_main_accel_group(GTK_WINDOW(data_window));
    gwy_help_add_to_window(GTK_WINDOW(data_window), "volume-data", NULL, GWY_HELP_DEFAULT);
    restore_data_window_zoom(data_window);

    data_view = gwy_data_window_get_data_view(data_window);
    g_signal_connect_swapped(data_view, "button-press-event",
                             G_CALLBACK(gwy_app_brick_popup_menu_popup_mouse), popup_menu);
    g_signal_connect_swapped(data_window, "popup-menu", G_CALLBACK(gwy_app_brick_popup_menu_popup_key), popup_menu);

    hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);

    vbox = gtk_bin_get_child(GTK_BIN(data_window));
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);
    gtk_box_reorder_child(GTK_BOX(vbox), hbox, 0);

    button = gtk_button_new_with_mnemonic(_("_Change Preview"));
    gtk_widget_set_can_focus(button, FALSE);
    gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    g_signal_connect_swapped(button, "clicked", G_CALLBACK(change_brick_preview), data_window);

    label = gtk_label_new(NULL);
    gtk_label_set_xalign(GTK_LABEL(label), 0.0);
    gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);
    gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);

    g_object_set_data(G_OBJECT(data_window), "gwy-brick-info", label);

    g_signal_connect(data_window, "configure-event", G_CALLBACK(gwy_app_brick_window_configured), NULL);
    g_signal_connect(data_window, "size-allocate", G_CALLBACK(gwy_app_brick_window_configured), NULL);
}

static gboolean
gwy_app_brick_window_configured(GwyDataWindow *window)
{
    GwyDataView *view = GWY_DATA_VIEW(gwy_data_window_get_data_view(window));
    GwyContainer *container = gwy_data_view_get_data(view);
    const gchar *prefix = gwy_data_view_get_data_prefix(view);
    gchar *key;

    g_return_val_if_fail(container, FALSE);
    g_return_val_if_fail(prefix, FALSE);

    /* This leads to some odd keys containing .../preview/view/... */
    key = g_strconcat(prefix, "/view/scale", NULL);
    gwy_container_set_double_by_name(container, key, gwy_data_view_get_real_zoom(view));
    g_free(key);

    save_widget_screen_relative_size(GTK_WIDGET(view), container, prefix, FALSE);

    return FALSE;
}

static gboolean
brick_preview_filter(GwyContainer *data,
                     gint id,
                     gpointer user_data)
{
    BrickPreviewGUI *gui = (BrickPreviewGUI*)user_data;
    GwyDataField *field = gwy_container_get_object(data, gwy_app_get_data_key_for_id(id));

    if (gwy_data_field_check_compatibility_with_brick_xy(field, gui->brick,
                                                         GWY_DATA_COMPATIBILITY_RES
                                                         | GWY_DATA_COMPATIBILITY_REAL
                                                         | GWY_DATA_COMPATIBILITY_LATERAL))
        return FALSE;

    return TRUE;
}

static GwyParamDef*
define_brick_preview_parameters(GwyBrick *brick)
{
    static const GwyEnum types[] = {
        { N_("Mean"),          BRICK_PREVIEW_MEAN,    },
        { N_("Minimum"),       BRICK_PREVIEW_MINIMUM, },
        { N_("Maximum"),       BRICK_PREVIEW_MAXIMUM, },
        { N_("Min. position"), BRICK_PREVIEW_MINPOS,  },
        { N_("Max. position"), BRICK_PREVIEW_MAXPOS,  },
        { N_("RMS"),           BRICK_PREVIEW_RMS,     },
        { N_("Image:"),        BRICK_PREVIEW_IMAGE,   },
        { N_("Section:"),      BRICK_PREVIEW_SECTION, },
    };
    GwyParamDef *pardef = gwy_param_def_new();

    gwy_param_def_add_gwyenum(pardef, PARAM_PREVIEW_TYPE, NULL, _("Preview quantity"),
                              types, G_N_ELEMENTS(types), BRICK_PREVIEW_MEAN);
    gwy_param_def_add_int(pardef, PARAM_PREVIEW_LEVEL, NULL, NULL, 0, gwy_brick_get_zres(brick)-1, 0);
    gwy_param_def_add_image_id(pardef, PARAM_PREVIEW_IMAGE, NULL, NULL);
    gwy_param_def_add_instant_updates(pardef, PARAM_PREVIEW_UPDATE, NULL, NULL, FALSE);

    return pardef;
}

static void
recalculate_brick_preview(BrickPreviewGUI *gui)
{
    GwyBrick *brick = gui->brick;
    BrickPreviewType type = gwy_params_get_enum(gui->params, PARAM_PREVIEW_TYPE);
    GwyDataField *preview = gui->preview;

    if (type == BRICK_PREVIEW_MEAN)
        gwy_brick_mean_xy_plane(brick, preview);
    else if (type == BRICK_PREVIEW_MINIMUM)
        gwy_brick_min_xy_plane(brick, preview);
    else if (type == BRICK_PREVIEW_MAXIMUM)
        gwy_brick_max_xy_plane(brick, preview);
    else if (type == BRICK_PREVIEW_MINPOS)
        gwy_brick_minpos_xy_plane(brick, preview);
    else if (type == BRICK_PREVIEW_MAXPOS)
        gwy_brick_maxpos_xy_plane(brick, preview);
    else if (type == BRICK_PREVIEW_RMS)
        gwy_brick_rms_xy_plane(brick, preview);
    else if (type == BRICK_PREVIEW_IMAGE)
        gwy_data_field_assign(preview, gwy_params_get_image(gui->params, PARAM_PREVIEW_IMAGE));
    else if (type == BRICK_PREVIEW_SECTION)
        gwy_brick_extract_xy_plane(brick, preview, gwy_params_get_int(gui->params, PARAM_PREVIEW_LEVEL));
    else {
        g_assert_not_reached();
    }
    gwy_container_set_object(gui->data, gui->quark, gui->preview);
    gwy_data_field_data_changed(gui->preview);
}

static void
brick_preview_param_changed(BrickPreviewGUI *gui, gint id, GwyParamTable *table)
{
    gboolean has_image = !!gwy_params_get_image(gui->params, PARAM_PREVIEW_IMAGE);
    BrickPreviewType type = gwy_params_get_enum(gui->params, PARAM_PREVIEW_TYPE);

    if (id < 0 && !has_image) {
        if (type == BRICK_PREVIEW_IMAGE)
            gwy_param_table_set_enum(table, PARAM_PREVIEW_TYPE, (type = BRICK_PREVIEW_MEAN));
        gwy_param_table_radio_set_sensitive(table, PARAM_PREVIEW_TYPE, BRICK_PREVIEW_IMAGE, FALSE);
    }
    if (id < 0 || id == PARAM_PREVIEW_TYPE) {
        gwy_param_table_set_sensitive(table, PARAM_PREVIEW_LEVEL, type == BRICK_PREVIEW_SECTION);
        gwy_param_table_set_sensitive(table, PARAM_PREVIEW_IMAGE, type == BRICK_PREVIEW_IMAGE);
    }
    gwy_dialog_invalidate(GWY_DIALOG(gui->dialog));
}

static void
brick_preview_preview(gpointer user_data)
{
    BrickPreviewGUI *gui = (BrickPreviewGUI*)user_data;

    /* GwyDialog runs an initial preview on startup, but we do not really want it here without instant updates enabled
     * because it just feels disorienting. */
    if (gui->initial_preview) {
        gui->initial_preview = FALSE;
        return;
    }
    recalculate_brick_preview(gui);
    gwy_dialog_have_result(GWY_DIALOG(gui->dialog));
}

static void
change_brick_preview(GwyDataWindow *data_window)
{
    static const guint simple_types[] = {
        BRICK_PREVIEW_MEAN, BRICK_PREVIEW_MINIMUM, BRICK_PREVIEW_MAXIMUM,
        BRICK_PREVIEW_MINPOS, BRICK_PREVIEW_MAXPOS,
        BRICK_PREVIEW_RMS,
    };
    BrickPreviewGUI gui;
    GwyDialog *dialog;
    GwyParamDef *pardef;
    GwyParams *params;
    GwyParamTable *table;
    GwyContainer *data;
    GwyBrick *brick;
    GwyDialogOutcome outcome;
    gboolean update = FALSE;
    gint id;
    guint i;

    gwy_app_data_browser_get_current(GWY_APP_BRICK, &brick,
                                     GWY_APP_BRICK_ID, &id,
                                     GWY_APP_CONTAINER, &data,
                                     0);
    g_return_if_fail(GWY_IS_BRICK(brick));
    g_return_if_fail(id >= 0);

    gui.brick = brick;
    gui.data = data;
    gui.id = id;
    gui.quark = gwy_app_get_brick_preview_key_for_id(id);
    gui.preview = gwy_container_get_object(data, gui.quark);
    gui.old_preview = gwy_data_field_copy(gui.preview);

    pardef = define_brick_preview_parameters(brick);
    params = gui.params = gwy_params_new();
    gwy_params_set_def(params, pardef);

    /* These two are safe because zeros (from NULL) are good defaults. */
    gwy_params_set_enum(params, PARAM_PREVIEW_TYPE,
                        GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(brick), "gwy-preview-type")));
    gwy_params_set_int(params, PARAM_PREVIEW_LEVEL,
                       GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(brick), "gwy-preview-level")));
    if (gwy_container_gis_boolean_by_name(gwy_app_settings_get(), "/app/brick-preview/update", &update))
        gwy_params_set_boolean(params, PARAM_PREVIEW_UPDATE, update);
    gui.initial_preview = !update;

    gui.dialog = gwy_dialog_new(_("Change Volume Data Preview"));
    dialog = GWY_DIALOG(gui.dialog);
    gtk_window_set_transient_for(GTK_WINDOW(dialog), GTK_WINDOW(data_window));
    gwy_dialog_add_buttons(dialog, GTK_RESPONSE_CANCEL, GTK_RESPONSE_OK, 0);

    table = gwy_param_table_new(params);
    gwy_param_table_append_radio_header(table, PARAM_PREVIEW_TYPE);
    for (i = 0; i < G_N_ELEMENTS(simple_types); i++)
        gwy_param_table_append_radio_item(table, PARAM_PREVIEW_TYPE, simple_types[i]);
    gwy_param_table_append_radio_item(table, PARAM_PREVIEW_TYPE, BRICK_PREVIEW_IMAGE);
    gwy_param_table_append_image_id(table, PARAM_PREVIEW_IMAGE);
    gwy_param_table_data_id_set_filter(table, PARAM_PREVIEW_IMAGE, brick_preview_filter, &gui, NULL);
    gwy_param_table_append_radio_item(table, PARAM_PREVIEW_TYPE, BRICK_PREVIEW_SECTION);
    gwy_param_table_append_slider(table, PARAM_PREVIEW_LEVEL);
    gwy_param_table_append_separator(table);
    gwy_param_table_append_checkbox(table, PARAM_PREVIEW_UPDATE);

    gwy_dialog_add_content(dialog, gwy_param_table_widget(table), TRUE, TRUE, 0);
    gwy_dialog_add_param_table(dialog, table);
    gwy_dialog_set_preview_func(dialog, GWY_PREVIEW_IMMEDIATE, brick_preview_preview, &gui, NULL);
    g_signal_connect_swapped(table, "param-changed", G_CALLBACK(brick_preview_param_changed), &gui);

    outcome = gwy_dialog_run(dialog);
    if (outcome == GWY_DIALOG_CANCEL) {
        /* Make sure we get back the old preview, regardless what the user did in the dialog. */
        gwy_data_field_assign(gui.preview, gui.old_preview);
        gwy_data_field_data_changed(gui.preview);
        goto end;
    }
    if (outcome != GWY_DIALOG_HAVE_RESULT)
        recalculate_brick_preview(&gui);

    g_object_set_data(G_OBJECT(brick), "gwy-preview-type",
                      GUINT_TO_POINTER(gwy_params_get_enum(params, PARAM_PREVIEW_TYPE)));
    g_object_set_data(G_OBJECT(brick), "gwy-preview-level",
                      GUINT_TO_POINTER(gwy_params_get_int(params, PARAM_PREVIEW_LEVEL)));
    gwy_container_set_boolean_by_name(gwy_app_settings_get(),
                                      "/app/brick-preview/update",
                                      gwy_params_get_boolean(params, PARAM_PREVIEW_UPDATE));

end:
    g_object_unref(gui.old_preview);
    g_object_unref(params);
    g_object_unref(pardef);
}

static GtkWidget*
gwy_app_menu_brick_popup_create(GtkAccelGroup *accel_group)
{
    static struct {
        const gchar *label;
        gpointer callback;
        gpointer cbdata;
        guint key;
        GdkModifierType mods;
    }
    const menu_items[] = {
        {
            N_("Zoom _1:1"), gwy_app_any_window_reset_zoom, GUINT_TO_POINTER(GWY_APP_VOLUME_VIEW),
            0, 0
        },
        {
            N_("Metadata _Browser..."), metadata_browser, GUINT_TO_POINTER(GWY_APP_BRICK),
            GDK_KEY_B, GDK_CONTROL_MASK | GDK_SHIFT_MASK
        },
        {
            N_("View _Log..."), log_browser, GUINT_TO_POINTER(GWY_APP_BRICK),
            0, 0
        },
    };
    GwySensitivityGroup *sensgroup;
    GtkWidget *menu, *item;
    const gchar *name;
    guint i, mask;

    menu = gtk_menu_new();
    if (accel_group)
        gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    sensgroup = gwy_app_sensitivity_get_group();
    for (i = 0; i < G_N_ELEMENTS(menu_items); i++) {
        if (menu_items[i].callback == gwy_app_run_volume_func
            && !gwy_volume_func_get_run_types((gchar*)menu_items[i].cbdata)) {
            g_warning("Brick function <%s> for data view context menu is not available.",
                      (const gchar*)menu_items[i].cbdata);
            continue;
        }
        if (menu_items[i].callback == gwy_app_run_volume_func) {
            name = _(gwy_volume_func_get_menu_path(menu_items[i].cbdata));
            name = strrchr(name, '/');
            if (!name) {
                g_warning("Invalid translated menu path for <%s>", (const gchar*)menu_items[i].cbdata);
                continue;
            }
            item = gtk_menu_item_new_with_mnemonic(name + 1);
            mask = gwy_volume_func_get_sensitivity_mask(menu_items[i].cbdata);
            gwy_sensitivity_group_add_widget(sensgroup, item, mask);
        }
        else {
            item = gtk_menu_item_new_with_mnemonic(_(menu_items[i].label));
        }

        if (menu_items[i].key)
            gtk_widget_add_accelerator(item, "activate", accel_group,
                                       menu_items[i].key, menu_items[i].mods, GTK_ACCEL_VISIBLE | GTK_ACCEL_LOCKED);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
        g_signal_connect_swapped(item, "activate", G_CALLBACK(menu_items[i].callback), menu_items[i].cbdata);
    }

    return menu;
}

static gboolean
gwy_app_brick_popup_menu_popup_mouse(GtkWidget *menu,
                                     GdkEventButton *event,
                                     GwyDataView *data_view)
{
    if (event->button != 3)
        return FALSE;

    gwy_app_data_browser_select_volume(data_view);
    gtk_menu_popup_at_pointer(GTK_MENU(menu), (GdkEvent*)event);

    return TRUE;
}

static void
gwy_app_brick_popup_menu_popup_key(GtkWidget *menu,
                                   GtkWidget *data_window)
{
    gtk_menu_popup_at_widget(GTK_MENU(menu), data_window, GDK_GRAVITY_NORTH_WEST, GDK_GRAVITY_NORTH_WEST, NULL);
}

/********************************************************************************************************************
 *                                                                                                                  *
 *     XYZs                                                                                                         *
 *                                                                                                                  *
 ********************************************************************************************************************/

void
_gwy_app_surface_window_setup(GwyDataWindow *data_window)
{
    static GtkWidget *popup_menu = NULL;

    GtkAccelGroup *accel_group;
    GtkWidget *data_view, *main_window, *vbox, *hbox, *button, *label;

    if (!popup_menu && (main_window = gwy_app_main_window_get())) {
        g_return_if_fail(GTK_IS_WINDOW(main_window));
        accel_group = GTK_ACCEL_GROUP(g_object_get_data(G_OBJECT(main_window), "accel_group"));

        if (accel_group && !popup_menu) {
            popup_menu = gwy_app_menu_surface_popup_create(accel_group);
            gtk_widget_show_all(popup_menu);
        }
    }

    gwy_app_add_main_accel_group(GTK_WINDOW(data_window));
    gwy_help_add_to_window(GTK_WINDOW(data_window), "xyz-data", NULL, GWY_HELP_DEFAULT);

    data_view = gwy_data_window_get_data_view(data_window);
    restore_other_window_size(GTK_WINDOW(data_window),
                              gwy_data_view_get_data(GWY_DATA_VIEW(data_view)),
                              gwy_data_view_get_data_prefix(GWY_DATA_VIEW(data_view)));

    g_signal_connect_swapped(data_view, "button-press-event",
                             G_CALLBACK(gwy_app_surface_popup_menu_popup_mouse), popup_menu);
    g_signal_connect_swapped(data_window, "popup-menu", G_CALLBACK(gwy_app_surface_popup_menu_popup_key), popup_menu);

    hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);

    vbox = gtk_bin_get_child(GTK_BIN(data_window));
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);
    gtk_box_reorder_child(GTK_BOX(vbox), hbox, 0);

    button = gtk_button_new_with_mnemonic(_("_Update Preview"));
    gtk_widget_set_can_focus(button, FALSE);
    gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    g_signal_connect_swapped(button, "clicked", G_CALLBACK(update_surface_preview), data_window);

    button = gtk_toggle_button_new_with_mnemonic(_("_Density Map"));
    gtk_widget_set_can_focus(button, FALSE);
    gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    g_signal_connect_swapped(button, "toggled", G_CALLBACK(surface_density_map_toggled), data_window);

    label = gtk_label_new(NULL);
    gtk_label_set_xalign(GTK_LABEL(label), 0.0);
    gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);
    gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);

    g_object_set_data(G_OBJECT(data_window), "gwy-surface-info", label);

    g_signal_connect(data_window, "configure-event", G_CALLBACK(gwy_app_surface_window_configured), NULL);
    g_signal_connect(data_window, "size-allocate", G_CALLBACK(gwy_app_surface_window_configured), NULL);
}

static gboolean
gwy_app_surface_window_configured(GwyDataWindow *window)
{
    GwyDataView *view = GWY_DATA_VIEW(gwy_data_window_get_data_view(window));
    GwyContainer *container = gwy_data_view_get_data(view);
    const gchar *prefix = gwy_data_view_get_data_prefix(view);

    g_return_val_if_fail(container, FALSE);
    g_return_val_if_fail(prefix, FALSE);

    save_widget_screen_relative_size(GTK_WIDGET(window), container, prefix, TRUE);

    return FALSE;
}

static void
update_surface_preview(GwyDataWindow *data_window)
{
    GwyPreviewSurfaceFlags flags = GWY_PREVIEW_SURFACE_FILL;
    GwyDataView *data_view = GWY_DATA_VIEW(gwy_data_window_get_data_view(data_window));
    GwyContainer *data;
    GwySurface *surface;
    GwyDataField *raster;
    gint id;
    GQuark quark;

    gwy_app_data_browser_get_current(GWY_APP_SURFACE, &surface,
                                     GWY_APP_SURFACE_ID, &id,
                                     GWY_APP_CONTAINER, &data,
                                     0);
    g_return_if_fail(GWY_IS_SURFACE(surface));
    g_return_if_fail(id >= 0);

    quark = gwy_app_get_surface_preview_key_for_id(id);
    raster = gwy_container_get_object(data, quark);
    g_return_if_fail(raster);

    if (g_object_get_data(G_OBJECT(data_view), "gwy-app-surface-density-map"))
        flags |= GWY_PREVIEW_SURFACE_DENSITY;

    GdkRectangle allocation;
    gtk_widget_get_allocation(GTK_WIDGET(data_view), &allocation);
    gwy_preview_surface_to_datafield(surface, raster, allocation.width, allocation.height, flags);
    gwy_data_view_set_zoom(data_view, 1.0);
    gwy_data_field_data_changed(raster);
}

static void
surface_density_map_toggled(GwyDataWindow *data_window,
                            GtkToggleButton *toggle)
{
    GtkWidget *data_view = gwy_data_window_get_data_view(data_window);
    gboolean active = gtk_toggle_button_get_active(toggle);

    g_object_set_data(G_OBJECT(data_view), "gwy-app-surface-density-map", GINT_TO_POINTER(active));
    update_surface_preview(data_window);
}

static GtkWidget*
gwy_app_menu_surface_popup_create(GtkAccelGroup *accel_group)
{
    static struct {
        const gchar *label;
        gpointer callback;
        gpointer cbdata;
        guint key;
        GdkModifierType mods;
    }
    const menu_items[] = {
        {
            N_("Metadata _Browser..."), metadata_browser, GUINT_TO_POINTER(GWY_APP_SURFACE),
            GDK_KEY_B, GDK_CONTROL_MASK | GDK_SHIFT_MASK
        },
        {
            N_("View _Log..."), log_browser, GUINT_TO_POINTER(GWY_APP_SURFACE),
            0, 0
        },
    };
    GwySensitivityGroup *sensgroup;
    GtkWidget *menu, *item;
    const gchar *name;
    guint i, mask;

    menu = gtk_menu_new();
    if (accel_group)
        gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    sensgroup = gwy_app_sensitivity_get_group();
    for (i = 0; i < G_N_ELEMENTS(menu_items); i++) {
        if (menu_items[i].callback == gwy_app_run_xyz_func
            && !gwy_xyz_func_get_run_types((gchar*)menu_items[i].cbdata)) {
            g_warning("Surface function <%s> for data view context menu is not available.",
                      (const gchar*)menu_items[i].cbdata);
            continue;
        }
        if (menu_items[i].callback == gwy_app_run_xyz_func) {
            name = _(gwy_xyz_func_get_menu_path(menu_items[i].cbdata));
            name = strrchr(name, '/');
            if (!name) {
                g_warning("Invalid translated menu path for <%s>", (const gchar*)menu_items[i].cbdata);
                continue;
            }
            item = gtk_menu_item_new_with_mnemonic(name + 1);
            mask = gwy_xyz_func_get_sensitivity_mask(menu_items[i].cbdata);
            gwy_sensitivity_group_add_widget(sensgroup, item, mask);
        }
        else {
            item = gtk_menu_item_new_with_mnemonic(_(menu_items[i].label));
        }

        if (menu_items[i].key)
            gtk_widget_add_accelerator(item, "activate", accel_group,
                                       menu_items[i].key, menu_items[i].mods, GTK_ACCEL_VISIBLE | GTK_ACCEL_LOCKED);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
        g_signal_connect_swapped(item, "activate", G_CALLBACK(menu_items[i].callback), menu_items[i].cbdata);
    }

    return menu;
}

static gboolean
gwy_app_surface_popup_menu_popup_mouse(GtkWidget *menu,
                                       GdkEventButton *event,
                                       GwyDataView *data_view)
{
    if (event->button != 3)
        return FALSE;

    gwy_app_data_browser_select_xyz(data_view);
    gtk_menu_popup_at_pointer(GTK_MENU(menu), (GdkEvent*)event);

    return TRUE;
}

static void
gwy_app_surface_popup_menu_popup_key(GtkWidget *menu,
                                     GtkWidget *data_window)
{
    gtk_menu_popup_at_widget(GTK_MENU(menu), data_window, GDK_GRAVITY_NORTH_WEST, GDK_GRAVITY_NORTH_WEST, NULL);
}

/********************************************************************************************************************
 *                                                                                                                  *
 *     Lawns                                                                                                        *
 *                                                                                                                  *
 ********************************************************************************************************************/

void
_gwy_app_lawn_window_setup(GwyDataWindow *data_window)
{
    static GtkWidget *popup_menu = NULL;
    static GtkWidget *corner_menu = NULL;

    GtkAccelGroup *accel_group;
    GtkWidget *data_view;
    GtkWidget *corner, *ebox, *main_window;
    GtkWidget *vbox, *hbox, *button, *label;

    if ((!popup_menu || !corner_menu) && (main_window = gwy_app_main_window_get())) {
        g_return_if_fail(GTK_IS_WINDOW(main_window));
        accel_group = GTK_ACCEL_GROUP(g_object_get_data(G_OBJECT(main_window), "accel_group"));

        if (accel_group && !popup_menu) {
            popup_menu = gwy_app_menu_lawn_popup_create(accel_group);
            gtk_widget_show_all(popup_menu);
        }
        if (accel_group && !corner_menu) {
            corner_menu = gwy_app_menu_data_corner_create(accel_group, GWY_PAGE_CURVE_MAPS);
            gtk_widget_show_all(corner_menu);
        }
    }

    gwy_app_add_main_accel_group(GTK_WINDOW(data_window));
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
    corner = gtk_arrow_new(GTK_ARROW_RIGHT, GTK_SHADOW_ETCHED_OUT);
#pragma GCC diagnostic pop
    gwy_set_widget_padding(corner, 2, 2, 2, 2);

    ebox = gtk_event_box_new();
    gtk_container_add(GTK_CONTAINER(ebox), corner);
    gtk_widget_add_events(ebox, GDK_BUTTON_PRESS_MASK);
    gtk_widget_show_all(ebox);

    gwy_data_window_set_ul_corner_widget(data_window, ebox);
    gwy_help_add_to_window(GTK_WINDOW(data_window), "curve-maps", NULL, GWY_HELP_DEFAULT);
    restore_data_window_zoom(data_window);

    data_view = gwy_data_window_get_data_view(data_window);
    g_signal_connect_swapped(data_view, "button-press-event",
                             G_CALLBACK(gwy_app_lawn_popup_menu_popup_mouse), popup_menu);
    g_signal_connect_swapped(ebox, "button-press-event", G_CALLBACK(gwy_app_data_corner_menu_popup_mouse), corner_menu);
    g_signal_connect_swapped(data_window, "popup-menu", G_CALLBACK(gwy_app_lawn_popup_menu_popup_key), popup_menu);

    hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);

    vbox = gtk_bin_get_child(GTK_BIN(data_window));
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);
    gtk_box_reorder_child(GTK_BOX(vbox), hbox, 0);

    button = gtk_button_new_with_mnemonic(_("_Change Preview"));
    gtk_widget_set_can_focus(button, FALSE);
    gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    g_signal_connect_swapped(button, "clicked", G_CALLBACK(change_lawn_preview), data_window);

    label = gtk_label_new(NULL);
    gtk_label_set_xalign(GTK_LABEL(label), 0.0);
    gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);
    gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);

    g_object_set_data(G_OBJECT(data_window), "gwy-lawn-info", label);

    g_signal_connect(data_window, "configure-event", G_CALLBACK(gwy_app_lawn_window_configured), NULL);
    g_signal_connect(data_window, "size-allocate", G_CALLBACK(gwy_app_lawn_window_configured), NULL);
}

static gboolean
gwy_app_lawn_window_configured(GwyDataWindow *window)
{
    GwyDataView *view = GWY_DATA_VIEW(gwy_data_window_get_data_view(window));
    GwyContainer *container = gwy_data_view_get_data(view);
    const gchar *prefix = gwy_data_view_get_data_prefix(view);
    gchar *key;

    g_return_val_if_fail(container, FALSE);
    g_return_val_if_fail(prefix, FALSE);

    /* This leads to some odd keys containing .../preview/view/... */
    key = g_strconcat(prefix, "/view/scale", NULL);
    gwy_container_set_double_by_name(container, key, gwy_data_view_get_real_zoom(view));
    g_free(key);

    save_widget_screen_relative_size(GTK_WIDGET(view), container, prefix, FALSE);

    return FALSE;
}

static gboolean
lawn_preview_filter(GwyContainer *data,
                    gint id,
                    gpointer user_data)
{
    GwyDataField *field;
    GQuark key = gwy_app_get_data_key_for_id(id);
    GwySIUnit *funit, *bunit;
    GwyLawn *lawn = (GwyLawn*)user_data;

    field = GWY_DATA_FIELD(gwy_container_get_object(data, key));
    g_return_val_if_fail(field, FALSE);

    if (field->xres != lawn->xres || field->yres != lawn->yres
        || ABS(log(field->xreal/lawn->xreal)) > 1e-6 || ABS(log(field->yreal/lawn->yreal)) > 1e-6)
        return FALSE;

    bunit = gwy_lawn_get_si_unit_xy(lawn);
    funit = gwy_data_field_get_si_unit_xy(field);
    if (!gwy_si_unit_equal(bunit, funit))
        return FALSE;

    return TRUE;
}

static void
update_lawn_preview_sens(LawnPreviewType type,
                         GObject *dialog)
{
    GwyDataChooser *chooser;
    GtkWidget *widget;
    gboolean ok = TRUE;

    widget = GTK_WIDGET(g_object_get_data(dialog, "channel-chooser"));
    chooser = GWY_DATA_CHOOSER(widget);
    gtk_widget_set_sensitive(widget, type == LAWN_PREVIEW_IMAGE);

    if (type == LAWN_PREVIEW_IMAGE) {
        gint id;
        GwyContainer *container = gwy_data_chooser_get_active(chooser, &id);
        ok = !!container;
    }
    gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog), GTK_RESPONSE_OK, ok);
}

static void
lawn_preview_type_changed(GtkRadioButton *button,
                          GObject *dialog)
{
    GSList *group = gtk_radio_button_get_group(button);
    LawnPreviewType type = gwy_radio_buttons_get_current(group);
    update_lawn_preview_sens(type, dialog);
}

static gdouble
lawn_reduce_length(G_GNUC_UNUSED gint ncurves, gint curvelength, G_GNUC_UNUSED const gdouble *curvedata,
                   G_GNUC_UNUSED gpointer user_data)
{
    return curvelength;
}

static gdouble
lawn_reduce_avg(gint ncurves, gint curvelength, const gdouble *curvedata, gpointer user_data)
{
    guint i, idx = GPOINTER_TO_UINT(user_data);
    gdouble s = 0.0;

    g_return_val_if_fail(idx < ncurves, 0.0);
    if (!curvelength)
        return 0.0;

    curvedata += idx*curvelength;
    for (i = 0; i < curvelength; i++)
        s += curvedata[i];
    return s/curvelength;
}

static gdouble
lawn_reduce_rms(gint ncurves, gint curvelength, const gdouble *curvedata, gpointer user_data)
{
    guint i, idx = GPOINTER_TO_UINT(user_data);
    gdouble m, s = 0.0;

    g_return_val_if_fail(idx < ncurves, 0.0);
    if (!curvelength)
        return 0.0;

    m = lawn_reduce_avg(ncurves, curvelength, curvedata, user_data);
    curvedata += idx*curvelength;
    for (i = 0; i < curvelength; i++)
        s += (curvedata[i] - m)*(curvedata[i] - m);
    return sqrt(s/curvelength);
}

static gdouble
lawn_reduce_min(gint ncurves, gint curvelength, const gdouble *curvedata, gpointer user_data)
{
    guint i, idx = GPOINTER_TO_UINT(user_data);
    gdouble m = G_MAXDOUBLE;

    g_return_val_if_fail(idx < ncurves, 0.0);
    if (!curvelength)
        return 0.0;

    curvedata += idx*curvelength;
    for (i = 0; i < curvelength; i++)
        m = fmin(m, curvedata[i]);
    return m;
}

static gdouble
lawn_reduce_max(gint ncurves, gint curvelength, const gdouble *curvedata, gpointer user_data)
{
    guint i, idx = GPOINTER_TO_UINT(user_data);
    gdouble m = -G_MAXDOUBLE;

    g_return_val_if_fail(idx < ncurves, 0.0);
    if (!curvelength)
        return 0.0;

    curvedata += idx*curvelength;
    for (i = 0; i < curvelength; i++)
        m = fmax(m, curvedata[i]);
    return m;
}

static void
change_lawn_preview(GwyDataWindow *data_window)
{
    GtkWidget *dialog, *label, *grid, *chooser, *curvecombo;
    GwyCurveReduceFunction reduce_func = NULL;
    GwyContainer *data, *cdata;
    GwyDataField *preview, *mask;
    GwyLawn *lawn;
    GSList *group, *l;
    gint response, id, cid, i;
    guint curveno;
    LawnPreviewType type;
    GQuark quark;

    gwy_app_data_browser_get_current(GWY_APP_LAWN, &lawn,
                                     GWY_APP_LAWN_ID, &id,
                                     GWY_APP_CONTAINER, &data,
                                     0);
    g_return_if_fail(GWY_IS_LAWN(lawn));
    g_return_if_fail(id >= 0);

    type = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(lawn), "gwy-preview-type"));
    curveno = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(lawn), "gwy-preview-curveno"));

    dialog = gtk_dialog_new_with_buttons(_("Change Curve Map Preview"), GTK_WINDOW(data_window),
                                         GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                                         NULL, NULL);
    gwy_add_cancel_button_to_dialog(GTK_DIALOG(dialog));
    gwy_add_ok_button_to_dialog(GTK_DIALOG(dialog));
    grid = gtk_grid_new();
    gtk_container_set_border_width(GTK_CONTAINER(grid), 8);
    gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(GTK_DIALOG(dialog))), grid, TRUE, TRUE, 0);

    label = gtk_label_new(_("Curve:"));
    gtk_label_set_xalign(GTK_LABEL(label), 0.0);
    gtk_grid_attach(GTK_GRID(grid), label, 0, 0, 1, 1);
    curvecombo = gwy_combo_box_lawn_curve_new(NULL, NULL, lawn, curveno);
    gtk_grid_attach(GTK_GRID(grid), curvecombo, 1, 0, 1, 1);

    label = gtk_label_new(_("Preview quantity:"));
    gtk_label_set_xalign(GTK_LABEL(label), 0.0);
    gtk_widget_set_margin_top(label, 8);
    gtk_grid_attach(GTK_GRID(grid), label, 0, 2, 1, 1);

    group = gwy_radio_buttons_createl(G_CALLBACK(lawn_preview_type_changed), dialog, type,
                                      _("Mean"), LAWN_PREVIEW_MEAN,
                                      _("Minimum"), LAWN_PREVIEW_MINIMUM,
                                      _("Maximum"), LAWN_PREVIEW_MAXIMUM,
                                      _("RMS"), LAWN_PREVIEW_RMS,
                                      _("Image:"), LAWN_PREVIEW_IMAGE,
                                      NULL);
    l = group;
    i = 2;
    while (l) {
        gtk_grid_attach(GTK_GRID(grid), GTK_WIDGET(l->data), 0, i, 1, 1);
        i++;
        l = g_slist_next(l);
    }
    chooser = gwy_data_chooser_new_channels();
    gwy_data_chooser_set_filter(GWY_DATA_CHOOSER(chooser), &lawn_preview_filter, lawn, NULL);
    gtk_grid_attach(GTK_GRID(grid), chooser, 1, i-1, 1, 1);
    g_object_set_data(G_OBJECT(dialog), "channel-chooser", chooser);

    update_lawn_preview_sens(gwy_radio_buttons_get_current(group), G_OBJECT(dialog));

    gtk_widget_show_all(dialog);
    response = gtk_dialog_run(GTK_DIALOG(dialog));
    type = gwy_radio_buttons_get_current(group);
    cdata = gwy_data_chooser_get_active(GWY_DATA_CHOOSER(chooser), &cid);
    curveno = gwy_enum_combo_box_get_active(GTK_COMBO_BOX(curvecombo));
    curveno = MIN(curveno, gwy_lawn_get_n_curves(lawn)-1);
    gtk_widget_destroy(dialog);
    if (response != GTK_RESPONSE_OK)
        return;

    quark = gwy_app_get_lawn_preview_key_for_id(id);
    if (!gwy_container_gis_object(data, quark, (GObject**)&preview)) {
        g_warning("No preview field found for lawn %d.", id);
        preview = gwy_data_field_new(gwy_lawn_get_xres(lawn), gwy_lawn_get_yres(lawn), 1.0, 1.0, FALSE);
        gwy_container_pass_object(data, quark, preview);
    }

    if (type == LAWN_PREVIEW_MEAN)
        reduce_func = lawn_reduce_avg;
    else if (type == LAWN_PREVIEW_MINIMUM)
        reduce_func = lawn_reduce_min;
    else if (type == LAWN_PREVIEW_MAXIMUM)
        reduce_func = lawn_reduce_max;
    else if (type == LAWN_PREVIEW_RMS)
        reduce_func = lawn_reduce_rms;
    else if (type == LAWN_PREVIEW_IMAGE) {
        GwyDataField *field = gwy_container_get_object(cdata, gwy_app_get_data_key_for_id(cid));
        gwy_data_field_assign(preview, field);
    }
    else {
        g_return_if_reached();
    }

    if (reduce_func) {
        gwy_lawn_reduce_to_plane(lawn, preview, reduce_func, GUINT_TO_POINTER(curveno));
        gwy_si_unit_assign(gwy_data_field_get_si_unit_z(preview), gwy_lawn_get_si_unit_curve(lawn, curveno));

        mask = gwy_data_field_new_alike(preview, FALSE);
        gwy_lawn_reduce_to_plane(lawn, mask, lawn_reduce_length, GUINT_TO_POINTER(curveno));
        gwy_data_field_laplace_solve(preview, mask, 0, 0.5);
        g_object_unref(mask);
    }

    g_object_set_data(G_OBJECT(lawn), "gwy-preview-type", GUINT_TO_POINTER(type));
    g_object_set_data(G_OBJECT(lawn), "gwy-preview-curveno", GUINT_TO_POINTER(curveno));
    gwy_data_field_data_changed(preview);
}

static GtkWidget*
gwy_app_menu_lawn_popup_create(GtkAccelGroup *accel_group)
{
    static struct {
        const gchar *label;
        gpointer callback;
        gpointer cbdata;
        guint key;
        GdkModifierType mods;
    }
    const menu_items[] = {
        {
            N_("Zoom _1:1"), gwy_app_any_window_reset_zoom, GUINT_TO_POINTER(GWY_APP_CURVE_MAP_VIEW),
            0, 0
        },
        {
            N_("Metadata _Browser..."), metadata_browser, GUINT_TO_POINTER(GWY_APP_LAWN),
            GDK_KEY_B, GDK_CONTROL_MASK | GDK_SHIFT_MASK
        },
        {
            N_("View _Log..."), log_browser, GUINT_TO_POINTER(GWY_APP_LAWN),
            0, 0
        },
    };
    GwySensitivityGroup *sensgroup;
    GtkWidget *menu, *item;
    const gchar *name;
    guint i, mask;

    menu = gtk_menu_new();
    if (accel_group)
        gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    sensgroup = gwy_app_sensitivity_get_group();
    for (i = 0; i < G_N_ELEMENTS(menu_items); i++) {
        if (menu_items[i].callback == gwy_app_run_curve_map_func
            && !gwy_curve_map_func_get_run_types((gchar*)menu_items[i].cbdata)) {
            g_warning("Lawn function <%s> for data view context menu is not available.",
                      (const gchar*)menu_items[i].cbdata);
            continue;
        }
        if (menu_items[i].callback == gwy_app_run_curve_map_func) {
            name = _(gwy_curve_map_func_get_menu_path(menu_items[i].cbdata));
            name = strrchr(name, '/');
            if (!name) {
                g_warning("Invalid translated menu path for <%s>", (const gchar*)menu_items[i].cbdata);
                continue;
            }
            item = gtk_menu_item_new_with_mnemonic(name + 1);
            mask = gwy_curve_map_func_get_sensitivity_mask(menu_items[i].cbdata);
            gwy_sensitivity_group_add_widget(sensgroup, item, mask);
        }
        else {
            item = gtk_menu_item_new_with_mnemonic(_(menu_items[i].label));
        }

        if (menu_items[i].key)
            gtk_widget_add_accelerator(item, "activate", accel_group, menu_items[i].key, menu_items[i].mods,
                                       GTK_ACCEL_VISIBLE | GTK_ACCEL_LOCKED);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
        g_signal_connect_swapped(item, "activate", G_CALLBACK(menu_items[i].callback), menu_items[i].cbdata);
    }

    return menu;
}

static gboolean
gwy_app_lawn_popup_menu_popup_mouse(GtkWidget *menu,
                                    GdkEventButton *event,
                                    GwyDataView *data_view)
{
    if (event->button != 3)
        return FALSE;

    gwy_app_data_browser_select_curve_map(data_view);
    gtk_menu_popup_at_pointer(GTK_MENU(menu), (GdkEvent*)event);

    return TRUE;
}

static void
gwy_app_lawn_popup_menu_popup_key(GtkWidget *menu,
                                  GtkWidget *data_window)
{
    gtk_menu_popup_at_widget(GTK_MENU(menu), data_window, GDK_GRAVITY_NORTH_WEST, GDK_GRAVITY_NORTH_WEST, NULL);
}

/********************************************************************************************************************
 *                                                                                                                  *
 *     Miscellaneous                                                                                                *
 *                                                                                                                  *
 ********************************************************************************************************************/

/**
 * gwy_app_switch_tool:
 * @toolname: Tool name, that is type name of the tool object in the GLib
 *            type system.  This can be for instance "GwyToolGrainMeasure".
 *
 * Switches the current tool to given tool.
 *
 * If the tool is already active it is shown when hidden and hidden when
 * visible.
 **/
void
gwy_app_switch_tool(const gchar *toolname)
{
    GwyContainer *settings;
    GwyTool *newtool;
    GwyDataView *data_view;
    GType type;
    gboolean do_restore = FALSE;

    gwy_debug("%s", toolname ? toolname : "NONE");
    type = g_type_from_name(toolname);
    g_return_if_fail(type);

    gwy_app_data_browser_get_current(GWY_APP_DATA_VIEW, &data_view, 0);
    if (current_tool && type == G_TYPE_FROM_INSTANCE(current_tool)) {
        if (!gwy_tool_is_visible(current_tool))
            gwy_tool_show(current_tool);
        else
            gwy_tool_hide(current_tool);
        return;
    }

    g_clear_object(&current_tool);
    newtool = (GwyTool*)g_object_new(type, NULL);
    current_tool = newtool;
    g_return_if_fail(GWY_IS_TOOL(newtool));

    settings = gwy_app_settings_get();
    gwy_container_gis_boolean_by_name(settings, "/app/restore-tool-position", &do_restore);
    if (do_restore)
        gwy_tool_restore_screen_position(newtool);

    if (data_view) {
        GwySpectra *spectra;

        gwy_tool_data_switched(current_tool, data_view);
        gwy_app_data_browser_get_current(GWY_APP_SPECTRA, &spectra, 0);
        gwy_tool_spectra_switched(current_tool, spectra);
        gwy_tool_show(current_tool);
    }
}

/**
 * gwy_app_current_tool_name:
 *
 * Obtains the name of currently active tool.
 *
 * See gwy_app_switch_tool() for the name description.
 *
 * In some rare circumstances, this function can return %NULL because no tool is active.  This includes program
 * startup and shutdown and during the construction of a new #GwyTool object while switching tools.  Also, %NULL is
 * typically returned outside Gwyddion in programs just using the libraries.
 *
 * Returns: The tool name.
 **/
const gchar*
gwy_app_current_tool_name(void)
{
    if (!current_tool)
        return NULL;

    return G_OBJECT_TYPE_NAME(current_tool);
}

/**
 * gwy_app_current_tool:
 *
 * Obtains the currently active tool object.
 *
 * In some rare circumstances, this function can return %NULL because no tool is active.  This includes program
 * startup and shutdown and during the construction of a new #GwyTool object while switching tools.  Also, %NULL is
 * typically returned outside Gwyddion in programs just using the libraries.
 *
 * Returns: The current tool object.
 **/
GwyTool*
gwy_app_current_tool(void)
{
    return current_tool;
}

static void
gwy_app_any_window_reset_zoom(gpointer user_data)
{
    GtkWidget *window, *view;

    gwy_app_data_browser_get_current(GPOINTER_TO_UINT(user_data), &view, 0);
    window = gtk_widget_get_ancestor(view, GWY_TYPE_DATA_WINDOW);
    g_return_if_fail(window);
    gwy_data_window_set_zoom(GWY_DATA_WINDOW(window), 10000);
}

static void
get_current_view_data_and_id(GwyAppWhat what, GtkWidget **view, GwyContainer **data, gint *id)
{
    if (what == GWY_APP_DATA_FIELD) {
        gwy_app_data_browser_get_current(GWY_APP_DATA_VIEW, view,
                                         GWY_APP_CONTAINER, data,
                                         GWY_APP_DATA_FIELD_ID, id,
                                         0);
    }
    else if (what == GWY_APP_BRICK) {
        gwy_app_data_browser_get_current(GWY_APP_VOLUME_VIEW, view,
                                         GWY_APP_CONTAINER, data,
                                         GWY_APP_BRICK_ID, id,
                                         0);
    }
    else if (what == GWY_APP_SURFACE) {
        gwy_app_data_browser_get_current(GWY_APP_XYZ_VIEW, view,
                                         GWY_APP_CONTAINER, data,
                                         GWY_APP_SURFACE_ID, id,
                                         0);
    }
    else if (what == GWY_APP_LAWN) {
        gwy_app_data_browser_get_current(GWY_APP_CURVE_MAP_VIEW, view,
                                         GWY_APP_CONTAINER, data,
                                         GWY_APP_LAWN_ID, id,
                                         0);
    }
    else {
        g_return_if_reached();
    }
}

static void
metadata_browser(gpointer pwhat)
{
    GwyAppWhat what = GPOINTER_TO_UINT(pwhat);
    GtkWidget *view;
    GwyContainer *container;
    gint id;

    get_current_view_data_and_id(what, &view, &container, &id);
    if (!view || !container || id == -1)
        return;

    if (what == GWY_APP_DATA_FIELD)
        gwy_app_metadata_browser_for_channel(container, id);
    else if (what == GWY_APP_BRICK)
        gwy_app_metadata_browser_for_volume(container, id);
    else if (what == GWY_APP_SURFACE)
        gwy_app_metadata_browser_for_xyz(container, id);
    else if (what == GWY_APP_LAWN)
        gwy_app_metadata_browser_for_curve_map(container, id);
}

static void
log_browser(gpointer pwhat)
{
    GwyAppWhat what = GPOINTER_TO_UINT(pwhat);
    GtkWidget *view;
    GwyContainer *container;
    gint id;

    get_current_view_data_and_id(what, &view, &container, &id);
    if (!view || !container || id == -1)
        return;

    if (what == GWY_APP_DATA_FIELD)
        gwy_app_log_browser_for_channel(container, id);
    else if (what == GWY_APP_BRICK)
        gwy_app_log_browser_for_volume(container, id);
    else if (what == GWY_APP_SURFACE)
        gwy_app_log_browser_for_xyz(container, id);
    else if (what == GWY_APP_LAWN)
        gwy_app_log_browser_for_curve_map(container, id);
}

static void
gwy_app_change_mask_color(void)
{
    GwyDataView *data_view;

    gwy_app_data_browser_get_current(GWY_APP_DATA_VIEW, &data_view, 0);
    g_return_if_fail(data_view);
    gwy_app_data_view_change_mask_color(data_view);
}

/**
 * gwy_app_data_view_change_mask_color:
 * @data_view: A data view (of application's data window).  It must have a mask.
 *
 * Runs mask color selector on a data view.
 *
 * This is a convenience function to run gwy_color_selector_for_mask(), possibly taking the initial color from
 * settings.
 **/
void
gwy_app_data_view_change_mask_color(GwyDataView *data_view)
{
    GwyDataViewLayer *layer;
    GwyContainer *data, *settings;
    const gchar *key;
    GwyRGBA rgba;

    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));
    data = gwy_data_view_get_data(data_view);
    g_return_if_fail(GWY_IS_CONTAINER(data));
    layer = gwy_data_view_get_alpha_layer(data_view);
    g_return_if_fail(GWY_IS_LAYER_MASK(layer));
    key = gwy_layer_mask_get_color_key(GWY_LAYER_MASK(layer));
    g_return_if_fail(key);
    gwy_debug("<%s>", key);

    /* copy defaults to data container if necessary */
    if (!gwy_rgba_get_from_container(&rgba, data, key)) {
        settings = gwy_app_settings_get();
        gwy_rgba_get_from_container(&rgba, settings, "/mask");
        gwy_rgba_store_to_container(&rgba, data, key);
    }
    gwy_color_selector_for_mask(NULL, NULL, data, key);
}

/**
 * gwy_app_save_window_position:
 * @window: A window to save position of.
 * @prefix: Unique prefix in settings to store the information under.
 * @position: %TRUE to save position information.
 * @size: %TRUE to save size information.
 *
 * Saves position and/or size of a window to settings.
 *
 * Some sanity checks are included, therefore if window position and/or size is too suspicious, it is not saved.
 **/
void
gwy_app_save_window_position(GtkWindow *window,
                             const gchar *prefix,
                             gboolean position,
                             gboolean size)
{
    GwyContainer *settings;
    gint x, y, w, h, scw, sch;
    guint len;
    gchar *key;
    gint64 mconf;

    g_return_if_fail(GTK_IS_WINDOW(window));
    g_return_if_fail(prefix);
    if (!(position || size))
        return;

    len = strlen(prefix);
    key = g_newa(gchar, len + sizeof("/position/height"));  /* The longest suffix */
    strcpy(key, prefix);

    settings = gwy_app_settings_get();
    scw = gwy_get_screen_width(GTK_WIDGET(window));
    sch = gwy_get_screen_height(GTK_WIDGET(window));
    /* FIXME: read the gtk_window_get_position() docs about how this is a broken approach. But it is better than
     * nothing. We should detect broken environments like Wayland which do not support programmatic positioning of
     * windows at all and avoid saving the poisition. On the other hand, gtk_window_move() should also be noop then
     * so we may not care. */
    if (position) {
        mconf = monitor_config_checksum(GTK_WIDGET(window));
        gtk_window_get_position(window, &x, &y);
        if (x >= 0 && y >= 0 && x+1 < scw && y+1 < sch) {
            strcpy(key + len, "/position/x");
            gwy_container_set_int32_by_name(settings, key, x);
            strcpy(key + len, "/position/y");
            gwy_container_set_int32_by_name(settings, key, y);
            strcpy(key + len, "/position/mconf");
            gwy_container_set_int64_by_name(settings, key, mconf);
        }
    }
    if (size) {
        gtk_window_get_size(window, &w, &h);
        if (w > 1 && h > 1) {
            strcpy(key + len, "/position/width");
            gwy_container_set_int32_by_name(settings, key, w);
            strcpy(key + len, "/position/height");
            gwy_container_set_int32_by_name(settings, key, h);
        }
    }
}

typedef struct {
    gulong hid;
    gint x;
    gint y;
} WindowPos;

static void
restore_pos(GtkWindow *window, WindowPos *wp)
{
    g_printerr("moving window %p(%s) to (%d, %d)\n", window, gtk_window_get_title(window), wp->x, wp->y);
    GdkWindow *gdkwin = gtk_widget_get_window(GTK_WIDGET(window));
    gdk_window_move(gdkwin, wp->x, wp->y);
    g_signal_handler_disconnect(window, wp->hid);
    g_free(wp);
}

/**
 * gwy_app_restore_window_position:
 * @window: A window to restore position of.
 * @prefix: Unique prefix in settings to get the information from (the same as
 *          in gwy_app_save_window_position()).
 * @grow_only: %TRUE to only attempt set the window default size bigger than it
 *              requests, never smaller.
 *
 * Restores a window position and/or size from settings.
 *
 * Unlike gwy_app_save_window_position(), this function has no @position and @size arguments, it simply restores all
 * attributes that were saved.
 *
 * Note to restore position (not size) it should be called twice for each window to accommodate sloppy window
 * managers: once before the window is shown, second time immediately after showing the window.
 *
 * Some sanity checks are included, therefore if saved window position and/or size is too suspicious, it is not
 * restored.
 **/
void
gwy_app_restore_window_position(GtkWindow *window,
                                const gchar *prefix,
                                gboolean grow_only)
{
    GwyContainer *settings;
    GtkRequisition req;
    gint x, y, w, h, scw, sch;
    guint len;
    gchar *key;
    gint64 mconf, savedmconf = 0;

    g_return_if_fail(GTK_IS_WINDOW(window));
    g_return_if_fail(prefix);

    len = strlen(prefix);
    key = g_newa(gchar, len + sizeof("/position/height"));  /* The longest suffix */
    strcpy(key, prefix);

    settings = gwy_app_settings_get();
    mconf = monitor_config_checksum(GTK_WIDGET(window));
    scw = gwy_get_screen_width(GTK_WIDGET(window));
    sch = gwy_get_screen_height(GTK_WIDGET(window));
    x = y = w = h = -1;
    strcpy(key + len, "/position/x");
    gwy_container_gis_int32_by_name(settings, key, &x);
    strcpy(key + len, "/position/y");
    gwy_container_gis_int32_by_name(settings, key, &y);
    strcpy(key + len, "/position/mconf");
    gwy_container_gis_int64_by_name(settings, key, &savedmconf);
    strcpy(key + len, "/position/width");
    gwy_container_gis_int32_by_name(settings, key, &w);
    strcpy(key + len, "/position/height");
    gwy_container_gis_int32_by_name(settings, key, &h);
    g_printerr("mcof = %lx, savedmconf = %lx\n", mconf, savedmconf);
    if (mconf == savedmconf && x >= 0 && y >= 0 && x+1 < scw && y+1 < sch) {
        gtk_window_move(window, x, y);
        WindowPos *wp = g_new(WindowPos, 1);
        wp->x = x;
        wp->y = y;
        wp->hid = g_signal_connect_after(window, "map", G_CALLBACK(restore_pos), wp);
        /* FIXME: We lost wp if window is never mapped. */
    }
    if (w > 1 && h > 1) {
        if (grow_only) {
            gtk_widget_get_preferred_size(GTK_WIDGET(window), NULL, &req);
            w = MAX(w, req.width);
            h = MAX(h, req.height);
        }
        gtk_window_set_default_size(window, w, h);
    }
}

static void
save_widget_screen_relative_size(GtkWidget *widget,
                                 GwyContainer *container, const gchar *prefix,
                                 gboolean absolute_too)
{

    if (!widget)
        return;

    gdouble scw = gwy_get_screen_width(widget);
    gdouble sch = gwy_get_screen_height(widget);
    GdkRectangle allocation;
    gtk_widget_get_allocation(widget, &allocation);
    gint w = allocation.width;
    gint h = allocation.height;
    gdouble relsize = MAX(w/scw, h/sch);
    gchar *key = g_strconcat(prefix, "/view/relative-size", NULL);
    gwy_container_set_double_by_name(container, key, relsize);
    g_free(key);

    if (absolute_too) {
        key = g_strconcat(prefix, "/view/width", NULL);
        gwy_container_set_int32_by_name(container, key, w);
        g_free(key);

        key = g_strconcat(prefix, "/view/height", NULL);
        gwy_container_set_int32_by_name(container, key, h);
        g_free(key);
    }
}

static void
restore_data_window_zoom(GwyDataWindow *data_window)
{
    GwyDataView *data_view = GWY_DATA_VIEW(gwy_data_window_get_data_view(data_window));
    GwyContainer *container = gwy_data_view_get_data(data_view);
    const gchar *prefix = gwy_data_view_get_data_prefix(data_view);
    GtkRequisition req;
    gdouble scw, sch, scale, relsize, newrelsize;
    gchar *key;

    if (!container || !prefix)
        return;

    key = g_strconcat(prefix, "/view/relative-size", NULL);
    if (!gwy_container_gis_double_by_name(container, key, &relsize)) {
        g_free(key);
        return;
    }
    g_free(key);

    key = g_strconcat(prefix, "/view/scale", NULL);
    if (!gwy_container_gis_double_by_name(container, key, &scale)) {
        g_free(key);
        return;
    }
    g_free(key);

    if (scale <= 0.0 || relsize <= 0.0)
        return;

    gtk_widget_get_preferred_size(GTK_WIDGET(data_view), NULL, &req);
    scw = gwy_get_screen_width(GTK_WIDGET(data_window));
    sch = gwy_get_screen_height(GTK_WIDGET(data_window));
    newrelsize = MAX(scale*req.width/scw, scale*req.height/sch);
    gwy_debug("restoring data window: relsize %g, zoom %g, request %dx%d, newrelsize %g",
              relsize, scale, req.width, req.height, newrelsize);

    /* If the data view will be small we can just apply the saved zoom.  Should
     * it be larger though, we must check if it is not too large and better
     * show it at defaut size than huge. */
    if (newrelsize > 1.2*relsize || newrelsize > 0.85)
        return;
    gwy_data_view_set_zoom(data_view, scale);
}

static void
restore_other_window_size(GtkWindow *window,
                          GwyContainer *container, const gchar *prefix)
{
    gdouble scw, sch, relsize, newrelsize;
    gint w, h;
    gchar *key;

    if (!container || !prefix)
        return;

    key = g_strconcat(prefix, "/view/relative-size", NULL);
    if (!gwy_container_gis_double_by_name(container, key, &relsize)) {
        g_free(key);
        return;
    }
    g_free(key);

    key = g_strconcat(prefix, "/view/width", NULL);
    if (!gwy_container_gis_int32_by_name(container, key, &w)) {
        g_free(key);
        return;
    }
    g_free(key);

    key = g_strconcat(prefix, "/view/height", NULL);
    if (!gwy_container_gis_int32_by_name(container, key, &h)) {
        g_free(key);
        return;
    }
    g_free(key);

    if (w <= 0 || h <= 0 || relsize <= 0.0)
        return;

    scw = gwy_get_screen_width(GTK_WIDGET(window));
    sch = gwy_get_screen_height(GTK_WIDGET(window));
    newrelsize = MAX(w/scw, h/sch);
    gwy_debug("restoring other window: relsize %g, size %dx%d, newrelsize %g",
              relsize, w, h, newrelsize);

    /* If the window will be small we can just apply the saved zoom.  Should it be larger though, we must check if it
     * is not too large and better show it at defaut size than huge. */
    if (newrelsize > 1.2*relsize || newrelsize > 0.9)
        return;
    gtk_window_set_default_size(window, w, h);
}

/* We probably do not have to this repeatedly but GDK has the monitor properties tabulated so it is just a few extra
 * function calls... */
static gint64
monitor_config_checksum(GtkWidget *widget)
{
    static GChecksum *checksummer = NULL;

    if (!widget)
        return 0;

    GdkDisplay *display = gtk_widget_get_display(widget);
    if (!display)
        display = gdk_display_get_default();
    if (!display)
        return 0;

    gint n = gdk_display_get_n_monitors(display);
    if (!n)
        return 0;

    GdkRectangle *mdims, mdim0;
    if (n == 1)
        mdims = &mdim0;
    else
        mdims = g_new(GdkRectangle, n);

    for (gint i = 0; i < n; i++) {
        GdkMonitor *monitor = gdk_display_get_monitor(display, i);
        gdk_monitor_get_geometry(monitor, mdims + i);
    }

    if (!checksummer)
        checksummer = g_checksum_new(G_CHECKSUM_MD5);
    else
        g_checksum_reset(checksummer);

    g_checksum_update(checksummer, (guchar*)mdims, n*sizeof(GdkRectangle));

    guchar digest[16];
    gsize len = G_N_ELEMENTS(digest);
    gwy_clear(digest, len);
    g_checksum_get_digest(checksummer, digest, &len);
    if (n != 1)
        g_free(mdims);

    guint64 r = digest[0] & 0x7f;
    for (gint i = 1; i < 8; i++) {
        r <<= 8ul;
        r |= digest[i];
    }

    return r;
}

gint
_gwy_app_get_n_recent_files(void)
{
    return 10;
}

/**
 * gwy_app_init_widget_styles:
 * @screen: Gdk screen to set the styles for.
 *
 * Sets up style properties for special Gwyddion widgets.
 *
 * Normally not needed to call explicitly.
 **/
void
gwy_app_init_widget_styles(GdkScreen *screen)
{
    /* TODO GTK3: use CSS theming as needed. It does not seem we use toolboxheader and cornerbutton anywhere.
     * But toolboxmenubar and toolboxbutton are used. */
#if 0
    static const gchar gwyrcstyle[] =
        /* data window corner buttons */
        "style \"cornerbutton\" {\n"
        "GtkButton::focus_line_width = 0\n"
        "GtkButton::focus_padding = 0\n"
        "}\n"
        "widget \"*.cornerbutton\" style \"cornerbutton\"\n"
        "\n"
        /* toolbox group header buttons */
        "style \"toolboxheader\" {\n"
        "GtkButton::focus_line_width = 0\n"
        "GtkButton::focus_padding = 0\n"
        "}\n"
        "widget \"*.toolboxheader\" style \"toolboxheader\"\n"
        "\n"
        /* toolbox single-item menubars */
        "style \"toolboxmenubar\" {\n"
        "GtkMenuBar::shadow_type = 0\n"
        "}\n"
        "widget \"*.toolboxmenubar\" style \"toolboxmenubar\"\n"
        "\n"
        /* toolbox buttons (XXX: perhaps we would like to inherit the style for
         * toolbar buttons, but I am unable to extract the right style using
         * gtk_rc_get_style_by_paths() and gtk_style_get()...) */
        "style \"toolboxbutton\" {\n"
        "GtkButton::inner_border = {1, 1, 2, 2}\n"
        "}\n"
        "widget \"*.toolboxbutton\" style \"toolboxbutton\"\n"
        "\n";

    gtk_rc_parse_string(gwyrcstyle);
#endif
    static const gchar gwycssstyle[] =
        "progressbar trough {\n"
        "    min-height: 3.2ex;\n"
        "}\n"
        "progressbar trough progress {\n"
        "    min-height: 3ex;\n"
        "}\n"
        "\n";
    static GtkCssProvider *provider = NULL;

    if (!provider) {
        provider = gtk_css_provider_new();
        gtk_css_provider_load_from_data(provider, gwycssstyle, sizeof(gwycssstyle)-1, NULL);
    }
    gtk_style_context_add_provider_for_screen(screen, GTK_STYLE_PROVIDER(provider),
                                              GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
}

/**
 * gwy_app_init_i18n:
 *
 * Initializes internationalization.
 *
 * Normally not needed to call explicitly.
 **/
void
gwy_app_init_i18n(void)
{
#ifdef ENABLE_NLS
    gchar *locdir;

    locdir = gwy_find_self_dir("locale");
    bindtextdomain(PACKAGE, locdir);
    g_free(locdir);
    textdomain(PACKAGE);
    if (!bind_textdomain_codeset(PACKAGE, "UTF-8"))
        g_critical("Cannot bind gettext `%s' codeset to UTF-8", PACKAGE);
#endif
}

/**
 * gwy_app_init_common:
 * @error: Error location for settings loading error.
 * @...: List of module types to load, terminated with %NULL.  Possible types are "cmap", "file", "graph", "layer",
 *       "process", "tool", "volume" and "xyz", plus two special values "" and "all" for untyped modules (like pygwy
 *       or plug-in proxy) and all modules, respectively.
 *
 * Performs common application initialization.
 *
 * This function can be useful in Gwyddion-based programs and includes: widget class initialisation,
 * internationalisation initialisation, stock item and resource class loading, settings loading, and module
 * registration.
 *
 * Returns: Settings loading status.
 **/
gboolean
gwy_app_init_common(GError **error,
                    ...)
{
    gchar *settings_file;
    va_list ap;
    const gchar *dir;
    gboolean ok = TRUE;

    gwy_widgets_type_init();
    //gwy_app_init_widget_styles();
    gwy_app_init_i18n();

    /* Register resources */
    gwy_resource_class_load(g_type_class_peek(GWY_TYPE_GRADIENT));
    gwy_resource_class_load(g_type_class_peek(GWY_TYPE_GL_MATERIAL));
    gwy_resource_class_load(g_type_class_peek(GWY_TYPE_GRAIN_VALUE));

    /* Load settings */
    settings_file = gwy_app_settings_get_settings_filename();
    if (g_file_test(settings_file, G_FILE_TEST_IS_REGULAR))
        ok = gwy_app_settings_load(settings_file, error);
    g_free(settings_file);
    gwy_app_settings_get();

    /* Register modules */
    va_start(ap, error);
    dir = va_arg(ap, const gchar*);
    va_end(ap);
    if (dir && gwy_strequal(dir, "all")) {
        gchar **module_dirs;

        module_dirs = gwy_app_settings_get_module_dirs();
        gwy_module_register_modules((const gchar**)module_dirs);
        g_strfreev(module_dirs);
    }
    else {
        GPtrArray *module_dirs;
        gchar *p;
        guint i;

        module_dirs = g_ptr_array_new();

        p = gwy_find_self_dir("modules");
        va_start(ap, error);
        while ((dir = va_arg(ap, const gchar*)))
            g_ptr_array_add(module_dirs, g_build_filename(p, *dir ? dir : NULL, NULL));
        va_end(ap);
        g_free(p);

        va_start(ap, error);
        while ((dir = va_arg(ap, const gchar*)))
            g_ptr_array_add(module_dirs, g_build_filename(p, "modules", *dir ? dir : NULL, NULL));
        va_end(ap);

        g_ptr_array_add(module_dirs, NULL);
        gwy_module_register_modules((const gchar**)module_dirs->pdata);

        for (i = 0; i < module_dirs->len-1; i++)
            g_free(module_dirs->pdata[i]);
        g_ptr_array_free(module_dirs, TRUE);
    }

    return ok;
}

/**
 * gwy_app_init_nongui:
 * @option: Name of the first initialisation option, or %NULL if there are no
 *          options.
 * @...: Value of the first option, if any, followed by another option name
 *       and value, etc.  Terminated by %NULL.
 *
 * Performs common initialisations useful in non-GUI (batch) programs utilising Gwyddion libraries.
 *
 * This function does not initialise GTK+ and does not set up anything related or requiring a display.  It sets up
 * logging to console, runs gwy_widgets_type_init(), disables undo, disables GUI for the data browser and waiting, and
 * loads resources, settings and all modules.  However, it prevents the plugin-proxy and pygwy modules from loading.
 *
 * There are currently the following options:
 *
 * "enable-threads" with #gboolean value.  If %TRUE or %FALSE is passed gwy_threads_set_enabled() will be called with
 * this settings.  By default, the multithread processing state is unchanged.
 *
 * "enable-pygwy" with #gboolean value.  If %TRUE is passed then pygwy will not be prevented from loading.  Passing
 * %FALSE is the same as not setting the option at all.
 *
 * So, for instance a program wanting to enable multithread processing in Gwyddion functions (provided it is built in)
 * can run
 *
 * |[
 * gwy_app_init_nongui("enable-threads", TRUE, NULL);
 * ]|
 **/
void
gwy_app_init_nongui(const gchar *option,
                    ...)
{
    gboolean want_threads = FALSE, set_thread_state = FALSE;
    gboolean want_pygwy = FALSE;
    va_list ap;
    gchar *settings_file;
    gchar **module_dirs;

    gwy_app_setup_logging(GWY_APP_LOGGING_TO_CONSOLE);

    va_start(ap, option);
    while (option) {
        if (gwy_strequal(option, "enable-threads")) {
            set_thread_state = TRUE;
            want_threads = va_arg(ap, gint);
        }
        else if (gwy_strequal(option, "enable-pygwy")) {
            want_pygwy = va_arg(ap, gint);
        }
        else {
            g_warning("Unknown init option %s.\n", option);
            break;
        }
        option = va_arg(ap, const gchar*);
    }
    va_end(ap);

    if (set_thread_state)
        gwy_threads_set_enabled(want_threads);

    gwy_module_disable_registration("plugin-proxy");
    if (!want_pygwy)
        gwy_module_disable_registration("pygwy");

    gwy_widgets_type_init();
    gwy_undo_set_enabled(FALSE);
    gwy_app_wait_set_enabled(FALSE);
    gwy_app_data_browser_set_gui_enabled(FALSE);

    /* Register resources */
    gwy_resource_class_load(g_type_class_peek(GWY_TYPE_GRADIENT));
    gwy_resource_class_load(g_type_class_peek(GWY_TYPE_GL_MATERIAL));
    gwy_resource_class_load(g_type_class_peek(GWY_TYPE_GRAIN_VALUE));

    /* Load settings */
    settings_file = gwy_app_settings_get_settings_filename();
    if (g_file_test(settings_file, G_FILE_TEST_IS_REGULAR))
        gwy_app_settings_load(settings_file, NULL);
    g_free(settings_file);
    gwy_app_settings_get();

    /* Register modules */
    module_dirs = gwy_app_settings_get_module_dirs();
    gwy_module_register_modules((const gchar**)module_dirs);
    g_strfreev(module_dirs);

    /* The Python initialisation somehow overrides SIGINT and Gwyddion can no
     * longer be terminated with Ctrl-C.  Fix it. */
    if (want_pygwy)
        signal(SIGINT, SIG_DFL);
}

/************************** Documentation ****************************/

/**
 * SECTION:app
 * @title: app
 * @short_description: Core application interface, window management
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
