/*
 *  $Id: gwyglmaterial.c 27696 2025-04-02 09:36:13Z yeti-dn $
 *  Copyright (C) 2005-2025 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>

#include "libgwyddion/gwymacros.h"
#include "libgwyddion/gwymath.h"
#include "libgwyddion/gwyutils.h"

#include "gwyglmaterial.h"

#include "gwydrawinternal.h"

G_STATIC_ASSERT(sizeof(GwyRGBA) == 4*sizeof(gdouble));

#define float_to_hex(c) ((guint8)((c)*255.9999999))

struct _GwyGLMaterialPrivate {
    GwyRGBA ambient;
    GwyRGBA diffuse;
    GwyRGBA specular;
    GwyRGBA emission;
    gdouble shininess;
};

static gpointer       resource_copy     (gpointer);
static void           reset_data        (GwyGLMaterial *gl_material);
static void           data_changed      (GwyGLMaterial *gl_material);
static GwyGLMaterial* create_gl_material(const gchar *name,
                                         const GwyRGBA *ambient,
                                         const GwyRGBA *diffuse,
                                         const GwyRGBA *specular,
                                         const GwyRGBA *emission,
                                         gdouble shininess,
                                         gboolean is_const);
static void           resource_dump     (GwyResource *resource,
                                         GString *str);
static GwyResource*   resource_parse    (const gchar *text);

static GwyResourceClass *parent_class = NULL;
static GwyInventory *inventory = NULL;

static const GwyRGBA null_color = { 0, 0, 0, 0 };

G_DEFINE_TYPE_WITH_CODE(GwyGLMaterial, gwy_gl_material, GWY_TYPE_RESOURCE,
                        G_ADD_PRIVATE(GwyGLMaterial))

static void
gwy_gl_material_class_init(GwyGLMaterialClass *klass)
{
    GwyResourceClass *res_class = GWY_RESOURCE_CLASS(klass);

    parent_class = gwy_gl_material_parent_class;

    res_class->item_type = *gwy_resource_class_get_item_type(parent_class);
    res_class->item_type.type = G_TYPE_FROM_CLASS(klass);
    res_class->item_type.copy = resource_copy;

    res_class->name = "glmaterials";
    res_class->inventory = inventory = gwy_inventory_new(&res_class->item_type);
    gwy_inventory_set_default_item_name(inventory, GWY_GL_MATERIAL_DEFAULT);
    res_class->dump = resource_dump;
    res_class->parse = resource_parse;
}

static void
gwy_gl_material_init(GwyGLMaterial *gl_material)
{
    gl_material->priv = gwy_gl_material_get_instance_private(gl_material);
    reset_data(gl_material);
}

/**
 * set_rgba_validated:
 * @src: A color.
 * @dest: A color to update.
 *
 * Assigns two colors, fixing color components to range -1..1.
 *
 * Returns: Whether the color has actually changed.
 **/
static inline gboolean
set_rgba_validated(const GwyRGBA *src, GwyRGBA *dest)
{
    gboolean changed = FALSE;
    gdouble c;

    if ((c = CLAMP(src->r, -1.0, 1.0)) != dest->r) {
        if (c != src->r)
            g_warning("Red component outside -1..1 range");
        dest->r = c;
        changed = TRUE;
    }
    if ((c = CLAMP(src->g, -1.0, 1.0)) != dest->g) {
        if (c != src->g)
            g_warning("Green component outside -1..1 range");
        dest->g = c;
        changed = TRUE;
    }
    if ((c = CLAMP(src->b, -1.0, 1.0)) != dest->b) {
        if (c != src->b)
            g_warning("Blue component outside -1..1 range");
        dest->b = c;
        changed = TRUE;
    }
    if ((c = CLAMP(src->a, -1.0, 1.0)) != dest->a) {
        if (c != src->a)
            g_warning("Alpha component outside -1..1 range");
        dest->a = c;
        changed = TRUE;
    }

    return changed;
}

/**
 * gwy_gl_material_get_ambient:
 * @gl_material: A GL material.
 *
 * Gets the ambient reflectance of a GL material.
 *
 * Returns: Ambient reflectance (owned by GL material, must not be modified nor freed).
 **/
const GwyRGBA*
gwy_gl_material_get_ambient(GwyGLMaterial *gl_material)
{
    g_return_val_if_fail(GWY_IS_GL_MATERIAL(gl_material), &null_color);
    return &gl_material->priv->ambient;
}

/**
 * gwy_gl_material_set_ambient:
 * @gl_material: A GL material.
 * @ambient: Ambient reflectance.
 *
 * Sets the ambient reflectance of a GL material.
 **/
void
gwy_gl_material_set_ambient(GwyGLMaterial *gl_material,
                            const GwyRGBA *ambient)
{
    g_return_if_fail(GWY_IS_GL_MATERIAL(gl_material));
    g_return_if_fail(gwy_resource_is_modifiable(GWY_RESOURCE(gl_material)));
    if (set_rgba_validated(ambient, &gl_material->priv->ambient))
        data_changed(gl_material);
}

/**
 * gwy_gl_material_get_diffuse:
 * @gl_material: A GL material.
 *
 * Gets the diffuse reflectance of a GL material.
 *
 * Returns: Diffuse reflectance (owned by GL material, must not be modified nor freed).
 **/
const GwyRGBA*
gwy_gl_material_get_diffuse(GwyGLMaterial *gl_material)
{
    g_return_val_if_fail(GWY_IS_GL_MATERIAL(gl_material), &null_color);
    return &gl_material->priv->diffuse;
}

/**
 * gwy_gl_material_set_diffuse:
 * @gl_material: A GL material.
 * @diffuse: Diffuse reflectance.
 *
 * Sets the diffuse reflectance of a GL material.
 **/
void
gwy_gl_material_set_diffuse(GwyGLMaterial *gl_material,
                            const GwyRGBA *diffuse)
{
    g_return_if_fail(GWY_IS_GL_MATERIAL(gl_material));
    g_return_if_fail(gwy_resource_is_modifiable(GWY_RESOURCE(gl_material)));
    if (set_rgba_validated(diffuse, &gl_material->priv->diffuse))
        data_changed(gl_material);
}

/**
 * gwy_gl_material_get_specular:
 * @gl_material: A GL material.
 *
 * Gets the specular reflectance of a GL material.
 *
 * Returns: Specular reflectance (owned by GL material, must not be modified nor freed).
 **/
const GwyRGBA*
gwy_gl_material_get_specular(GwyGLMaterial *gl_material)
{
    g_return_val_if_fail(GWY_IS_GL_MATERIAL(gl_material), &null_color);
    return &gl_material->priv->specular;
}

/**
 * gwy_gl_material_set_specular:
 * @gl_material: A GL material.
 * @specular: Specular reflectance.
 *
 * Sets the specular reflectance of a GL material.
 **/
void
gwy_gl_material_set_specular(GwyGLMaterial *gl_material,
                             const GwyRGBA *specular)
{
    g_return_if_fail(GWY_IS_GL_MATERIAL(gl_material));
    g_return_if_fail(gwy_resource_is_modifiable(GWY_RESOURCE(gl_material)));
    if (set_rgba_validated(specular, &gl_material->priv->specular))
        data_changed(gl_material);
}

/**
 * gwy_gl_material_get_emission:
 * @gl_material: A GL material.
 *
 * Gets the emission component of a GL material.
 *
 * Returns: Emission component (owned by GL material, must not be modified nor freed).
 **/
const GwyRGBA*
gwy_gl_material_get_emission(GwyGLMaterial *gl_material)
{
    g_return_val_if_fail(GWY_IS_GL_MATERIAL(gl_material), &null_color);
    return &gl_material->priv->emission;
}

/**
 * gwy_gl_material_set_emission:
 * @gl_material: A GL material.
 * @emission: Emission component.
 *
 * Sets the emission component of a GL material.
 **/
void
gwy_gl_material_set_emission(GwyGLMaterial *gl_material,
                             const GwyRGBA *emission)
{
    g_return_if_fail(GWY_IS_GL_MATERIAL(gl_material));
    g_return_if_fail(gwy_resource_is_modifiable(GWY_RESOURCE(gl_material)));
    if (set_rgba_validated(emission, &gl_material->priv->emission))
        data_changed(gl_material);
}

/**
 * gwy_gl_material_get_shininess:
 * @gl_material: A GL material.
 *
 * Gets the shininess value of a GL material.
 *
 * Returns: The shininess value as a value in interval [0,1].
 **/
gdouble
gwy_gl_material_get_shininess(GwyGLMaterial *gl_material)
{
    g_return_val_if_fail(GWY_IS_GL_MATERIAL(gl_material), 0.0);
    return gl_material->priv->shininess;
}

/**
 * gwy_gl_material_set_shininess:
 * @gl_material: A GL material.
 * @shininess: Shinniness value in interval [0,1].
 *
 * Sets the shininess value of a GL material.
 **/
void
gwy_gl_material_set_shininess(GwyGLMaterial *gl_material,
                              gdouble shininess)
{
    g_return_if_fail(GWY_IS_GL_MATERIAL(gl_material));
    g_return_if_fail(gwy_resource_is_modifiable(GWY_RESOURCE(gl_material)));
    GwyGLMaterialPrivate *priv = gl_material->priv;
    shininess = gwy_clamp(shininess, 0.0, 1.0);
    if (shininess != priv->shininess) {
        priv->shininess = shininess;
        data_changed(gl_material);
    }
}

/**
 * gwy_gl_material_reset:
 * @gl_material: A GL material.
 *
 * Resets a GL material to default values.
 **/
void
gwy_gl_material_reset(GwyGLMaterial *gl_material)
{
    g_return_if_fail(GWY_IS_GL_MATERIAL(gl_material));
    g_return_if_fail(gwy_resource_is_modifiable(GWY_RESOURCE(gl_material)));
    reset_data(gl_material);
    data_changed(gl_material);
}

static void
reset_data(GwyGLMaterial *gl_material)
{
    GwyGLMaterialPrivate *priv = gl_material->priv;
    /* OpenGL reference defines the following default. */
    priv->ambient = (GwyRGBA){ 0.2, 0.2, 0.2, 1.0 };
    priv->diffuse = (GwyRGBA){ 0.8, 0.8, 0.8, 1.0 };
    priv->specular = (GwyRGBA){ 0.0, 0.0, 0.0, 1.0 };
    priv->emission = (GwyRGBA){ 0.0, 0.0, 0.0, 1.0 };
    priv->shininess = 0.0;
}

/**
 * gwy_gl_material_sample_to_pixbuf:
 * @gl_material: A GL material to sample.
 * @pixbuf: A pixbuf to sample gl_material to (in horizontal direction).
 *
 * Samples GL material to a provided pixbuf.
 **/
void
gwy_gl_material_sample_to_pixbuf(GwyGLMaterial *gl_material,
                                 GdkPixbuf *pixbuf)
{
    g_return_if_fail(GWY_IS_GL_MATERIAL(gl_material));
    g_return_if_fail(GDK_IS_PIXBUF(pixbuf));

    GwyGLMaterialPrivate *priv = gl_material->priv;
    gint width = gdk_pixbuf_get_width(pixbuf);
    gint height = gdk_pixbuf_get_height(pixbuf);
    gint rowstride = gdk_pixbuf_get_rowstride(pixbuf);
    gboolean has_alpha = gdk_pixbuf_get_has_alpha(pixbuf);
    gint bpp = 3 + (has_alpha ? 1 : 0);
    guchar *pdata = gdk_pixbuf_get_pixels(pixbuf);

    gdouble q = (width == 1) ? 0.0 : 1.0/(width - 1.0);
    gdouble p = (height == 1) ? 0.0 : 1.0/(height - 1.0);
    guint8 alpha = float_to_hex(priv->ambient.a);

    for (gint j = 0; j < width; j++) {
        gdouble VRp = j*q*(2.0 - j*q);
        gdouble s = pow(VRp, 128.0*priv->shininess);
        GwyRGBA s0;

        s0.r = priv->emission.r + 0.3*priv->ambient.r + priv->specular.r*s;
        s0.g = priv->emission.g + 0.3*priv->ambient.g + priv->specular.g*s;
        s0.b = priv->emission.b + 0.3*priv->ambient.b + priv->specular.b*s;

        for (gint i = 0; i < height; i++) {
            guchar *row = pdata + i*rowstride + j*bpp;
            gdouble LNp = 1.0 - i*p;
            gdouble v;

            v = s0.r + priv->diffuse.r*LNp;
            *(row++) = float_to_hex(v);

            v = s0.g + priv->diffuse.g*LNp;
            *(row++) = float_to_hex(v);

            v = s0.b + priv->diffuse.b*LNp;
            *(row++) = float_to_hex(v);

            if (has_alpha)
                *(row++) = alpha;
        }
    }
}

static void
data_changed(GwyGLMaterial *gl_material)
{
    gwy_resource_data_changed(GWY_RESOURCE(gl_material));
}

void
_gwy_gl_material_class_setup_presets(void)
{
    /* Force class instantiation, this function is called before it's first referenced. */
    GwyResourceClass *klass = g_type_class_ref(GWY_TYPE_GL_MATERIAL);

    GwyGLMaterial *default_material = create_gl_material(GWY_GL_MATERIAL_DEFAULT,
                                                         NULL, NULL, NULL, NULL, 0.0,
                                                         TRUE);
    gwy_inventory_insert_item(klass->inventory, default_material);
    g_object_unref(default_material);

    GwyGLMaterial *none_material = create_gl_material(GWY_GL_MATERIAL_NONE,
                                                      &null_color, &null_color, &null_color, &null_color, 0.0,
                                                      TRUE);
    gwy_inventory_insert_item(klass->inventory, none_material);
    g_object_unref(none_material);

    /* The gl_material added a reference so we can safely unref it again */
    g_type_class_unref(klass);
}

static GwyGLMaterial*
create_gl_material(const gchar *name,
                   const GwyRGBA *ambient, const GwyRGBA *diffuse, const GwyRGBA *specular, const GwyRGBA *emission,
                   gdouble shininess,
                   gboolean is_const)
{
    GwyGLMaterial *gl_material = g_object_new(GWY_TYPE_GL_MATERIAL,
                                              "name", name,
                                              "const", is_const,
                                              NULL);
    GwyGLMaterialPrivate *priv = gl_material->priv;
    if (ambient)
        priv->ambient = *ambient;
    if (diffuse)
        priv->diffuse = *diffuse;
    if (specular)
        priv->specular = *specular;
    if (emission)
        priv->emission = *emission;
    if (shininess >= 0)
        priv->shininess = shininess;

    return gl_material;
}

static gpointer
resource_copy(gpointer item)
{
    g_return_val_if_fail(GWY_IS_GL_MATERIAL(item), NULL);

    GwyGLMaterial *gl_material = GWY_GL_MATERIAL(item);
    GwyGLMaterialPrivate *priv = gl_material->priv;
    GwyGLMaterial *copy = create_gl_material(gwy_resource_get_name(GWY_RESOURCE(item)),
                                             &priv->ambient, &priv->diffuse, &priv->specular, &priv->emission,
                                             priv->shininess,
                                             FALSE);

    return copy;
}

static void
resource_dump(GwyResource *resource, GString *str)
{
    g_return_if_fail(GWY_IS_GL_MATERIAL(resource));
    GwyGLMaterial *gl_material = GWY_GL_MATERIAL(resource);
    GwyGLMaterialPrivate *priv = gl_material->priv;

    gwy_append_doubles_to_gstring(str, (gdouble*)&priv->ambient, 4, 6, " ", TRUE);
    g_string_append_c(str, '\n');
    gwy_append_doubles_to_gstring(str, (gdouble*)&priv->diffuse, 4, 6, " ", TRUE);
    g_string_append_c(str, '\n');
    gwy_append_doubles_to_gstring(str, (gdouble*)&priv->specular, 4, 6, " ", TRUE);
    g_string_append_c(str, '\n');
    gwy_append_doubles_to_gstring(str, (gdouble*)&priv->emission, 4, 6, " ", TRUE);
    g_string_append_c(str, '\n');
    gwy_append_doubles_to_gstring(str, &priv->shininess, 1, 6, " ", TRUE);
    g_string_append_c(str, '\n');
}

static GwyResource*
resource_parse(const gchar *text)
{
    GwyGLMaterial *gl_material = NULL;
    gint nrows = 4, ncols = 4;
    gdouble data[4*4], shininess;

    g_return_val_if_fail(text, NULL);
    GwyGLMaterialClass *klass = g_type_class_peek(GWY_TYPE_GL_MATERIAL);
    g_return_val_if_fail(klass, NULL);

    gchar *end;
    if (!gwy_parse_doubles(text, data, 0, &nrows, &ncols, &end, NULL)) {
        g_warning("Cannot parse GL material.");
        return NULL;
    }
    nrows = ncols = 1;
    if (!gwy_parse_doubles(end, &shininess, GWY_PARSE_DOUBLES_COMPLETELY, &nrows, &ncols, NULL, NULL)) {
        g_warning("Cannot parse GL material.");
        return NULL;
    }
    shininess = CLAMP(shininess, 0.0, 1.0);

    gl_material = create_gl_material("", NULL, NULL, NULL, NULL, shininess, FALSE);
    GwyGLMaterialPrivate *priv = gl_material->priv;

    set_rgba_validated(&priv->ambient, (GwyRGBA*)(data + 0));
    set_rgba_validated(&priv->diffuse, (GwyRGBA*)(data + 4));
    set_rgba_validated(&priv->specular, (GwyRGBA*)(data + 8));
    set_rgba_validated(&priv->emission, (GwyRGBA*)(data + 12));

    return GWY_RESOURCE(gl_material);
}

/**
 * gwy_gl_materials:
 *
 * Gets inventory with all the GL materials.
 *
 * Returns: GL material inventory.
 **/
GwyInventory*
gwy_gl_materials(void)
{
    return inventory;
}

/**
 * gwy_gl_materials_get_gl_material:
 * @name: GL material name.  May be %NULL to get the default GL material.
 *
 * Convenience function to get a GL material from gwy_gl_materials() by name.
 *
 * Returns: GL material identified by @name or the default GL material if @name does not exist.
 **/
GwyGLMaterial*
gwy_gl_materials_get_gl_material(const gchar *name)
{
    g_return_val_if_fail(inventory, NULL);
    return (GwyGLMaterial*)gwy_inventory_get_item_or_default(inventory, name);
}

/************************** Documentation ****************************/

/**
 * SECTION:gwyglmaterial
 * @title: GwyGLMaterial
 * @short_description: OpenGL material representation
 * @see_also: #Gwy3DView -- 3D data display widget,
 *            #GwyRGBA -- representation of color components,
 *            #GwyInventory -- the container holding all GL materials
 *
 * #GwyGLMaterial represents an OpenGL material.  Its properties directly map to corresponding OpenGL material
 * characteristics, all are in the range [0,1].
 *
 * Gradient objects can be obtained from gwy_gl_materials_get_gl_material(). New GL materials can be created with
 * gwy_inventory_new_item() on the #GwyInventory returned by gwy_gl_materials().
 **/

/**
 * GWY_GL_MATERIAL_DEFAULT:
 *
 * The name of the default OpenGL material.
 *
 * It is guaranteed to always exist.
 *
 * Note this is not the same as user's default material which corresponds to the default item in gwy_gl_materials()
 * inventory and it change over time.
 **/

/**
 * GWY_GL_MATERIAL_NONE:
 *
 * The name of special void material with all characteristics zero.
 *
 * It is guaranteed to exist, but you should rarely actually need it.
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
