/*
 *  $Id: serializable.c 27546 2025-03-17 18:08:22Z yeti-dn $
 *  Copyright (C) 2009-2025 David Nečas (Yeti).
 *  E-mail: yeti@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>
#include "serializable.h"

G_DEFINE_INTERFACE(GwySerializable, gwy_serializable, G_TYPE_OBJECT);

static void
gwy_serializable_default_init(G_GNUC_UNUSED GwySerializableInterface *iface)
{
}

/**
 * gwy_serializable_n_items:
 * @serializable: A serializable object.
 *
 * Provides an upper bound of the number of items the flattened representation of an object will take.
 *
 * This function wraps the virtual table method n_items().   It increases the returned value by the number of item
 * used to represent the object header (which is 1).
 *
 * Returns: An upper bound of the number of items @serializable will serialize to.
 **/
gsize
gwy_serializable_n_items(GwySerializable *serializable)
{
    GwySerializableInterface *iface = GWY_SERIALIZABLE_GET_INTERFACE(serializable);
    g_return_val_if_fail(iface && iface->n_items, 1);

    return iface->n_items(serializable) + 1;
}

/**
 * gwy_serializable_itemize:
 * @serializable: A serializable object.
 * @items: List of flattened object tree items to append the representation of @serializable to.
 *
 * Creates the flattened representation of a serializable object.
 *
 * This function wraps the virtual table method itemize().  It deals with recording of the object header item.  Method
 * itemize() then only serialises the actual data of the object.
 **/
void
gwy_serializable_itemize(GwySerializable *serializable,
                         GwySerializableItems *items)
{
    GwySerializableInterface *iface = GWY_SERIALIZABLE_GET_INTERFACE(serializable);
    g_return_if_fail(iface && iface->itemize);

    g_return_if_fail(items->n < items->len);
    GwySerializableItem *item = items->items + items->n;
    items->n++;
    item->name = G_OBJECT_TYPE_NAME(G_OBJECT(serializable));
    item->value.v_size = 0;
    item->ctype = GWY_SERIALIZABLE_HEADER;
    item->array_size = iface->itemize(serializable, items);
}

/**
 * gwy_serializable_done:
 * @serializable: A serializable object.
 *
 * Frees temporary storage allocated by object itemization.
 *
 * This function calls the virtual table method done(), if the class has any.
 **/
void
gwy_serializable_done(GwySerializable *serializable)
{
    GwySerializableInterface *iface = GWY_SERIALIZABLE_GET_INTERFACE(serializable);
    g_return_if_fail(iface);

    if (iface->done)
        iface->done(serializable);
}

/**
 * gwy_serializable_copy:
 * @serializable: A serializable object.
 *
 * Creates an object with identical value.
 *
 * This is a copy-constructor creating a deep copy.
 *
 * You can duplicate a %NULL, too, but you are discouraged from doing it.
 *
 * #GwySerializable implementations are encouraged to provide my_object_copy() convenience wrappers with argument and
 * return value of the specific type and the corresponding type checking.
 *
 * Returns: (transfer full):
 *          The newly created object copy.
 **/
GwySerializable*
gwy_serializable_copy(GwySerializable *serializable)
{
    if (!serializable)
        return NULL;

    GwySerializableInterface *iface = GWY_SERIALIZABLE_GET_INTERFACE(serializable);
    g_return_val_if_fail(iface && iface->copy, NULL);

    GwySerializable *copy = iface->copy(serializable);
    return copy ? copy : g_object_new(G_TYPE_FROM_INSTANCE(serializable), NULL);
}

/**
 * gwy_serializable_assign:
 * @destination: An object of the same type as @source. More precisely, @source may be subclass of @destination (the
 *               extra information is lost then).
 * @source: A serializable object.
 *
 * Copies the value of an object to another object.
 *
 * If @destination is the same object as @source the function is no-op.
 *
 * What constitutes ‘object value’ depends on the class. The general rule is that @destination will look as the result
 * of gwy_serializable_copy(). However, it will keep its identity (address, connected signal handlers, references,
 * object data, etc.). The operation may be more efficient than constructing a new object but it may be not.
 *
 * #GwySerializable implementations are encouraged to provide my_object_assign() convenience wrappers with arguments
 * of the specific type and the corresponding type checking.
 **/
void
gwy_serializable_assign(GwySerializable *destination,
                        GwySerializable *source)
{
    g_return_if_fail(source);
    g_return_if_fail(destination);
    if (source == destination)
        return;

    GwySerializableInterface *iface = GWY_SERIALIZABLE_GET_INTERFACE(destination);
    g_return_if_fail(iface && iface->assign);
    /* No need to check GWY_IS_SERIALIZABLE(source) if this holds. */
    g_return_if_fail(g_type_is_a(G_TYPE_FROM_INSTANCE(source), G_TYPE_FROM_INSTANCE(destination)));

    iface->assign(destination, source);
}

/**
 * SECTION: serializable
 * @title: GwySerializable
 * @short_description: Serialisable value-like object interface
 *
 * #GwySerializable is an abstract interface for data-like objects that can be serialised and deserialised.  You can
 * serialise any object implementing this interface with functions such as gwy_serialize_gio() and the restore
 * (deserialise) it with gwy_deserialize_memory().
 *
 * Gwyddion implements a simple serialisation model: only tree-like structures of objects, formed by ownership, can be
 * serialised and restored.  Any inter-object relations other than plain ownership must be stored in some weak form
 * and there is no explicit support for this.  Moreover, only object values are preserved, their identities are lost
 * (in particular, signals, user data and similar attributes are not subject to serialisation and deserialisation).
 * This, on the other hand, means that any saved object can be restored individually and independently and still be in
 * a meaningful state.
 *
 * Beside saving and restoration, all serialisable classes implement a copy-constructor that creates a duplicate of an
 * existing object.  This constructor is invoked with gwy_serializable_copy().  Furthermore, the value of one
 * such object can be transferred to another object of the same class (or a superclass), similarly to overriden
 * assignment operators in OO languages.  This assignment is performed with gwy_serializable_assign().
 *
 * Most classes that implement #GwySerializable define convenience macros for the copy-constructor and assignment
 * operator, called
 * <function>gwy_foo_copy<!-- -->()</function> and <function>gwy_foo_assign<!-- -->()</function>, respectively,
 * where <literal>foo</literal> is the lowercase class name.
 *
 * <refsect2 id='libgwy-serializable-implementing'>
 * <title>Implementing #GwySerializable</title>
 * <para>You can implement serialisation and deserialisation in your classes.  Use the %G_IMPLEMENT_ITERFACE macro
 * to generate the boilerplate interface implementation code. The copy constructor and assignment operators are
 * usually easy so we focus on serialisation and deserialisation (with public object data for simplicity). There
 * are a number of helper functions, such as gwy_deserialize_filter() for getting only the interesting items in
 * deserialisation or gwy_serialize_item_double() for handling the itemisation of a single value.</para>
 * |[
 * // Standard object structure definitions
 * typedef struct _MyObject      MyObject;
 * typedef struct _MyObjectClass MyObjectClass;
 *
 * struct _MyObject {
 *     GObject g_object;
 *     gint32 data;
 * };
 *
 * struct _MyObjectClass {
 *     GObjectClass g_object_class;
 * };
 *
 * // Define the type
 * G_DEFINE_TYPE_WITH_CODE(MyObject, my_object, G_TYPE_OBJECT,
 *                         G_IMPLEMENT_INTERFACE(GWY_TYPE_SERIALIZABLE, serializable_init));
 *
 * // Serialised data specification, used both in itemize() and construct()
 * static const GwySerializableItem default_items[] = {
 *     { .name = "data", .ctype = GWY_SERIALIZABLE_INT32, },
 * };
 *
 * // Implement the interface
 * static void
 * serializable_init(GwySerializableInterface *iface)
 * {
 *     iface->n_items   = serializable_n_items;
 *     iface->itemize   = serializable_itemize;
 *     iface->construct = serializable_construct;
 *     iface->copy      = serializable_copy;
 *     iface->assign    = serializable_assign;
 * }
 *
 * // Duplication is easy, create a new object and set data
 * static GwySerializable*
 * serializable_copy(GwySerializable *serializable)
 * {
 *     MyObject *myobject = GWY_MY_OBJECT(serializable);
 *     MyObject *copy = g_object_newv(MY_TYPE_OBJECT, 0, NULL);
 *
 *     copy->data = myobject->data;
 *     return GWY_SERIALIZABLE(copy);
 * }
 *
 * // Assigning is even easier, just set data
 * static void
 * serializable_assign(GwySerializable *destination,
 *                     GwySerializable *source)
 * {
 *     MyObject *myobject = GWY_MY_OBJECT(serializable);
 *     MyObject *src = GWY_MY_OBJECT(source);
 *
 *     src->data = myobject->data;
 * }
 *
 * // Our object needs at most one item and it does not contain any child objects.  Thus we can simply return
 * // a constant.
 * static gsize
 * serializable_n_items(G_GNUC_UNUSED GwySerializable *serializable)
 * {
 *     return 1;
 * }
 *
 * // Fill the item list with actual object data.  This is a bit
 * // overcomplicated for such a simple object as MyObject, of course.
 * static gsize
 * serializable_itemize(GwySerializable *serializable,
 *                      GwySerializableItems *items)
 * {
 *     MyObject *myobject = GWY_MY_OBJECT(serializable);
 *     GwySerializableItem it[G_N_ELEMENTS(default_items)];
 *     gsize n_items = 0;
 *
 *     memcpy(it, default_items, sizeof(default_items));
 *
 *     // Do not store the data if it has the default value
 *     if (myobject->data) {
 *         g_return_val_if_fail(items->len - items->n, 0);
 *         items->items[items->n].value.v_int32 = myobject->data;
 *         items->n++;
 *         n_items++;
 *     }
 *
 *     return n_items;
 * }
 *
 * // Construct the object from item list.  The data item does not need to be present, in this case the default zero
 * // value is used.  Also, for more complicated objects the deserialisation can fail or resources need to be released;
 * // see GwySerializableInterface for the clean-up rules.
 * static gboolean
 * serializable_construct(GwySerializable *serializable,
 *                        GwySerializableItems *items,
 *                        GwyErrorList **error_list)
 * {
 *     GwySerializableItem it[G_N_ELEMENTS(default_items)];
 *
 *     memcpy(it, default_items, sizeof(default_items));
 *     gwy_deserialize_filter_items(it, G_N_ELEMENTS(it), items, "MyObject", error_list);
 *
 *     MyObject *myobject = GWY_MY_OBJECT(serializable);
 *     myobject->data = it[0].value.v_int32;
 *
 *     return TRUE;
 * }
 * ]|
 * </refsect2>
 **/

/**
 * GwySerializable:
 *
 * Formal type of serialisable objects.
 **/

/**
 * GwySerializableCType:
 * @GWY_SERIALIZABLE_HEADER: Reserved type value which does not appear in files and denotes object header items
 *                           internally. The serialiser/deserialised sets @size to the object data size and
 *                           @array_size to the number of items.
 * @GWY_SERIALIZABLE_CONSUMED: Reserved type value which does not appear in files and denotes already consumed items
 *                             internally.
 * @GWY_SERIALIZABLE_INT8: Denotes a character (8bit integer).
 * @GWY_SERIALIZABLE_INT8_ARRAY: Denotes a character (8bit integer) array.
 * @GWY_SERIALIZABLE_BOOLEAN: Denotes a one-byte boolean.
 * @GWY_SERIALIZABLE_INT16: Denotes a 16bit integer.
 * @GWY_SERIALIZABLE_INT16_ARRAY: Denotes an array of 16bit integers.
 * @GWY_SERIALIZABLE_INT32: Denotes a 32bit integer.
 * @GWY_SERIALIZABLE_INT32_ARRAY: Denotes an array of 32bit integers.
 * @GWY_SERIALIZABLE_INT64: Denotes a 64bit integer.
 * @GWY_SERIALIZABLE_INT64_ARRAY: Denotes an array of 64bit integers.
 * @GWY_SERIALIZABLE_DOUBLE: Denotes a IEEE double.
 * @GWY_SERIALIZABLE_DOUBLE_ARRAY: Denotes an array of IEEE doubles.
 * @GWY_SERIALIZABLE_STRING: Denotes a UTF-8-encoded C string.  If you need a raw sequence of bytes, use
 *                           %GWY_SERIALIZABLE_INT8_ARRAY instead.
 * @GWY_SERIALIZABLE_STRING_ARRAY: Denotes an array of UTF-8 encoded C strings.
 * @GWY_SERIALIZABLE_OBJECT: Denotes an object.
 * @GWY_SERIALIZABLE_OBJECT_ARRAY: Denotes an array of objects.
 * @GWY_SERIALIZABLE_BOXED: Denotes a serialisable boxed type.
 *
 * Type of serialisable value.
 *
 * The type is a single byte, i.e. a value smaller than 256.  It is equal to the character used in GWY files to denote
 * the corresponding type.
 *
 * Signed and usinged types are not distinguished here as their byte order is handled the same way.
 **/

/**
 * GwySerializableValue:
 * @v_boolean: The value as a #gboolean.
 * @v_int8: The value as a #gint8.
 * @v_uint8: The value as a #guint8.
 * @v_int16: The value as a #gint16.
 * @v_uint16: The value as a #guint16.
 * @v_int32: The value as a #gint32.
 * @v_uint32: The value as a #guint32.
 * @v_int64: The value as a #gint64.
 * @v_uint64: The value as a #guint64.
 * @v_double: The value as a #gdouble.
 * @v_string: The value as a #gchar pointer.
 * @v_ustring: The value as a #guchar pointer.
 * @v_object: The value as an object.
 * @v_boxed: The value as a boxed type.
 * @v_size: The value as a #gsize.
 * @v_type: The value as a #GType.
 * @v_int8_array: The value as an array of #gint8<!-- -->s.
 * @v_uint8_array: The value as an array of #guint8<!-- -->s.
 * @v_int16_array: The value as an array of #gint16<!-- -->s.
 * @v_uint16_array: The value as an array of #guint16<!-- -->s.
 * @v_int32_array: The value as an array of #gint32<!-- -->s.
 * @v_uint32_array: The value as an array of #guint32<!-- -->s.
 * @v_int64_array: The value as an array of #gint64<!-- -->s.
 * @v_uint64_array: The value as an array of #guint64<!-- -->s.
 * @v_double_array: The value as an array of #gdouble<!-- -->s.
 * @v_string_array: The value as an array of #gchar pointers.
 * @v_ustring_array: The value as an array of #guchar pointers.
 * @v_object_array: The value as an array of objects.
 *
 * Representation of individual values of object serialisable data.
 *
 * See #GwySerializableCType for corresponding type specifiers.
 *
 * Member @v_size has no use in implementations (it corresponds to %GWY_SERIALIZABLE_HEADER), it is used only in
 * object header items in the flattened object tree.
 *
 * All string values must be valid UTF-8.  If you need raw byte sequences use arrays of #guint8.
 *
 * Signed and unsigned integer members are provided for convenience, the serialisation does not distinguish between
 * signed and unsigned integers.
 **/

/**
 * GwySerializableItem:
 * @name: Component name.
 * @array_size: Array size of the component if of an array type.  For boxed types that do not carry type information,
 *              deserialisation sets is to the #GType of the boxed type.  Unused otherwise.
 * @value: Item value, the interpreration depends on the @ctype member.
 * @ctype: Item type, one of the #GwySerializableCType enum.
 *
 * Information about one object component that is subject to serialisation or
 * deserialisation.
 **/

/**
 * GwySerializableItems:
 * @len: Allocated number of items.
 * @n: Number of items present.
 * @items: Array of items of total size @len with @n positions occupied.
 *
 * Flattened tree structure of object data used during serialisation and deserialisation.
 **/

/**
 * GwySerializableInterface:
 * @n_items: <para>Returns the number of items the object will flatten to, including items of all contained objects
 *           (but excluding the object header items of self).  It should use gwy_serializable_n_items() to calculate
 *           the numbers of items the contained objects will flatten to.</para>
 *           <para>This method may return a reasonable upper estimate instead of the exact number; for instance, if
 *           objects of some class are known to need 5 to 7 items, n_items() can simply return 7.</para>
 * @itemize: <para>Appends flattened representation of the object data to the @items list.  It should use
 *           gwy_serializable_itemize() to add the data of contained objects.</para>
 *           <para>The number of items corresponding to this object is returned, <emphasis>not including</emphasis>
 *           items of contained objects: every contained object counts only as one item.</para>
 * @done: <para>Frees all temporary data created by #GwySerializableInterface.itemize().  This method is optional but
 *        if it is defined it is guaranteed to be called after each #GwySerializableInterface.itemize().</para>
 * @construct: <para>Deserialises an object from array of flattened data items.  The first argument is the object of
 *             the final class (i.e. it may be a derived class), constructed with the default parameterless
 *             constructor.</para>
 *             <para>All strings, objects and arrays in the item list are newly allocated.  The method can (and,
 *             generally, should) take ownership by filling corresponding item values with %NULL, setting their
 *             @array_size to zero and setting the type to %GWY_SERIALIZABLE_CONSUMED.  The caller takes care of
 *             freeing them if they are not consumed by construct().</para>
 *             <para>This method must make any assumptions about the contents of @items.  Commonly, however, only
 *             items from a small predefined set are expected and then gwy_deserialize_filter_items() can be used to
 *             simplify the processing of the item list.</para>
 *             <para>It is possible to chain this method to the parent class and pass it the the item list or a part
 *             of it, if you are careful.</para>
 * @copy: <para>Creates a new object with all data identical to this object.  This method is expected to create
 *        a deep copy. Classes may provide other methods for shallow copies.</para>
 * @assign: <para>Makes all data of this object identical to the data of another object of the same class.
 *          Implementations may assume that the is-a relation is satisfied for the source object.  This method is
 *          expected to perform a deep copy.  Classes may provide other methods for shallow copies.</para>
 *
 * Interface implemented by serialisable objects.
 *
 * The object class must implement all the methods, except #GwySerializableInterface.done() which is optional.
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
