/*
 *  $Id: xyz.c 27641 2025-03-27 14:54:28Z yeti-dn $
 *  Copyright (C) 2025 David Necas (Yeti).
 *  E-mail: yeti@gwyddion.net.
 *
 *  The quicksort algorithm was copied from GNU C library, Copyright (C) 1991, 1992, 1996, 1997, 1999 Free Software
 *  Foundation, Inc.  See below.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>
#include <stdlib.h>

#include "gwymacros.h"
#include "gwymath.h"
#include "serializable-boxed.h"
#include "serializable-utils.h"
#include "xyz.h"

#define TYPE_NAME "GwyXYZ"

enum {
    ITEM_X,
    ITEM_Y,
    ITEM_Z,
    NUM_ITEMS,
};

static gsize    itemize  (gpointer boxed,
                          GwySerializableItems *items);
static gpointer construct(GwySerializableItems *items,
                          GwyErrorList **error_list);

static const GwySerializableItem serializable_items[NUM_ITEMS] = {
    { .name = "x", .ctype = GWY_SERIALIZABLE_DOUBLE, },
    { .name = "y", .ctype = GWY_SERIALIZABLE_DOUBLE, },
    { .name = "z", .ctype = GWY_SERIALIZABLE_DOUBLE, },
};

GType
gwy_xyz_get_type(void)
{
    /* Threads: type registered from gwy_types_init(). */
    static GType type = 0;

    if (G_UNLIKELY(!type)) {
        type = g_boxed_type_register_static("GwyXYZ", (GBoxedCopyFunc)gwy_xyz_copy, (GBoxedFreeFunc)gwy_xyz_free);
        static const GwySerializableBoxedInfo boxed_info = {
            sizeof(GwyXYZ), NUM_ITEMS, itemize, construct, NULL, NULL,
        };
        gwy_serializable_boxed_register_static(type, &boxed_info);
    }

    return type;
}

/**
 * gwy_xyz_new:
 * @x: X-coordinate.
 * @y: Y-coordinate.
 * @z: Z-coordinate.
 *
 * Creates Cartesian coordinates in space.
 *
 * This is mostly useful for language bindings.
 *
 * Returns: New XYZ structure.  The result should be freed using gwy_xyz_free().
 **/
GwyXYZ*
gwy_xyz_new(gdouble x, gdouble y, gdouble z)
{
    GwyXYZ *xyz = g_new(GwyXYZ, 1);
    xyz->x = x;
    xyz->y = y;
    xyz->z = z;
    return xyz;
}

/**
 * gwy_xyz_copy:
 * @xyz: Cartesian coordinates in space.
 *
 * Copies Cartesian coordinates in space.
 *
 * Returns: A copy of @xyz. The result should be freed using gwy_xyz_free().
 **/
GwyXYZ*
gwy_xyz_copy(const GwyXYZ *xyz)
{
    g_return_val_if_fail(xyz, NULL);
    return g_memdup2(xyz, sizeof(GwyXYZ));
}

/**
 * gwy_xyz_free:
 * @xyz: Cartesian coordinates in space.
 *
 * Frees Cartesian coordinates in space created with gwy_xyz_copy().
 **/
void
gwy_xyz_free(GwyXYZ *xyz)
{
    g_free(xyz);
}

static gsize
itemize(gpointer boxed, GwySerializableItems *items)
{
    g_return_val_if_fail(items->len - items->n >= NUM_ITEMS, 0);

    GwyXYZ *xyz = (GwyXYZ*)boxed;
    guint n = 0;
    gwy_serialize_append_double(xyz->x, serializable_items + ITEM_X, items, &n);
    gwy_serialize_append_double(xyz->y, serializable_items + ITEM_Y, items, &n);
    gwy_serialize_append_double(xyz->z, serializable_items + ITEM_Z, items, &n);
    return n;
}

static gpointer
construct(GwySerializableItems *items, GwyErrorList **error_list)
{
    GwySerializableItem its[NUM_ITEMS];
    gwy_assign(its, serializable_items, NUM_ITEMS);
    gwy_deserialize_filter_items(its, NUM_ITEMS, items, TYPE_NAME, error_list);

    GwyXYZ *xyz = g_new(GwyXYZ, 1);
    xyz->x = its[ITEM_X].value.v_double;
    xyz->y = its[ITEM_Y].value.v_double;
    xyz->z = its[ITEM_Z].value.v_double;
    return xyz;
}

/**
 * SECTION:gwyxyz
 * @title: GwyXYZ
 * @short_description: Cartesian coordinates in space
 **/

/**
 * GwyXYZ:
 * @x: X-coordinate.
 * @y: Y-coordinate.
 * @z: Z-coordinate.
 *
 * Representation of Cartesian coordinates in space.
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
