/*
 *  $Id: gwydataview.c 27380 2025-02-27 17:00:56Z yeti-dn $
 *  Copyright (C) 2003-2024 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
#define DEBUG 1
#include "config.h"
#include <glib-object.h>
#include <gtk/gtk.h>

#include "libgwyddion/gwymacros.h"
#include "libgwyddion/gwymath.h"
#include "libprocess/datafield.h"

#include "gwydgettypes.h"
#include "gwydataview.h"
#include "widget-impl-utils.h"

#define BITS_PER_SAMPLE 8

/* Rounding errors */
#define EPS 1e-6

enum {
    SGNL_REDRAWN,
    SGNL_RESIZED,
    SGNL_LAYER_PLUGGED,
    SGNL_LAYER_UNPLUGGED,
    NUM_SIGNALS
};

enum {
    PROP_0,
    PROP_ZOOM,
    PROP_DATA_PREFIX,
    NUM_PROPERTIES
};

struct _GwyDataViewPrivate {
    GwyContainer *data;
    GdkWindow *event_window;
    gchar *cursor_name;

    GQuark data_prefix;
    gulong square_hid;

    GwyDataViewLayer *base_layer;
    GwyDataViewLayer *alpha_layer;
    GwyDataViewLayer *top_layer;

    gulong base_hid;
    gulong alpha_hid;
    gulong top_hid;

    gboolean layers_changed;
    gboolean size_requested;

    gboolean resizable;
    gdouble zoom;    /* real zoom (larger number means larger pixmaps) */
    gdouble newzoom;    /* requested (ideal) zoom value */
    gdouble xmeasure;    /* physical units per pixel */
    gdouble ymeasure;    /* physical units per pixel */
    gint xoff;    /* x offset of the pixbuf from widget->allocation.x */
    gint yoff;    /* y offset of the pixbuf from widget->allocation.y */

    gdouble xoffset; /* offset of top left corner in data field coordinates */
    gdouble yoffset; /* offset of top left corner in data field coordinates */

    gboolean realsquare;
    gint xres;
    gint yres;
    gdouble xreal;
    gdouble yreal;
    gint scwidth;
    gint scheight;

    GdkPixbuf *base_pixbuf;
    GdkPixbuf *alpha_pixbuf;
};

static void     dispose             (GObject *object);
static void     set_property        (GObject *object,
                                     guint prop_id,
                                     const GValue *value,
                                     GParamSpec *pspec);
static void     get_property        (GObject *object,
                                     guint prop_id,
                                     GValue *value,
                                     GParamSpec *pspec);
static void     realize             (GtkWidget *widget);
static void     unrealize           (GtkWidget *widget);
static void     map                 (GtkWidget *widget);
static void     unmap               (GtkWidget *widget);
static void     get_preferred_width (GtkWidget *widget,
                                     gint *minimum,
                                     gint *natural);
static void     get_preferred_height(GtkWidget *widget,
                                     gint *minimum,
                                     gint *natural);
static void     size_allocate       (GtkWidget *widget,
                                     GdkRectangle *allocation);
static void     make_pixbufs        (GwyDataView *data_view);
static gboolean draw                (GtkWidget *widget,
                                     cairo_t *cr);
static void     set_layer           (GwyDataView *data_view,
                                     GwyDataViewLayer **which_layer,
                                     gulong *hid,
                                     GwyDataViewLayer *layer,
                                     GwyDataViewLayerType type);
static void     square_changed      (GwyDataView *data_view,
                                     GQuark quark);
static void     connect_data        (GwyDataView *data_view);
static void     set_cursor          (GwyDataView *data_view);

static guint signals[NUM_SIGNALS];
static GParamSpec *properties[NUM_PROPERTIES] = { NULL, };
static GtkWidgetClass *parent_class = NULL;

G_DEFINE_TYPE_WITH_CODE(GwyDataView, gwy_data_view, GTK_TYPE_WIDGET,
                        G_ADD_PRIVATE(GwyDataView))

static void
gwy_data_view_class_init(GwyDataViewClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);
    GType type = G_TYPE_FROM_CLASS(klass);

    parent_class = gwy_data_view_parent_class;

    gobject_class->dispose = dispose;
    gobject_class->set_property = set_property;
    gobject_class->get_property = get_property;

    widget_class->realize = realize;
    widget_class->unrealize = unrealize;
    widget_class->map = map;
    widget_class->unmap = unmap;
    widget_class->draw = draw;
    widget_class->get_preferred_width = get_preferred_width;
    widget_class->get_preferred_height = get_preferred_height;
    widget_class->size_allocate = size_allocate;

    /**
     * GwyDataView::redrawn:
     * @gwydataview: The #GwyDataView which received the signal.
     *
     * The ::redrawn signal is emitted when #GwyDataView redraws pixbufs after an update.  That is, when it's the
     * right time to get a new pixbuf from gwy_data_view_get_pixbuf().
     **/
    signals[SGNL_REDRAWN] = g_signal_new("redrawn", type,
                                         G_SIGNAL_RUN_FIRST,
                                         G_STRUCT_OFFSET(GwyDataViewClass, redrawn),
                                         NULL, NULL,
                                         g_cclosure_marshal_VOID__VOID,
                                         G_TYPE_NONE, 0);
    g_signal_set_va_marshaller(signals[SGNL_REDRAWN], type, g_cclosure_marshal_VOID__VOIDv);

    /**
     * GwyDataView::resized:
     * @gwydataview: The #GwyDataView which received the signal.
     *
     * The ::resized signal is emitted when #GwyDataView wants to be resized, that is when the dimensions of base data
     * field changes or square mode is changed. Its purpose is to subvert the normal resizing logic of #GwyDataWindow
     * (due to geometry hints, its size requests are generally ignored, so an explicit resize is needed).  You should
     * usually ignore it.
     **/
    signals[SGNL_RESIZED] = g_signal_new("resized", type,
                                         G_SIGNAL_RUN_FIRST,
                                         G_STRUCT_OFFSET(GwyDataViewClass, resized),
                                         NULL, NULL,
                                         g_cclosure_marshal_VOID__VOID,
                                         G_TYPE_NONE, 0);
    g_signal_set_va_marshaller(signals[SGNL_RESIZED], type, g_cclosure_marshal_VOID__VOIDv);

    /**
     * GwyDataView::layer-plugged:
     * @arg1: Which layer was plugged (a #GwyDataViewLayerType value).
     * @gwydataview: The #GwyDataView which received the signal.
     *
     * The ::layer-plugged signal is emitted when a layer is plugged into a #GwyDataView.
     *
     * When a layer replaces an existing layer, ::layer-unplugged is emitted once the old layer is unplugged, then
     * ::layer-plugged when the new is plugged.
     **/
    signals[SGNL_LAYER_PLUGGED] = g_signal_new("layer-plugged", type,
                                               G_SIGNAL_RUN_FIRST,
                                               G_STRUCT_OFFSET(GwyDataViewClass, layer_plugged),
                                               NULL, NULL,
                                               g_cclosure_marshal_VOID__ENUM,
                                               G_TYPE_NONE, 1, GWY_TYPE_DATA_VIEW_LAYER_TYPE);
    g_signal_set_va_marshaller(signals[SGNL_LAYER_PLUGGED], type, g_cclosure_marshal_VOID__ENUMv);

    /**
     * GwyDataView::layer-unplugged:
     * @arg1: Which layer was unplugged (a #GwyDataViewLayerType value).
     * @gwydataview: The #GwyDataView which received the signal.
     *
     * The ::layer-unplugged signal is emitted when a layer is unplugged from a #GwyDataView.
     *
     * When a layer replaces an existing layer, ::layer-unplugged is emitted once the old layer is unplugged, then
     * ::layer-plugged when the new is plugged.
     **/
    signals[SGNL_LAYER_UNPLUGGED] = g_signal_new("layer-unplugged", type,
                                                 G_SIGNAL_RUN_FIRST,
                                                 G_STRUCT_OFFSET(GwyDataViewClass, layer_unplugged),
                                                 NULL, NULL,
                                                 g_cclosure_marshal_VOID__ENUM,
                                                 G_TYPE_NONE, 1, GWY_TYPE_DATA_VIEW_LAYER_TYPE);
    g_signal_set_va_marshaller(signals[SGNL_LAYER_UNPLUGGED], type, g_cclosure_marshal_VOID__ENUMv);

    properties[PROP_ZOOM] = g_param_spec_double("zoom", "Zoom",
                                                "The size of a pixel on screen, in horizontal direction",
                                                1/16.0, 16.0, 1.0, GWY_GPARAM_RWE);

    /**
     * GwyDataView:data-prefix:
     *
     * The :data-prefix property is the container prefix the data displayed by this view lies under.
     *
     * Note it is only used for items used by #GwyDataView itself, layers have their own keys.
     **/
    properties[PROP_DATA_PREFIX] = g_param_spec_string("data-prefix", "Data prefix",
                                                       "Container data prefix", NULL, GWY_GPARAM_RWE);

    g_object_class_install_properties(gobject_class, NUM_PROPERTIES, properties);
}

static void
gwy_data_view_init(GwyDataView *data_view)
{
    GwyDataViewPrivate *priv;

    data_view->priv = priv = gwy_data_view_get_instance_private(data_view);
    priv->zoom = 1.0;
    priv->newzoom = 1.0;
    priv->resizable = TRUE;

    gtk_widget_set_can_focus(GTK_WIDGET(data_view), TRUE);
    gtk_widget_set_has_window(GTK_WIDGET(data_view), FALSE);
}

static void
set_property(GObject *object,
             guint prop_id,
             const GValue *value,
             GParamSpec *pspec)
{
    GwyDataView *data_view = GWY_DATA_VIEW(object);

    switch (prop_id) {
        case PROP_ZOOM:
        gwy_data_view_set_zoom(data_view, g_value_get_double(value));
        break;

        case PROP_DATA_PREFIX:
        gwy_data_view_set_data_prefix(data_view, g_value_get_string(value));
        break;

        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
get_property(GObject *object,
             guint prop_id,
             GValue *value,
             GParamSpec *pspec)
{
    GwyDataView *data_view = GWY_DATA_VIEW(object);

    switch (prop_id) {
        case PROP_ZOOM:
        g_value_set_double(value, gwy_data_view_get_zoom(data_view));
        break;

        case PROP_DATA_PREFIX:
        g_value_set_string(value, gwy_data_view_get_data_prefix(data_view));
        break;

        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
dispose(GObject *object)
{
    GwyDataView *data_view = GWY_DATA_VIEW(object);
    GwyDataViewPrivate *priv = data_view->priv;

    g_clear_signal_handler(&priv->square_hid, priv->data);
    set_layer(data_view, &priv->top_layer, NULL, NULL, GWY_DATA_VIEW_LAYER_TOP);
    set_layer(data_view, &priv->alpha_layer, &priv->alpha_hid, NULL, GWY_DATA_VIEW_LAYER_ALPHA);
    set_layer(data_view, &priv->base_layer, &priv->base_hid, NULL, GWY_DATA_VIEW_LAYER_BASE);

    g_clear_object(&priv->base_layer);
    g_clear_object(&priv->alpha_layer);
    g_clear_object(&priv->top_layer);
    g_clear_object(&priv->data);

    G_OBJECT_CLASS(parent_class)->dispose(object);
}

/**
 * gwy_data_view_new:
 * @data: A #GwyContainer containing the data to display.
 *
 * Creates a new data-displaying widget for @data.
 *
 * A newly created #GwyDataView doesn't display anything.  You have to add some layers to it, at least a base layer
 * with gwy_data_view_set_base_layer(), and possibly others with gwy_data_view_set_alpha_layer() and
 * gwy_data_view_set_top_layer().
 *
 * The top layer is special. It must be a vector layer and can receive mouse and keyboard events.
 *
 * The base layer it also special. It must be always present, and must not be transparent or vector.
 *
 * Returns: A newly created data view as a #GtkWidget.
 **/
GtkWidget*
gwy_data_view_new(GwyContainer *data)
{
    GtkWidget *data_view = gtk_widget_new(GWY_TYPE_DATA_VIEW, NULL);

    g_return_val_if_fail(GWY_IS_CONTAINER(data), data_view);
    g_object_ref(data);
    GWY_DATA_VIEW(data_view)->priv->data = data;

    return data_view;
}

static void
realize(GtkWidget *widget)
{
    GwyDataView *data_view = GWY_DATA_VIEW(widget);
    GwyDataViewPrivate *priv = data_view->priv;

    /* FIXME GTK3 widgets generally do not call parent's realize. Not sure why because it does a couple of things for
     * us like setting the widget realized. */
    parent_class->realize(widget);
    /* TODO GTK3 scroll and touch events */
    priv->event_window = gwy_create_widget_input_window(widget,
                                                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
                                                        | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK
                                                        | GDK_POINTER_MOTION_MASK);

    /* FIXME GTK3 do we still need these? Cursors should be updated in map(), not realize() anyway. Some of them
     * draw text, but that should be probably handled by the parent by giving them a PangoContext in the draw
     * function. (There are more instances of such calls here, all commented out.) */
#if 0
    if (priv->base_layer)
        gwy_data_view_layer_realize(GWY_DATA_VIEW_LAYER(priv->base_layer));
    if (priv->alpha_layer)
        gwy_data_view_layer_realize(GWY_DATA_VIEW_LAYER(priv->alpha_layer));
    if (priv->top_layer)
        gwy_data_view_layer_realize(GWY_DATA_VIEW_LAYER(priv->top_layer));
#endif

    make_pixbufs(data_view);
    set_cursor(data_view);
}

static void
unrealize(GtkWidget *widget)
{
    GwyDataView *data_view = GWY_DATA_VIEW(widget);
    GwyDataViewPrivate *priv = data_view->priv;

#if 0
    if (priv->base_layer)
        gwy_data_view_layer_unrealize(GWY_DATA_VIEW_LAYER(priv->base_layer));
    if (priv->alpha_layer)
        gwy_data_view_layer_unrealize(GWY_DATA_VIEW_LAYER(priv->alpha_layer));
    if (priv->top_layer)
        gwy_data_view_layer_unrealize(GWY_DATA_VIEW_LAYER(priv->top_layer));
#endif

    gwy_destroy_widget_input_window(widget, &priv->event_window);

    g_clear_object(&priv->base_pixbuf);

    parent_class->unrealize(widget);
}

static void
map(GtkWidget *widget)
{
    GwyDataViewPrivate *priv = GWY_DATA_VIEW(widget)->priv;
    parent_class->map(widget);
    gdk_window_show(priv->event_window);
}

static void
unmap(GtkWidget *widget)
{
    GwyDataViewPrivate *priv = GWY_DATA_VIEW(widget)->priv;
    gdk_window_hide(priv->event_window);
    parent_class->unmap(widget);
}

static void
get_preferred_width(GtkWidget *widget,
                    gint *minimum, gint *natural)
{
    GwyDataView *data_view = GWY_DATA_VIEW(widget);
    GwyDataViewPrivate *priv = data_view->priv;

    *minimum = *natural = 2;
    if (!priv->base_layer)
        return;

    const gchar *key = gwy_pixmap_layer_get_data_key(GWY_PIXMAP_LAYER(priv->base_layer));
    if (!key)
        return;

    if (priv->realsquare) {
        gdouble scale = fmax(priv->xres/priv->xreal, priv->yres/priv->yreal);
        scale *= priv->newzoom;
        *natural = GWY_ROUND(scale * priv->xreal);
    }
    else
        *natural = GWY_ROUND(priv->newzoom * priv->xres);

    if (!priv->resizable)
        *minimum = *natural;

    priv->size_requested = TRUE;
    gwy_debug("requesting width min: %d, nat: %d", *minimum, *natural);
}

static void
get_preferred_height(GtkWidget *widget,
                     gint *minimum, gint *natural)
{
    GwyDataView *data_view = GWY_DATA_VIEW(widget);
    GwyDataViewPrivate *priv = data_view->priv;

    *minimum = *natural = 2;
    if (!priv->base_layer)
        return;

    const gchar *key = gwy_pixmap_layer_get_data_key(GWY_PIXMAP_LAYER(priv->base_layer));
    if (!key)
        return;

    if (priv->realsquare) {
        gdouble scale = fmax(priv->xres/priv->xreal, priv->yres/priv->yreal);
        scale *= priv->newzoom;
        *natural = GWY_ROUND(scale * priv->yreal);
    }
    else {
        *natural = GWY_ROUND(priv->newzoom * priv->yres);
    }

    if (!priv->resizable)
        *minimum = *natural;

    priv->size_requested = TRUE;
    gwy_debug("requesting height min: %d, nat: %d", *minimum, *natural);
}

static void
size_allocate(GtkWidget *widget,
              GdkRectangle *allocation)
{
    GwyDataView *data_view = GWY_DATA_VIEW(widget);
    GwyDataViewPrivate *priv = data_view->priv;

    gwy_debug("allocating %d x %d", allocation->width, allocation->height);

    parent_class->size_allocate(widget, allocation);
    if (priv->event_window)
        gdk_window_move_resize(priv->event_window, allocation->x, allocation->y, allocation->width, allocation->height);

    make_pixbufs(data_view);
    /* Update ideal zoom after a `spontanoues' size-allocate when someone simply changed the size w/o asking us.  But
     * if we were queried first, be persistent and request the same zoom also next time */
    if (!priv->size_requested) {
        priv->newzoom = priv->zoom;
        g_object_notify_by_pspec(G_OBJECT(data_view), properties[PROP_ZOOM]);
    }
    priv->size_requested = FALSE;
}

static void
make_pixbufs(GwyDataView *data_view)
{
    GwyDataViewPrivate *priv = data_view->priv;

    g_clear_object(&priv->base_pixbuf);
    g_clear_object(&priv->alpha_pixbuf);
    if (!priv->xres || !priv->yres)
        return;

    /* Base layer is always present, however pixmap layers may return NULL if they do not have corresponding data
     * fields. If the layers do not think they need repaining this all is just some referencing and unreferencing
     * because above we do not actually free anything and here we get it back again. */
    priv->base_pixbuf = gwy_pixmap_layer_paint(GWY_PIXMAP_LAYER(priv->base_layer));
    g_object_ref(priv->base_pixbuf);

    if (priv->alpha_layer) {
        priv->alpha_pixbuf = gwy_pixmap_layer_paint(GWY_PIXMAP_LAYER(priv->alpha_layer));
        if (priv->alpha_pixbuf)
            g_object_ref(priv->alpha_pixbuf);
    }
}

static void
calculate_scaling(GwyDataView *data_view)
{
    GdkRectangle allocation;
    gtk_widget_get_allocation(GTK_WIDGET(data_view), &allocation);
    gdouble allocwidth = allocation.width, allocheight = allocation.height;

    GwyDataViewPrivate *priv = data_view->priv;
    if (priv->realsquare) {
        gdouble scale = fmax(priv->xres/priv->xreal, priv->yres/priv->yreal);
        priv->zoom = fmin(allocwidth/(scale*priv->xreal), allocheight/(scale*priv->yreal));
        scale *= priv->zoom;
        priv->scwidth = GWY_ROUND(scale * priv->xreal);
        priv->scheight = GWY_ROUND(scale * priv->yreal);
    }
    else {
        priv->zoom = fmin(allocwidth/priv->xres, allocheight/priv->yres);
        priv->scwidth = GWY_ROUND(priv->xres * priv->zoom);
        priv->scheight = GWY_ROUND(priv->yres * priv->zoom);
    }
    priv->scwidth = MAX(priv->scwidth, 1);
    priv->scheight = MAX(priv->scheight, 1);
    priv->xmeasure = priv->xreal/priv->scwidth;
    priv->ymeasure = priv->yreal/priv->scheight;
    priv->xoff = (allocwidth - priv->scwidth)/2;
    priv->yoff = (allocheight - priv->scheight)/2;
}

static gboolean
draw(GtkWidget *widget, cairo_t *cr)
{
    GwyDataView *data_view = GWY_DATA_VIEW(widget);
    GwyDataViewPrivate *priv = data_view->priv;
    gboolean emit_redrawn = FALSE;

    if (!priv->xres || !priv->yres) {
        cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
        cairo_paint(cr);
        return FALSE;
    }

    /* XXX GTK3 we are doing a bunch of stupid redraws every window switch event, even though nothing changes in the
     * data view. It is because there are some animations of stupid box shadow decorations which queue resize (sic)
     * and of course redraw. */
    {
        GdkRectangle rect;
        gdk_cairo_get_clip_rectangle(cr, &rect);
        g_printerr("DataView %p draw %d×%d at (%d,%d)\n", data_view, rect.width, rect.height, rect.x, rect.y);
    }

    /* This means we requested new size, but received no allocation -- because the new size was identical to the old
     * one.  BUT we had a reason for that.  Typically this happens after a rotation of realsquare-displayed data: the
     * new widget size is the same, but the the stretched dimension is now the other one and thus pixmap sizes have to
     * be recalculated. */
    if (priv->size_requested) {
        /* TODO: If we want to cache a Cairo pattern rendered at the correct size, we would do something here. */
        priv->size_requested = FALSE;
    }

    if (priv->layers_changed
        || (priv->base_layer && gwy_pixmap_layer_wants_repaint(GWY_PIXMAP_LAYER(priv->base_layer)))
        || (priv->alpha_layer && gwy_pixmap_layer_wants_repaint(GWY_PIXMAP_LAYER(priv->alpha_layer)))) {
        emit_redrawn = TRUE;
        priv->layers_changed = FALSE;
    }

    /* This is a no-op when we are not repainting anything. Run it uncoditionally. */
    make_pixbufs(data_view);
    calculate_scaling(data_view);

    cairo_save(cr);
    cairo_translate(cr, priv->xoff, priv->yoff);
    cairo_scale(cr, (gdouble)priv->scwidth/priv->xres, (gdouble)priv->scheight/priv->yres);
    gdk_cairo_set_source_pixbuf(cr, priv->base_pixbuf, 0, 0);
    cairo_pattern_set_filter(cairo_get_source(cr), CAIRO_FILTER_NEAREST);
    cairo_paint(cr);
    if (priv->alpha_pixbuf) {
        gdk_cairo_set_source_pixbuf(cr, priv->alpha_pixbuf, 0, 0);
        cairo_pattern_set_filter(cairo_get_source(cr), CAIRO_FILTER_NEAREST);
        cairo_paint(cr);
    }
    cairo_restore(cr);

    if (priv->top_layer)
        gwy_vector_layer_draw(GWY_VECTOR_LAYER(priv->top_layer), cr);

    if (emit_redrawn)
        g_signal_emit(data_view, signals[SGNL_REDRAWN], 0);

    return FALSE;
}

static void
update(GwyDataView *data_view)
{
    GwyDataViewPrivate *priv = data_view->priv;
    GwyDataField *data_field;
    const gchar *key;
    gint pxres, pyres;
    gboolean need_resize = FALSE;

    if (!priv->base_layer
        || !(key = gwy_pixmap_layer_get_data_key(GWY_PIXMAP_LAYER(priv->base_layer)))
        || !(data_field = gwy_container_get_object_by_name(priv->data, key)))
        return;

    priv->xres = gwy_data_field_get_xres(data_field);
    priv->yres = gwy_data_field_get_yres(data_field);
    priv->xreal = gwy_data_field_get_xreal(data_field);
    priv->yreal = gwy_data_field_get_yreal(data_field);
    priv->xoffset = gwy_data_field_get_xoffset(data_field);
    priv->yoffset = gwy_data_field_get_yoffset(data_field);

    GtkWidget *widget = GTK_WIDGET(data_view);
    if (!gtk_widget_get_realized(widget))
        return;

    if (priv->base_pixbuf) {
        pxres = gdk_pixbuf_get_width(priv->base_pixbuf);
        pyres = gdk_pixbuf_get_height(priv->base_pixbuf);
        gwy_debug("field: %dx%d, pixbuf: %dx%d", priv->xres, priv->yres, pxres, pyres);
        if (pxres != priv->xres || pyres != priv->yres)
            need_resize = TRUE;
    }

    calculate_scaling(data_view);
    if (need_resize) {
        gwy_debug("needs resize");
        gtk_widget_queue_resize(widget);
        g_signal_emit(widget, signals[SGNL_RESIZED], 0);
    }
    else
        gtk_widget_queue_draw(widget);
}

/**
 * gwy_data_view_get_base_layer:
 * @data_view: A data view.
 *
 * Returns the base layer this data view currently uses.
 *
 * A base layer should be always present.
 *
 * Returns: The currently used base layer.
 **/
GwyDataViewLayer*
gwy_data_view_get_base_layer(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), NULL);
    return data_view->priv->base_layer;
}

/**
 * gwy_data_view_get_alpha_layer:
 * @data_view: A data view.
 *
 * Returns the alpha layer this data view currently uses, or %NULL if none
 * is present.
 *
 * Returns: The currently used alpha layer.
 **/
GwyDataViewLayer*
gwy_data_view_get_alpha_layer(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), NULL);
    return data_view->priv->alpha_layer;
}

/**
 * gwy_data_view_get_top_layer:
 * @data_view: A data view.
 *
 * Returns the top layer this data view currently uses, or %NULL if none
 * is present.
 *
 * Returns: The currently used top layer.
 **/
GwyDataViewLayer*
gwy_data_view_get_top_layer(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), NULL);
    return data_view->priv->top_layer;
}

static void
set_layer(GwyDataView *data_view,
          GwyDataViewLayer **which_layer,
          gulong *hid,
          GwyDataViewLayer *layer,
          GwyDataViewLayerType type)
{
    GwyDataViewPrivate *priv = data_view->priv;
    GwyDataViewLayer *old_layer = *which_layer;

    if (layer == old_layer)
        return;
    if (old_layer) {
        if (hid) {
            g_signal_handler_disconnect(old_layer, *hid);
            *hid = 0;
        }
#if 0
        if (gtk_widget_get_realized(GTK_WIDGET(data_view)))
            gwy_data_view_layer_unrealize(GWY_DATA_VIEW_LAYER(*which_layer));
#endif
        /* XXX: Some things used to expect we send the signal first, actually unplug later. But we do not do that
         * any more. */
        gwy_data_view_layer_set_parent(old_layer, NULL);
        g_object_unref(old_layer);
        g_signal_emit(data_view, signals[SGNL_LAYER_UNPLUGGED], 0, type);
    }
    if (layer) {
        g_assert(!gwy_data_view_layer_get_parent(layer));
        /* FIXME: Too many refs here? */
        g_object_ref(layer);
        g_object_ref_sink(layer);
        if (hid)
            *hid = g_signal_connect_swapped(layer, "updated", G_CALLBACK(update), data_view);
        gwy_data_view_layer_set_parent(layer, GTK_WIDGET(data_view));
#if 0
        if (gtk_widget_get_realized(GTK_WIDGET(data_view)))
            gwy_data_view_layer_realize(GWY_DATA_VIEW_LAYER(layer));
#endif
    }
    priv->layers_changed = TRUE;
    *which_layer = layer;
    g_signal_emit(data_view, signals[SGNL_LAYER_PLUGGED], 0, type);
    update(data_view);
}

/**
 * gwy_data_view_set_base_layer:
 * @data_view: A data view.
 * @layer: A layer to be used as the base layer for @data_view.
 *
 * Plugs @layer to @data_view as the base layer.
 *
 * If another base layer is present, it's unplugged.
 *
 * The layer must not be a vector layer.  Theoretically, it can be %NULL to
 * use no base layer, but then @data_view will probably display garbage.
 **/
void
gwy_data_view_set_base_layer(GwyDataView *data_view,
                             GwyDataViewLayer *layer)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));
    g_return_if_fail(!layer || GWY_IS_PIXMAP_LAYER(layer));

    GwyDataViewPrivate *priv = data_view->priv;
    set_layer(data_view, &priv->base_layer, &priv->base_hid, GWY_DATA_VIEW_LAYER(layer), GWY_DATA_VIEW_LAYER_BASE);
    gtk_widget_queue_resize(GTK_WIDGET(data_view));
}

/**
 * gwy_data_view_set_alpha_layer:
 * @data_view: A data view.
 * @layer: A layer to be used as the alpha layer for @data_view.
 *
 * Plugs @layer to @data_view as the alpha layer.
 *
 * If another alpha layer is present, it's unplugged.
 *
 * The layer must not be a vector layer.  It can be %NULL, meaning no alpha
 * layer is to be used.
 **/
void
gwy_data_view_set_alpha_layer(GwyDataView *data_view,
                              GwyDataViewLayer *layer)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));
    g_return_if_fail(!layer || GWY_IS_PIXMAP_LAYER(layer));

    GwyDataViewPrivate *priv = data_view->priv;
    set_layer(data_view, &priv->alpha_layer, &priv->alpha_hid, GWY_DATA_VIEW_LAYER(layer), GWY_DATA_VIEW_LAYER_ALPHA);
}

/**
 * gwy_data_view_set_top_layer:
 * @data_view: A data view.
 * @layer: A layer to be used as the top layer for @data_view.
 *
 * Plugs @layer to @data_view as the top layer.
 *
 * If another top layer is present, it's unplugged.
 *
 * The layer must be a vector layer.  It can be %NULL, meaning no top
 * layer is to be used.
 **/
void
gwy_data_view_set_top_layer(GwyDataView *data_view,
                            GwyDataViewLayer *layer)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));
    g_return_if_fail(!layer || GWY_IS_VECTOR_LAYER(layer));

    GwyDataViewPrivate *priv = data_view->priv;
    set_layer(data_view, &priv->top_layer, &priv->top_hid, GWY_DATA_VIEW_LAYER(layer), GWY_DATA_VIEW_LAYER_TOP);
}

/**
 * gwy_data_view_get_hexcess:
 * @data_view: A data view.
 *
 * Return the horizontal excess of widget size to data size.
 *
 * Do not use.  Only useful for #GwyDataWindow implementation.
 *
 * Returns: The execess.
 **/
gdouble
gwy_data_view_get_hexcess(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), 0);
    return (gdouble)gtk_widget_get_allocated_width(GTK_WIDGET(data_view))/data_view->priv->scwidth - 1.0;
}

/**
 * gwy_data_view_get_vexcess:
 * @data_view: A data view.
 *
 * Return the vertical excess of widget size to data size.
 *
 * Do not use.  Only useful for #GwyDataWindow implementation.
 *
 * Returns: The execess.
 **/
gdouble
gwy_data_view_get_vexcess(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), 0);
    return (gdouble)gtk_widget_get_allocated_height(GTK_WIDGET(data_view))/data_view->priv->scheight - 1.0;
}

/**
 * gwy_data_view_set_zoom:
 * @data_view: A data view.
 * @zoom: A new zoom value.
 *
 * Sets zoom of @data_view to @zoom.
 *
 * Zoom greater than 1 means larger image on screen and vice versa.
 *
 * Note window manager can prevent the window from resize and thus the zoom from change.
 **/
void
gwy_data_view_set_zoom(GwyDataView *data_view,
                       gdouble zoom)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));

    GwyDataViewPrivate *priv = data_view->priv;
    gwy_debug("zoom = %g, new = %g", priv->newzoom, zoom);
    if (fabs(log(priv->newzoom/zoom)) < 0.001)
        return;

    priv->newzoom = zoom;
    g_object_notify_by_pspec(G_OBJECT(data_view), properties[PROP_ZOOM]);
    gtk_widget_queue_resize(GTK_WIDGET(data_view));
}

/**
 * gwy_data_view_get_zoom:
 * @data_view: A data view.
 *
 * Returns current ideal zoom of a data view.
 *
 * More precisely the zoom value requested by gwy_data_view_set_zoom(), if it's in use (real zoom may differ a bit due
 * to pixel rounding).  If zoom was set by explicite widget size change, real and requested zoom are considered to be
 * the same.
 *
 * When a resize is queued, the new zoom value is returned.
 *
 * In other words, this is the zoom @data_view would like to have.  Use gwy_data_view_get_real_zoom() to get the real
 * zoom.
 *
 * Returns: The zoom as a ratio between ideal displayed size and base data field size.
 **/
gdouble
gwy_data_view_get_zoom(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), 1.0);
    return data_view->priv->newzoom;
}

/**
 * gwy_data_view_get_real_zoom:
 * @data_view: A data view.
 *
 * Returns current real zoom of a data view.
 *
 * This is the zoom value a data view may not wish to have, but was imposed by window manager or other constraints.
 * Unlike ideal zoom set by gwy_data_view_set_zoom(), this value cannot be set.
 *
 * When a resize is queued, the current (old) value is returned.
 *
 * Returns: The zoom as a ratio between real displayed size and base data field size.
 **/
gdouble
gwy_data_view_get_real_zoom(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), 1.0);
    return data_view->priv->zoom;
}

/**
 * gwy_data_view_get_xmeasure:
 * @data_view: A data view.
 *
 * Returns the ratio between horizontal physical lengths and horizontal screen lengths in pixels.
 *
 * Returns: The horizontal measure.
 **/
gdouble
gwy_data_view_get_xmeasure(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), 1.0);
    return data_view->priv->xmeasure;
}

/**
 * gwy_data_view_get_ymeasure:
 * @data_view: A data view.
 *
 * Returns the ratio between vertical physical lengths and horizontal screen lengths in pixels.
 *
 * Returns: The vertical measure.
 **/
gdouble
gwy_data_view_get_ymeasure(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), 1.0);
    return data_view->priv->ymeasure;
}

/**
 * gwy_data_view_get_data:
 * @data_view: A data view.
 *
 * Returns the data container used by @data_view.
 *
 * Returns: The data as a #GwyContainer.
 **/
GwyContainer*
gwy_data_view_get_data(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), NULL);
    return data_view->priv->data;
}

/**
 * gwy_data_view_coords_widget_clamp:
 * @data_view: A data view.
 * @xscr: A screen x-coordinate relative to widget origin.
 * @yscr: A screen y-coordinate relative to widget origin.
 *
 * Fixes screen coordinates @xscr and @yscr to be inside the data-displaying area (which can be smaller than widget
 * size).
 *
 * Returns: %TRUE if coordinates are inside the widget area.
 **/
gboolean
gwy_data_view_coords_widget_clamp(GwyDataView *data_view,
                                  gdouble *xscr, gdouble *yscr)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), FALSE);

    GwyDataViewPrivate *priv = data_view->priv;
    gint size;
    gdouble t;
    gboolean retval = TRUE;

    if (xscr) {
        size = priv->scwidth;
        t = fmin(fmax(*xscr, priv->xoff), priv->xoff + size-1);
        if (t != *xscr) {
            *xscr = t;
            retval = FALSE;
        }
    }
    if (yscr) {
        size = priv->scheight;
        t = fmin(fmax(*yscr, priv->yoff), priv->yoff + size-1);
        if (t != *yscr) {
            *yscr = t;
            retval = FALSE;
        }
    }

    return retval;
}

/* FIXME: There is a gwymath function now. It either does the same, or it should be corrected to do the same. */
/**
 * gwy_data_view_coords_widget_cut_line:
 * @data_view: A data view.
 * @x0scr: First point screen x-coordinate relative to widget origin.
 * @y0scr: First point screen y-coordinate relative to widget origin.
 * @x1scr: Second point screen x-coordinate relative to widget origin.
 * @y1scr: Second point screen y-coordinate relative to widget origin.
 *
 * Fixes screen coordinates of line endpoints to be inside the data-displaying area (which can be smaller than widget
 * size).
 *
 * Returns: %TRUE if line was entirely inside the area.
 **/
gboolean
gwy_data_view_coords_widget_cut_line(GwyDataView *data_view,
                                     gdouble *x0scr, gdouble *y0scr,
                                     gdouble *x1scr, gdouble *y1scr)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), FALSE);

    GwyDataViewPrivate *priv = data_view->priv;
    gdouble xsize, ysize, x0s, y0s, x1s, y1s;

    xsize = priv->scwidth;
    x0s = CLAMP(*x0scr, priv->xoff, priv->xoff + xsize-1);
    x1s = CLAMP(*x1scr, priv->xoff, priv->xoff + xsize-1);
    ysize = priv->scheight;
    y0s = CLAMP(*y0scr, priv->yoff, priv->yoff + ysize-1);
    y1s = CLAMP(*y1scr, priv->yoff, priv->yoff + ysize-1);

    /* All inside */
    if (x0s == *x0scr && y0s == *y0scr && x1s == *x1scr && y1s == *y1scr)
        return TRUE;

    /* Horizontal/vertical lines */
    if (*x1scr == *x0scr) {
        if (*x0scr != x0s)
            goto fail;
        *y0scr = y0s;
        *y1scr = y1s;
        return FALSE;
    }
    if (*y1scr == *y0scr) {
        if (*y0scr != y0s)
            goto fail;
        *x0scr = x0s;
        *x1scr = x1s;
        return FALSE;
    }

    /* The hard case */
    x0s = *x0scr;
    x1s = *x1scr;
    y0s = *y0scr;
    y1s = *y1scr;

    gdouble t[6];
    t[0] = -(x0s)/(gdouble)(x1s - x0s);
    t[1] = (xsize - 1 -(x0s))/(gdouble)(x1s - x0s);
    t[2] = -(y0s)/(gdouble)(y1s - y0s);
    t[3] = (ysize - 1 -(y0s))/(gdouble)(y1s - y0s);
    /* Include the endpoints */
    t[4] = 0.0;
    t[5] = 1.0;

    gwy_math_sort(G_N_ELEMENTS(t), t);

    gint i0 = -1, i1 = -1;
    for (gint i = 0; i < G_N_ELEMENTS(t); i++) {
        gdouble xy;

        if (t[i] < -EPS || t[i] > 1.0 + EPS)
            continue;

        xy = x0s + t[i]*(x1s - x0s);
        if (xy < -EPS || xy > xsize-1 + EPS)
            continue;

        xy = y0s + t[i]*(y1s - y0s);
        if (xy < -EPS || xy > ysize-1 + EPS)
            continue;

        /* i0 is the first index, once i0 != -1, do not change it any more */
        if (i0 == -1)
            i0 = i;

        /* i1 is the last index, move it as long as we are inside */
        i1 = i;
    }

    if (i0 == -1)
        goto fail;

    *x0scr = x0s + t[i0]*(x1s - x0s);
    *x1scr = x0s + t[i1]*(x1s - x0s);
    *y0scr = y0s + t[i0]*(y1s - y0s);
    *y1scr = y0s + t[i1]*(y1s - y0s);
    return FALSE;

fail:
    /* The line does not intersect the boundary at all.  Just return something and pray... */
    *x0scr = *x1scr = xsize/2;
    *y0scr = *y1scr = ysize/2;
    return FALSE;
}

/**
 * gwy_data_view_coords_widget_to_real:
 * @data_view: A data view.
 * @xscr: A screen x-coordinate relative to widget origin.
 * @yscr: A screen y-coordinate relative to widget origin.
 * @xreal: Where the physical x-coordinate in the data sample should be stored.
 * @yreal: Where the physical y-coordinate in the data sample should be stored.
 *
 * Recomputes screen coordinates relative to widget origin to physical coordinates in the sample.
 *
 * Note that data fields offsets are <emphasis>not</emphasis> taken into account.  Coordinates @xreal, @yreal are just
 * relative coordinates to the top left field corner.
 **/
void
gwy_data_view_coords_widget_to_real(GwyDataView *data_view,
                                    gdouble xscr, gdouble yscr,
                                    gdouble *xreal, gdouble *yreal)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));

    GwyDataViewPrivate *priv = data_view->priv;
    if (xreal)
        *xreal = (xscr + 0.5 - priv->xoff) * priv->xmeasure;
    if (yreal)
        *yreal = (yscr + 0.5 - priv->yoff) * priv->ymeasure;
}

/**
 * gwy_data_view_coords_real_to_widget:
 * @data_view: A data view.
 * @xreal: A physical x-coordinate in the data sample.
 * @yreal: A physical y-coordinate in the data sample.
 * @xscr: Where the screen x-coordinate relative to widget origin should be stored.
 * @yscr: Where the screen y-coordinate relative to widget origin should be stored.
 *
 * Recomputes physical coordinate in the sample to screen coordinate relative to widget origin, keeping them as
 * floating point values.
 *
 * Note that data fields offsets are <emphasis>not</emphasis> taken into account.  Coordinates @xreal, @yreal are just
 * relative coordinates to the top left field corner.
 **/
void
gwy_data_view_coords_real_to_widget(GwyDataView *data_view,
                                    gdouble xreal, gdouble yreal,
                                    gdouble *xscr, gdouble *yscr)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));

    GwyDataViewPrivate *priv = data_view->priv;
    if (xscr)
        *xscr = xreal/priv->xmeasure + priv->xoff;
    if (yscr)
        *yscr = yreal/priv->ymeasure + priv->yoff;
}

/**
 * gwy_data_view_get_xzoom:
 * @data_view: A data view.
 *
 * Gets the ratio between screen and image horizontal coordinates in a data view.
 *
 * Returns: The ratio between horizontal image and screen pixel sizes.
 **/
gdouble
gwy_data_view_get_xzoom(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), 1.0);
    GwyDataViewPrivate *priv = data_view->priv;
    /* xmeasure = xreal/scrsize, so the entire fraction reduces to scrsize/xres */
    return priv->xreal/priv->xres/priv->xmeasure;
}

/**
 * gwy_data_view_get_yzoom:
 * @data_view: A data view.
 *
 * Gets the ratio between screen and image vertical coordinates in a data view.
 *
 * Returns: The ratio between vertical image and screen pixel sizes.
 **/
gdouble
gwy_data_view_get_yzoom(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), 1.0);
    GwyDataViewPrivate *priv = data_view->priv;
    /* xmeasure = xreal/scrsize, so the entire fraction reduces to scrsize/xres */
    return priv->yreal/priv->yres/priv->ymeasure;
}

/**
 * gwy_data_view_set_resizable:
 * @data_view: A data view.
 * @resizable: %TRUE to make the data view user-resizable; %FALSE to request fixed size according to the zoom.
 *
 * Sets the resizability of a data view.
 *
 * By default, data views adapt to any size they get allocated and request very small minimum size. This is useful in
 * data windows which should be freely resizable, for instance.
 *
 * Module dialogs usually have small fixed-size previews. The data view never grows, but could be still shrinked. In
 * such case @resizable should be set to %FALSE to prevent the shrinking. The data view will then request minimum size
 * according to zoom.
 **/
void
gwy_data_view_set_resizable(GwyDataView *data_view,
                            gboolean resizable)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));
    GwyDataViewPrivate *priv = data_view->priv;
    if (!resizable == !priv->resizable)
        return;

    priv->resizable = !!resizable;
    gtk_widget_queue_resize(GTK_WIDGET(data_view));
}

/**
 * gwy_data_view_get_pixel_data_sizes:
 * @data_view: A data view.
 * @xres: Location to store x-resolution of displayed data (or %NULL).
 * @yres: Location to store y-resolution of displayed data (or %NULL).
 *
 * Obtains pixel dimensions of data displayed by a data view.
 *
 * This is a convenience method, the same values could be obtained by gwy_data_field_get_xres() and
 * gwy_data_field_get_yres() of the data field displayed by the base layer.
 **/
void
gwy_data_view_get_pixel_data_sizes(GwyDataView *data_view,
                                   gint *xres,
                                   gint *yres)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));

    GwyDataViewPrivate *priv = data_view->priv;
    if (xres)
        *xres = priv->xres;
    if (yres)
        *yres = priv->yres;
}

/**
 * gwy_data_view_get_real_data_sizes:
 * @data_view: A data view.
 * @xreal: Location to store physical x-dimension of the displayed data without excess (or %NULL).
 * @yreal: Location to store physical y-dimension of the displayed data without excess (or %NULL).
 *
 * Obtains physical dimensions of data displayed by a data view.
 *
 * Physical coordinates are always taken from data field displayed by the base layer.  This is a convenience method,
 * the same values could be obtained by gwy_data_field_get_xreal() and gwy_data_field_get_yreal() of the data field
 * displayed by the base layer.
 **/
void
gwy_data_view_get_real_data_sizes(GwyDataView *data_view,
                                  gdouble *xreal,
                                  gdouble *yreal)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));

    GwyDataViewPrivate *priv = data_view->priv;
    if (xreal)
        *xreal = priv->xreal;
    if (yreal)
        *yreal = priv->yreal;
}

/**
 * gwy_data_view_get_metric:
 * @data_view: A data view.
 * @metric: Metric matrix 2x2 (stored in sequentially by rows: m11, m12, m12, m22).
 *
 * Fills metric matrix for a data view.
 *
 * The metric matrix essentially transforms distances in physical coordinates to screen distances.  It is to be used
 * with functions like gwy_math_find_nearest_point() and gwy_math_find_nearest_line() when the distance should be
 * screen-Euclidean.
 **/
void
gwy_data_view_get_metric(GwyDataView *data_view,
                         gdouble *metric)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));
    g_return_if_fail(metric);

    GwyDataViewPrivate *priv = data_view->priv;
    metric[0] = 1.0/(priv->xmeasure*priv->xmeasure);
    metric[1] = metric[2] = 0.0;
    metric[3] = 1.0/(priv->ymeasure*priv->ymeasure);
}

/**
 * gwy_data_view_get_real_data_offsets:
 * @data_view: A data view.
 * @xoffset: Location to store physical x-offset of the top corner of
 *           displayed data without excess (or %NULL).
 * @yoffset: Location to store physical y-offset of the top corner of
 *           displayed data without excess (or %NULL).
 *
 * Obtains physical offsets of data displayed by a data view.
 *
 * Physical coordinates are always taken from data field displayed by the base layer.  This is a convenience method,
 * the same values could be obtained by gwy_data_field_get_xoffset() and gwy_data_field_get_yoffset() of the data
 * field displayed by the base layer.
 **/
void
gwy_data_view_get_real_data_offsets(GwyDataView *data_view,
                                    gdouble *xoffset,
                                    gdouble *yoffset)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));

    GwyDataViewPrivate *priv = data_view->priv;
    if (xoffset)
        *xoffset = priv->xoffset;
    if (yoffset)
        *yoffset = priv->yoffset;
}

/**
 * gwy_data_view_get_pixbuf:
 * @data_view: A data view.
 * @max_width: Pixbuf width that should not be exceeeded.  Value smaller than 1 means unlimited size.
 * @max_height: Pixbuf height that should not be exceeeded.  Value smaller than 1 means unlimited size.
 *
 * Creates and returns a pixbuf from the data view.
 *
 * If the data is not square, the resulting pixbuf is also nonsquare. The returned pixbuf also never has an alpha
 * channel.
 *
 * Returns: The pixbuf as a newly created #GdkPixbuf, it should be freed when no longer needed.  It is never larger
 *          than the actual data size, as @max_width and @max_height are only upper limits.
 **/
GdkPixbuf*
gwy_data_view_get_pixbuf(GwyDataView *data_view,
                         gint max_width,
                         gint max_height)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), NULL);
    GwyDataViewPrivate *priv = data_view->priv;
    g_return_val_if_fail(priv->base_pixbuf, NULL);

    gint width = gdk_pixbuf_get_width(priv->base_pixbuf);
    gint height = gdk_pixbuf_get_height(priv->base_pixbuf);
    gdouble xscale = (max_width > 0) ? (gdouble)max_width/width : 1.0;
    gdouble yscale = (max_height > 0) ? (gdouble)max_height/height : 1.0;
    gdouble scale = fmin(fmin(xscale, yscale), 1.0);
    gint width_scaled = (gint)(scale*width);
    gint height_scaled = (gint)(scale*height);
    if (max_width)
        width_scaled = CLAMP(width_scaled, 1, max_width);
    if (max_height)
        height_scaled = CLAMP(height_scaled, 1, max_height);

    if (!priv->alpha_pixbuf) {
        GdkPixbuf *pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, BITS_PER_SAMPLE, width_scaled, height_scaled);
        gdk_pixbuf_scale(priv->base_pixbuf, pixbuf, 0, 0, width_scaled, height_scaled, 0.0, 0.0,
                         scale, scale, GDK_INTERP_TILES);
        return pixbuf;
    }

    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, width_scaled, height_scaled);
    cairo_t *cr = cairo_create(surface);
    cairo_scale(cr, (gdouble)width_scaled/width, (gdouble)height_scaled/height);
    gdk_cairo_set_source_pixbuf(cr, priv->base_pixbuf, 0, 0);
    cairo_paint(cr);
    gdk_cairo_set_source_pixbuf(cr, priv->alpha_pixbuf, 0, 0);
    cairo_paint(cr);
    cairo_destroy(cr);

    GdkPixbuf *pixbuf = gdk_pixbuf_get_from_surface(surface, 0, 0, width_scaled, height_scaled);
    cairo_surface_destroy(surface);

    return pixbuf;
}

/**
 * gwy_data_view_set_data_prefix:
 * @data_view: A data view.
 * @prefix: Container prefix for data  (eg. "/0/data").
 *
 * Sets the prefix for the container data channel to display in a data view.
 *
 * This function only affects where the data view itself takes parameters from, it does not affect layer keys.
 **/
void
gwy_data_view_set_data_prefix(GwyDataView *data_view,
                              const gchar *prefix)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));

    GwyDataViewPrivate *priv = data_view->priv;
    GQuark quark = prefix ? g_quark_from_string(prefix) : 0;
    if (quark == priv->data_prefix)
        return;

    g_clear_signal_handler(&priv->square_hid, priv->data);
    priv->data_prefix = quark;
    if (quark)
        connect_data(data_view);

    g_object_notify_by_pspec(G_OBJECT(data_view), properties[PROP_DATA_PREFIX]);
}

/**
 * gwy_data_view_get_data_prefix:
 * @data_view: A data view.
 *
 * Gets the prefix for the container data channel that the data view is currently set to display.
 *
 * Returns: The container data prefix (eg. "/0/data").
 **/
const gchar*
gwy_data_view_get_data_prefix(GwyDataView *data_view)
{
    g_return_val_if_fail(GWY_IS_DATA_VIEW(data_view), NULL);

    return g_quark_to_string(data_view->priv->data_prefix);
}

/**
 * gwy_data_view_set_named_cursor:
 * @data_view: A data view.
 * @name: A cursor name, or %NULL.
 *
 * Sets the pointer cursor for a data view to a named cursor.
 *
 * The cursor name is the same as one would pass to gdk_cursor_new_from_name().
 *
 * The data view needs to be realised.
 *
 * This function is intended for the implementation of editable data view layers. If an application sets the cursor,
 * a data view layer will likely change it again soon.
 **/
void
gwy_data_view_set_named_cursor(GwyDataView *data_view,
                               const gchar *name)
{
    g_return_if_fail(GWY_IS_DATA_VIEW(data_view));

    GwyDataViewPrivate *priv = data_view->priv;
    /* This may be called very often, so just compare strings and quit if there is nothing to change. */
    if (!gwy_assign_string(&priv->cursor_name, name))
        return;
    if (priv->event_window)
        set_cursor(data_view);
}

static void
square_changed(GwyDataView *data_view,
               GQuark quark)
{
    GwyDataViewPrivate *priv = data_view->priv;
    gboolean realsquare = priv->realsquare;

    if (!gwy_container_gis_boolean(priv->data, quark, &realsquare))
        realsquare = FALSE;
    if (realsquare != priv->realsquare) {
        GtkWidget *widget = GTK_WIDGET(data_view);

        priv->realsquare = realsquare;
        if (gtk_widget_get_realized(GTK_WIDGET(widget))) {
            gtk_widget_queue_resize(widget);
            g_signal_emit(widget, signals[SGNL_RESIZED], 0);
        }
    }
}

static void
connect_data(GwyDataView *data_view)
{
    static const gchar prefix[] = "item-changed::";
    guint pfxlen = strlen(prefix);
    GwyDataViewPrivate *priv = data_view->priv;

    g_return_if_fail(priv->data);
    g_return_if_fail(priv->square_hid == 0);
    if (!priv->data_prefix)
        return;

    gchar *s = g_strconcat(prefix, g_quark_to_string(priv->data_prefix), "/realsquare", NULL);
    priv->square_hid = g_signal_connect_swapped(priv->data, s, G_CALLBACK(square_changed), data_view);
    square_changed(data_view, g_quark_from_string(s + pfxlen));
    g_free(s);
}

static void
set_cursor(GwyDataView *data_view)
{
    GwyDataViewPrivate *priv = data_view->priv;
    const gchar *name = priv->cursor_name;
    GdkDisplay *display = gtk_widget_get_display(GTK_WIDGET(data_view));

    g_return_if_fail(display);
    GdkCursor *cursor = name ? gdk_cursor_new_from_name(display, name) : NULL;
    gdk_window_set_cursor(priv->event_window, cursor);
    g_clear_object(&cursor);
}

/************************** Documentation ****************************/

/**
 * SECTION:gwydataview
 * @title: GwyDataView
 * @short_description: Data field displaying area
 * @see_also: #GwyDataWindow -- window combining data view with other controls,
 *            #GwyDataViewLayer -- layers a data view is composed of,
 *            <link linkend="libgwydraw-gwypixfield">gwypixfield</link> --
 *            low level functions for painting data fields,
 *            #Gwy3DView -- OpenGL 3D data display
 *
 * #GwyDataView is a basic two-dimensional data display widget.  The actual rendering is performed by one or more
 * #GwyDataViewLayer's, pluggable into the data view.  Each layer generally displays different data field from the
 * container supplied to gwy_data_view_new().
 *
 * A base layer (set with gwy_data_view_set_base_layer()) must be always present, and normally it is always
 * a #GwyLayerBasic.
 *
 * Other layers are optional.  Middle, or alpha, layer (set with gwy_data_view_set_alpha_layer()) used to display
 * masks is normally always a #GwyLayerMask.  Top layer, if present, is a #GwyVectorLayer allowing to draw selections
 * with mouse and otherwise interace with the view, it is set with gwy_data_view_set_top_layer().
 *
 * The size of a data view is affected by two factors: zoom and outer constraints. If an explicit size set by window
 * manager or by Gtk+ means, the view scales the displayed data to fit into this size (while keeping x/y ratio). Zoom
 * controlls the size a data view requests, and can be set with gwy_data_view_set_zoom().
 *
 * Several helper functions are available for transformation between screen coordinates in the view and physical
 * coordinates in the displayed data field: gwy_data_view_coords_widget_to_real(), gwy_data_view_get_xmeasure(),
 * gwy_data_view_get_hexcess(), and others. Physical coordinates are always taken from data field displayed by base
 * layer.
 **/

/**
 * GwyDataViewLayerType:
 * @GWY_DATA_VIEW_LAYER_BASE: Base (bottom) layer.
 * @GWY_DATA_VIEW_LAYER_ALPHA: Alpha (mask) layer.
 * @GWY_DATA_VIEW_LAYER_TOP: Top (vector, interactive) layer.
 *
 * #GwyDataView layer identification.
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
