/*
 *  $Id: gwydgets.c 27111 2025-01-02 13:51:15Z yeti-dn $
 *  Copyright (C) 2003-2024 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"

#include "libdraw/gwydraw.h"
#include "gwydgets.h"

/************************** Initialization ****************************/

/**
 * gwy_widgets_type_init:
 *
 * Makes libgwydgets types safe for deserialization and performs other initialization.  You have to call this function
 * before using widgets and objects from libgwydgets.
 *
 * Calls gwy_draw_type_init() first to make sure libgwydraw is initialized.
 *
 * It is safe to call this function more than once, subsequent calls are no-op.
 **/
void
gwy_widgets_type_init(void)
{
    static guint types_initialized = FALSE;

    if (types_initialized)
        return;

    gwy_draw_type_init();

    g_type_class_peek(GWY_TYPE_GRAPH_CURVE_MODEL);
    g_type_class_peek(GWY_TYPE_GRAPH_MODEL);
    g_type_class_peek(GWY_TYPE_GL_LABEL);
    g_type_class_peek(GWY_TYPE_GL_SETUP);

    g_type_class_peek(GWY_TYPE_SELECTION_AXIS);
    g_type_class_peek(GWY_TYPE_SELECTION_ELLIPSE);
    g_type_class_peek(GWY_TYPE_SELECTION_LATTICE);
    g_type_class_peek(GWY_TYPE_SELECTION_LINE);
    g_type_class_peek(GWY_TYPE_SELECTION_PATH);
    g_type_class_peek(GWY_TYPE_SELECTION_POINT);
    g_type_class_peek(GWY_TYPE_SELECTION_QUAD);
    g_type_class_peek(GWY_TYPE_SELECTION_RANGE);
    g_type_class_peek(GWY_TYPE_SELECTION_RECTANGLE);

    types_initialized = 1;

    GtkIconTheme *theme = gtk_icon_theme_get_default();

    gchar *path = gwy_find_self_dir("icons");
    // XXX: They say gtk_icon_theme_add_resource_path() should be used for application-specific icons. But that
    // requires compiling the icons to a resource file, loading it into a GResource and registering said resource.
    // Since we are doing none of that, just prepend our icons to the search path.
    //gtk_icon_theme_add_resource_path(theme, path);
    gtk_icon_theme_prepend_search_path(theme, path);
    g_free(path);

    path = g_build_filename(gwy_get_user_dir(), "icons", NULL);
    if (g_file_test(path, G_FILE_TEST_IS_DIR))
        gtk_icon_theme_add_resource_path(theme, path);
    g_free(path);

    /* FIXME GTK3 */
#if 0
    gtk_rc_parse_string
        (/* graph window statusbar */
         "style \"gwyflatstatusbar\" {\n"
         "  GtkStatusbar::shadow_type = 0\n"
         "}\n"
         "widget \"*.gwyflatstatusbar\" style \"gwyflatstatusbar\"\n"
         "\n"
         /* adjust bar internal check button */
         "style \"gwyadjbarcheck\" {\n"
         "  GtkCheckButton::focus_padding = 0\n"
         "  GtkCheckButton::focus_line_width = 0\n"
         "}\n"
         "widget \"*.gwyadjbarcheck\" style \"gwyadjbarcheck\"\n"
         "\n");
#endif
}

/**
 * gwy_widgets_gl_init:
 *
 * Configures an OpenGL-capable visual for 3D widgets.
 *
 * This function does not do anything any more.
 *
 * Returns: %TRUE if an appropriate visual was found.  If Gwyddion was compiled without OpenGL support, it always
 *          returns %FALSE.
 **/
gboolean
gwy_widgets_gl_init(void)
{
    /* TODO GTK3 Now we can create GdkGLContext for particular windows using gdk_window_create_gl_context().
     * There does not seem to be any global check, we must see if we succeed for any particular window.
     * See
     *
     * https://stackoverflow.com/questions/69201843/how-do-you-do-3d-rendering-in-gtk-using-opengl
     *
     * for a complete a GTK3 OpenGL example (including a shader). */
    return FALSE;
}

/************************** Documentation ****************************/

/**
 * SECTION:gwydgets
 * @title: gwydgets
 * @short_description: Base functions
 *
 * Gwyddion classes has to be initialized before they can be safely deserialized. The function gwy_type_init()
 * performs this initialization.
 *
 * Before 3D widgets (#Gwy3DView) can be used, OpenGL must be initialized with gwy_widgets_gl_init().
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
