/*
 *  $Id: gwyhruler.c 27110 2025-01-02 11:02:46Z yeti-dn $
 *  Copyright (C) 2003-2024 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>

#include "libgwyddion/gwymacros.h"
#include "libgwyddion/gwymath.h"

#include "cairo-utils.h"
#include "gwyhruler.h"

#define LINE_WIDTH 1
#define RULER_WIDTH 20

struct _GwyHRulerPrivate {
    gint min_width;
    gint min_height;
};

static void     get_preferred_width (GtkWidget *widget,
                                     gint *minimum,
                                     gint *natural);
static void     get_preferred_height(GtkWidget *widget,
                                     gint *minimum,
                                     gint *natural);
static gboolean motion_notify       (GtkWidget *widget,
                                     GdkEventMotion *event);
static void     prepare_sizes       (GwyRuler *ruler);
static void     draw_layout         (GwyRuler *ruler,
                                     cairo_t *cr,
                                     gint hpos,
                                     gint vpos);
static void     draw_tick           (GwyRuler *ruler,
                                     cairo_t *cr,
                                     gdouble pos,
                                     gdouble length);
static void     draw_pos            (GwyRuler *ruler,
                                     cairo_t *cr);
static void     queue_marker_redraw (GwyRuler *ruler,
                                     gdouble value);

G_DEFINE_TYPE_WITH_CODE(GwyHRuler, gwy_hruler, GWY_TYPE_RULER,
                        G_ADD_PRIVATE(GwyHRuler))

static void
gwy_hruler_class_init(GwyHRulerClass *klass)
{
    GtkWidgetClass *widget_class = (GtkWidgetClass*) klass;
    GwyRulerClass *ruler_class = (GwyRulerClass*) klass;

    widget_class->motion_notify_event = motion_notify;
    widget_class->get_preferred_width = get_preferred_width;
    widget_class->get_preferred_height = get_preferred_height;

    ruler_class->prepare_sizes = prepare_sizes;
    ruler_class->draw_layout = draw_layout;
    ruler_class->draw_tick = draw_tick;
    ruler_class->draw_pos = draw_pos;
    ruler_class->queue_marker_redraw = queue_marker_redraw;
}

static void
gwy_hruler_init(GwyHRuler *hruler)
{
    GwyHRulerPrivate *priv;

    hruler->priv = priv = gwy_hruler_get_instance_private(hruler);
    priv->min_width = 2*LINE_WIDTH + 1;
    priv->min_height = 2*LINE_WIDTH + RULER_WIDTH;
}

/**
 * gwy_hruler_new:
 *
 * Creates a new #GwyHRuler.
 *
 * Returns: The new ruler as a #GtkWidget.
 **/
GtkWidget*
gwy_hruler_new(void)
{
    return g_object_new(GWY_TYPE_HRULER, NULL);
}

static void
get_preferred_width(GtkWidget *widget,
                    gint *minimum, gint *natural)
{
    GwyHRuler *hruler = GWY_HRULER(widget);

    *minimum = *natural = hruler->priv->min_width;
}

static void
get_preferred_height(GtkWidget *widget,
                     gint *minimum, gint *natural)
{
    GwyHRuler *hruler = GWY_HRULER(widget);

    *minimum = *natural = hruler->priv->min_height;
}

static gboolean
motion_notify(GtkWidget *widget, GdkEventMotion *event)
{
    GwyRuler *ruler = GWY_RULER(widget);
    gwy_ruler_move_marker(ruler, gwy_ruler_coord_widget_to_real(ruler, event->x));
    gdk_event_request_motions(event);

    return FALSE;
}

static void
prepare_sizes(GwyRuler *ruler)
{
    GdkRectangle allocation;
    gtk_widget_get_allocation(GTK_WIDGET(ruler), &allocation);

    /* TODO GTK3 Protected, hide from public API. */
    ruler->hthickness = LINE_WIDTH;
    ruler->vthickness = LINE_WIDTH;
    ruler->height = allocation.height;
    ruler->pixelsize = allocation.width;
}

static void
draw_layout(GwyRuler *ruler, cairo_t *cr,
            gint hpos, gint vpos)
{
    GtkWidget *widget = GTK_WIDGET(ruler);
    GtkStyleContext *context = gtk_widget_get_style_context(widget);

    gtk_render_layout(context, cr, hpos + 3, vpos, ruler->layout);
}

static void
draw_tick(GwyRuler *ruler, cairo_t *cr,
          gdouble pos, gdouble length)
{
    gwy_cairo_line(cr, pos + 0.5, ruler->height, pos + 0.5, ruler->height - length);
    /* Let the caller to cairo_stroke(). */
}

static void
draw_pos(GwyRuler *ruler, cairo_t *cr)
{
    gdouble h = 0.5*ruler->height;
    gdouble x = gwy_ruler_coord_real_to_widget(ruler, ruler->position);

    gwy_cairo_triangle_down(cr, x, ruler->height - 0.5*h, 0.5*h);
    cairo_fill(cr);
}

static void
queue_marker_redraw(GwyRuler *ruler, gdouble value)
{
    gint h = (gint)ceil(0.5*ruler->height) + 1;
    gint x = GWY_ROUND(gwy_ruler_coord_real_to_widget(ruler, value));
    gint x0 = CLAMP(x - h/2 - 1, 0, ruler->pixelsize), x1 = CLAMP(x + h/2 + 1, 0, ruler->pixelsize);
    gint y0 = MAX(ruler->height - h, 0), y1 = ruler->height;
    GdkRectangle allocation;
    gtk_widget_get_allocation(GTK_WIDGET(ruler), &allocation);
    gtk_widget_queue_draw_area(GTK_WIDGET(ruler), allocation.x + x0, allocation.y + y0, x1 - x0, y1 - y0);
}

/************************** Documentation ****************************/

/**
 * SECTION:gwyhruler
 * @title: GwyHRuler
 * @short_description: Horizontal ruler, similar to GtkRuler
 * @see_also: #GwyVRuler -- vertical ruler
 *
 * Please see #GwyRuler for differences from #GtkHRuler.
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
