/*
 *  $Id: gwylayer-basic.c 27645 2025-03-27 18:08:13Z yeti-dn $
 *  Copyright (C) 2003-2024 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>

#include "libgwyddion/gwymacros.h"
#include "libprocess/stats.h"
#include "libdraw/gwypixfield.h"
#include "gwydgettypes.h"
#include "gwylayer-basic.h"

#define connect_swapped_after(obj, signal, cb, data) \
    g_signal_connect_object(obj, signal, G_CALLBACK(cb), data, G_CONNECT_SWAPPED | G_CONNECT_AFTER)

enum {
    SGNL_PRESENTATION_SWITCHED,
    NUM_SIGNALS
};

enum {
    PROP_0,
    PROP_GRADIENT_KEY,
    PROP_PRESENTATION_KEY,
    PROP_RANGE_TYPE_KEY,
    PROP_MIN_MAX_KEY,
    PROP_DEFAULT_RANGE_TYPE,
    NUM_PROPERTIES
};

struct _GwyLayerBasicPrivate {
    GQuark range_type_key;
    gulong range_type_id;

    GwyGradient *gradient;
    GQuark gradient_key;
    gulong gradient_id;
    gulong gradient_item_id;

    GQuark show_key;
    GwyDataField *show_field;
    gulong show_id;
    gulong show_item_id;

    GQuark fixed_key;
    gulong min_id;
    gulong max_id;
    gulong handler_id;

    GwyLayerBasicRangeType default_range_type;
};

static void       dispose               (GObject *object);
static void       set_property          (GObject *object,
                                         guint prop_id,
                                         const GValue *value,
                                         GParamSpec *pspec);
static void       get_property          (GObject *object,
                                         guint prop_id,
                                         GValue *value,
                                         GParamSpec *pspec);
static GdkPixbuf* paint                 (GwyPixmapLayer *layer);
static void       plugged               (GwyDataViewLayer *layer);
static void       unplugged             (GwyDataViewLayer *layer);
static void       gradient_connect      (GwyLayerBasic *layer);
static void       gradient_disconnect   (GwyLayerBasic *layer);
static void       get_fixed_range       (GwyLayerBasic *basic_layer,
                                         GwyContainer *container,
                                         GwyDataField *data_field,
                                         gdouble *rmin,
                                         gdouble *rmax);
static void       reconnect_fixed       (GwyLayerBasic *basic_layer);
static void       connect_fixed         (GwyLayerBasic *basic_layer);
static void       disconnect_fixed      (GwyLayerBasic *basic_layer);
static void       item_changed          (GwyLayerBasic *basic_layer);
static void       show_field_connect    (GwyLayerBasic *basic_layer);
static void       show_field_disconnect (GwyLayerBasic *basic_layer);
static void       container_connect     (GwyLayerBasic *basic_layer,
                                         const gchar *data_key_string,
                                         gulong *id,
                                         GCallback callback);
static void       gradient_item_changed (GwyLayerBasic *basic_layer);
static void       show_changed          (GwyLayerBasic *basic_layer);
static void       range_type_changed    (GwyLayerBasic *basic_layer);
static void       min_max_changed       (GwyLayerBasic *basic_layer);
static void       changed               (GwyPixmapLayer *pixmap_layer);
static void       set_default_range_type(GwyLayerBasic *basic_layer,
                                         GwyLayerBasicRangeType range_type);

static guint signals[NUM_SIGNALS];
static GParamSpec *properties[NUM_PROPERTIES] = { NULL, };
static GObjectClass *parent_class = NULL;

G_DEFINE_TYPE_WITH_CODE(GwyLayerBasic, gwy_layer_basic, GWY_TYPE_PIXMAP_LAYER,
                        G_ADD_PRIVATE(GwyLayerBasic))

static void
gwy_layer_basic_class_init(GwyLayerBasicClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GwyDataViewLayerClass *layer_class = GWY_DATA_VIEW_LAYER_CLASS(klass);
    GwyPixmapLayerClass *pixmap_class = GWY_PIXMAP_LAYER_CLASS(klass);
    GType type = G_TYPE_FROM_CLASS(klass);

    parent_class = gwy_layer_basic_parent_class;

    gobject_class->dispose = dispose;
    gobject_class->set_property = set_property;
    gobject_class->get_property = get_property;

    layer_class->plugged = plugged;
    layer_class->unplugged = unplugged;

    pixmap_class->paint = paint;

    /**
     * GwyLayerBasic::presentation-switched:
     * @gwydataview: The #GwyLayerBasic which received the signal.
     *
     * The ::presentation-switched signal is emitted when the presentation state changes, that is when a layer which
     * displayed data directly starts displaying a presentation, or conversely when it stops displaying
     * a presentation.
     **/
    signals[SGNL_PRESENTATION_SWITCHED] = g_signal_new("presentation-switched", type,
                                                       G_SIGNAL_RUN_FIRST,
                                                       G_STRUCT_OFFSET(GwyLayerBasicClass, presentation_switched),
                                                       NULL, NULL,
                                                       g_cclosure_marshal_VOID__VOID,
                                                       G_TYPE_NONE, 0);
    g_signal_set_va_marshaller(signals[SGNL_PRESENTATION_SWITCHED], type, g_cclosure_marshal_VOID__VOIDv);

    /**
     * GwyLayerBasic:gradient-key:
     *
     * The :gradient-key property is the container key used to identify #GwyGradient data is colored with.
     **/
    properties[PROP_GRADIENT_KEY] = g_param_spec_string("gradient-key",
                                                        "Gradient key",
                                                        "Key identifying gradient in container",
                                                        NULL,
                                                        G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    /**
     * GwyLayerBasic:presentation-key:
     *
     * The :presentation-key property is the container key used to identify presentation data field .
     **/
    properties[PROP_PRESENTATION_KEY] = g_param_spec_string("presentation-key",
                                                            "Presentation key",
                                                            "Key identifying presentation data in container",
                                                            NULL,
                                                            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    /**
     * GwyLayerBasic:range-type-key:
     *
     * The :range-type-key property is the container key used to identify color range type.
     **/
    properties[PROP_RANGE_TYPE_KEY] = g_param_spec_string("range-type-key",
                                                          "Range type key",
                                                          "Key identifying color range type in container",
                                                          NULL,
                                                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    /**
     * GwyLayerBasic:min-max-key:
     *
     * The :min-max-key property is the container key prefix used to identify fixed range minimum and maximum.
     **/
    properties[PROP_MIN_MAX_KEY] = g_param_spec_string("min-max-key",
                                                       "Min, max key",
                                                       "Key prefix identifying fixed range minimum and maximum in "
                                                       "container",
                                                       NULL,
                                                       G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    /**
     * GwyLayerBasic:default-range-type:
     *
     * The :default-range-type-key property determines the color mapping range type used when the type is unset in the
     * container.
     **/
    properties[PROP_DEFAULT_RANGE_TYPE] = g_param_spec_enum("default-range-type",
                                                            "Default range type",
                                                            "Default color range mapping type",
                                                            GWY_TYPE_LAYER_BASIC_RANGE_TYPE,
                                                            GWY_LAYER_BASIC_RANGE_FULL,
                                                            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    g_object_class_install_properties(gobject_class, NUM_PROPERTIES, properties);
}

static void
gwy_layer_basic_init(GwyLayerBasic *layer)
{
    layer->priv = gwy_layer_basic_get_instance_private(layer);
}

static void
dispose(GObject *object)
{
    GwyLayerBasicPrivate *priv = GWY_LAYER_BASIC(object)->priv;

    if (priv->gradient) {
        gwy_resource_release(GWY_RESOURCE(priv->gradient));
        priv->gradient = NULL;
    }

    G_OBJECT_CLASS(parent_class)->dispose(object);
}

static void
set_property(GObject *object,
             guint prop_id,
             const GValue *value,
             GParamSpec *pspec)
{
    GwyLayerBasic *basic_layer = GWY_LAYER_BASIC(object);

    switch (prop_id) {
        case PROP_GRADIENT_KEY:
        gwy_layer_basic_set_gradient_key(basic_layer, g_value_get_string(value));
        break;

        case PROP_PRESENTATION_KEY:
        gwy_layer_basic_set_presentation_key(basic_layer, g_value_get_string(value));
        break;

        case PROP_RANGE_TYPE_KEY:
        gwy_layer_basic_set_range_type_key(basic_layer, g_value_get_string(value));
        break;

        case PROP_MIN_MAX_KEY:
        gwy_layer_basic_set_min_max_key(basic_layer, g_value_get_string(value));
        break;

        case PROP_DEFAULT_RANGE_TYPE:
        set_default_range_type(basic_layer, g_value_get_enum(value));
        break;

        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
get_property(GObject *object,
             guint prop_id,
             GValue *value,
             GParamSpec *pspec)
{
    GwyLayerBasicPrivate *priv = GWY_LAYER_BASIC(object)->priv;

    switch (prop_id) {
        case PROP_GRADIENT_KEY:
        g_value_set_static_string(value, g_quark_to_string(priv->gradient_key));
        break;

        case PROP_PRESENTATION_KEY:
        g_value_set_static_string(value, g_quark_to_string(priv->show_key));
        break;

        case PROP_RANGE_TYPE_KEY:
        g_value_set_static_string(value, g_quark_to_string(priv->range_type_key));
        break;

        case PROP_MIN_MAX_KEY:
        g_value_set_static_string(value, g_quark_to_string(priv->fixed_key));
        break;

        case PROP_DEFAULT_RANGE_TYPE:
        g_value_set_enum(value, priv->default_range_type);
        break;

        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/**
 * gwy_layer_basic_new:
 *
 * Creates a new basic data displaying layer.
 *
 * Returns: The newly created layer.
 **/
GwyPixmapLayer*
gwy_layer_basic_new(void)
{
    return g_object_new(GWY_TYPE_LAYER_BASIC, NULL);
}

static GdkPixbuf*
paint(GwyPixmapLayer *layer)
{
    GwyLayerBasic *basic_layer = GWY_LAYER_BASIC(layer);
    GwyLayerBasicPrivate *priv = basic_layer->priv;
    GwyContainer *data = GWY_DATA_VIEW_LAYER(layer)->data;
    GwyDataField *data_field = (priv->show_field ? priv->show_field : layer->data_field);
    GwyGradient *gradient = priv->gradient;
    GwyLayerBasicRangeType range_type;
    gdouble min, max;

    g_return_val_if_fail(data && data_field, NULL);

    /* Special-case full range, as gwy_pixbuf_draw_data_field() is simplier,
     * it doesn't have to deal with outliers */
    range_type = gwy_layer_basic_get_range_type(basic_layer);
    gwy_pixmap_layer_make_pixbuf(layer, FALSE);

    GdkPixbuf *pixbuf = layer->pixbuf;
    if (range_type == GWY_LAYER_BASIC_RANGE_FULL)
        gwy_pixbuf_draw_data_field(pixbuf, data_field, gradient);
    else if (range_type == GWY_LAYER_BASIC_RANGE_ADAPT)
        gwy_pixbuf_draw_data_field_adaptive(pixbuf, data_field, gradient);
    else {
        if (priv->show_field) {
            /* Ignore fixed range in for presentations. */
            if (range_type == GWY_LAYER_BASIC_RANGE_FIXED)
                gwy_pixbuf_draw_data_field(pixbuf, data_field, gradient);
            else {
                gwy_data_field_get_autorange(data_field, &min, &max);
                gwy_pixbuf_draw_data_field_with_range(pixbuf, data_field, gradient, min, max);
            }
        }
        else {
            gwy_layer_basic_get_range(basic_layer, &min, &max);
            gwy_pixbuf_draw_data_field_with_range(pixbuf, data_field, gradient, min, max);
        }
    }

    return pixbuf;
}

static void
gradient_connect(GwyLayerBasic *basic_layer)
{
    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(basic_layer);
    GwyLayerBasicPrivate *priv = basic_layer->priv;
    const gchar *s = NULL;

    g_return_if_fail(!priv->gradient);
    if (priv->gradient_key)
        gwy_container_gis_string(layer->data, priv->gradient_key, &s);
    priv->gradient = gwy_gradients_get_gradient(s);
    gwy_resource_use(GWY_RESOURCE(priv->gradient));
    priv->gradient_id = g_signal_connect_swapped(priv->gradient, "data-changed", G_CALLBACK(changed), layer);
}

static void
gradient_disconnect(GwyLayerBasic *basic_layer)
{
    GwyLayerBasicPrivate *priv = basic_layer->priv;

    if (!priv->gradient)
        return;

    g_clear_signal_handler(&priv->gradient_id, priv->gradient);
    gwy_resource_release(GWY_RESOURCE(priv->gradient));
    priv->gradient = NULL;
}

static void
plugged(GwyDataViewLayer *layer)
{
    GwyLayerBasic *basic_layer = GWY_LAYER_BASIC(layer);
    GwyLayerBasicPrivate *priv = basic_layer->priv;

    GWY_DATA_VIEW_LAYER_CLASS(parent_class)->plugged(layer);

    container_connect(basic_layer, g_quark_to_string(priv->show_key),
                      &priv->show_item_id, G_CALLBACK(item_changed));
    show_field_connect(basic_layer);
    container_connect(basic_layer, g_quark_to_string(priv->gradient_key),
                      &priv->gradient_item_id, G_CALLBACK(gradient_item_changed));
    gradient_connect(basic_layer);
    container_connect(basic_layer, g_quark_to_string(priv->range_type_key),
                      &priv->range_type_id, G_CALLBACK(range_type_changed));
    reconnect_fixed(basic_layer);
}

static void
unplugged(GwyDataViewLayer *layer)
{
    GwyPixmapLayer *pixmap_layer = GWY_PIXMAP_LAYER(layer);
    GwyLayerBasic *basic_layer = GWY_LAYER_BASIC(layer);
    GwyLayerBasicPrivate *priv = basic_layer->priv;

    gwy_debug("disconnecting all handlers");

    disconnect_fixed(basic_layer);
    g_clear_signal_handler(&priv->range_type_id, layer->data);
    g_clear_signal_handler(&priv->gradient_item_id, layer->data);
    gradient_disconnect(basic_layer);
    g_clear_signal_handler(&priv->show_item_id, layer->data);
    show_field_disconnect(basic_layer);

    g_clear_object(&pixmap_layer->pixbuf);
    GWY_DATA_VIEW_LAYER_CLASS(parent_class)->unplugged(layer);
}

/**
 * gwy_layer_basic_set_gradient_key:
 * @basic_layer: A basic data view layer.
 * @key: Container string key identifying the color gradient to use.
 *
 * Sets the container key of the colour gradient to use to visualize data.
 **/
void
gwy_layer_basic_set_gradient_key(GwyLayerBasic *basic_layer,
                                 const gchar *key)
{
    g_return_if_fail(GWY_IS_LAYER_BASIC(basic_layer));

    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(basic_layer);
    GwyLayerBasicPrivate *priv = basic_layer->priv;
    GQuark quark = key ? g_quark_from_string(key) : 0;

    if (!layer->data || priv->gradient_key == quark) {
        priv->gradient_key = quark;
        return;
    }

    g_clear_signal_handler(&priv->gradient_item_id, layer->data);
    gradient_disconnect(basic_layer);
    priv->gradient_key = quark;
    gradient_connect(basic_layer);
    container_connect(basic_layer, g_quark_to_string(quark),
                      &priv->gradient_item_id, G_CALLBACK(gradient_item_changed));

    gwy_pixmap_layer_queue_repaint(GWY_PIXMAP_LAYER(basic_layer));
    gwy_data_view_layer_updated(layer);
}

/**
 * gwy_layer_basic_get_gradient_key:
 * @basic_layer: A basic data view layer.
 *
 * Gets key identifying color gradient.
 *
 * Returns: The string key, or %NULL if it isn't set.
 **/
const gchar*
gwy_layer_basic_get_gradient_key(GwyLayerBasic *basic_layer)
{
    g_return_val_if_fail(GWY_IS_LAYER_BASIC(basic_layer), NULL);
    return g_quark_to_string(basic_layer->priv->gradient_key);
}

/**
 * gwy_layer_basic_set_range_type_key:
 * @basic_layer: A basic data view layer.
 * @key: Container string key identifying the range type to use.
 *
 * Sets the container key of the color range mapping type to use to visualize
 * data.
 **/
void
gwy_layer_basic_set_range_type_key(GwyLayerBasic *basic_layer,
                                   const gchar *key)
{
    g_return_if_fail(GWY_IS_LAYER_BASIC(basic_layer));

    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(basic_layer);
    GwyLayerBasicPrivate *priv = basic_layer->priv;
    GQuark q = key ? g_quark_from_string(key) : 0;

    if (!layer->data || priv->range_type_key == q) {
        priv->range_type_key = q;
        return;
    }

    g_clear_signal_handler(&priv->range_type_id, layer->data);
    priv->range_type_key = q;
    if (q)
        container_connect(basic_layer, g_quark_to_string(q), &priv->range_type_id, G_CALLBACK(range_type_changed));
    reconnect_fixed(basic_layer);

    gwy_pixmap_layer_queue_repaint(GWY_PIXMAP_LAYER(basic_layer));
    gwy_data_view_layer_updated(GWY_DATA_VIEW_LAYER(basic_layer));
}

/**
 * gwy_layer_basic_get_range_type_key:
 * @basic_layer: A basic data view layer.
 *
 * Gets key identifying color range mapping type.
 *
 * Returns: The string key, or %NULL if it isn't set.
 **/
const gchar*
gwy_layer_basic_get_range_type_key(GwyLayerBasic *basic_layer)
{
    g_return_val_if_fail(GWY_IS_LAYER_BASIC(basic_layer), NULL);
    return g_quark_to_string(basic_layer->priv->range_type_key);
}

/**
 * gwy_layer_basic_set_min_max_key:
 * @basic_layer: A basic data view layer.
 * @prefix: Prefix of keys identifying minimum and maximum values for fixed range, "/min" and "/max" is appended to it
 *          to get the individual minimum and maximum keys.
 *
 * Sets basic layer fixed range minimum and maximum.
 **/
void
gwy_layer_basic_set_min_max_key(GwyLayerBasic *basic_layer,
                                const gchar *prefix)
{
    g_return_if_fail(GWY_IS_LAYER_BASIC(basic_layer));

    GwyLayerBasicPrivate *priv = basic_layer->priv;
    GQuark quark = prefix ? g_quark_from_string(prefix) : 0;

    if (quark == priv->fixed_key)
        return;

    priv->fixed_key = quark;
    if (!quark)
        disconnect_fixed(basic_layer);
    else
        reconnect_fixed(basic_layer);

    gwy_pixmap_layer_queue_repaint(GWY_PIXMAP_LAYER(basic_layer));
    gwy_data_view_layer_updated(GWY_DATA_VIEW_LAYER(basic_layer));
}

/**
 * gwy_layer_basic_get_min_max_key:
 * @basic_layer: A basic data view layer.
 *
 * Gets prefix identifying fixed range minimum and maximum.
 *
 * Returns: The prefix, or %NULL if it isn't set.
 **/
const gchar*
gwy_layer_basic_get_min_max_key(GwyLayerBasic *basic_layer)
{
    g_return_val_if_fail(GWY_IS_LAYER_BASIC(basic_layer), NULL);
    return g_quark_to_string(basic_layer->priv->fixed_key);
}

/**
 * gwy_layer_basic_get_range:
 * @basic_layer: A basic data view layer.
 * @min: Location to store range minimum to.
 * @max: Location to store range maximum to.
 *
 * Gets the range colors are mapped from in current mode.
 *
 * This function does not take presentations into account.  It always returns ther range corresponding to the
 * underlying data even if a presentation is shown instead.
 **/
void
gwy_layer_basic_get_range(GwyLayerBasic *basic_layer,
                          gdouble *min,
                          gdouble *max)
{
    g_return_if_fail(GWY_IS_LAYER_BASIC(basic_layer));

    GwyPixmapLayer *pixmap_layer = GWY_PIXMAP_LAYER(basic_layer);
    GwyContainer *data = GWY_DATA_VIEW_LAYER(basic_layer)->data;
    GwyDataField *data_field = GWY_DATA_FIELD(pixmap_layer->data_field);
    gdouble rmin, rmax;

    g_return_if_fail(data && data_field);

    switch (gwy_layer_basic_get_range_type(basic_layer)) {
        case GWY_LAYER_BASIC_RANGE_FULL:
        case GWY_LAYER_BASIC_RANGE_ADAPT:
        gwy_data_field_get_min_max(data_field, &rmin, &rmax);
        break;

        case GWY_LAYER_BASIC_RANGE_FIXED:
        get_fixed_range(basic_layer, data, data_field, &rmin, &rmax);
        break;

        case GWY_LAYER_BASIC_RANGE_AUTO:
        gwy_data_field_get_autorange(data_field, &rmin, &rmax);
        break;

        default:
        g_assert_not_reached();
        break;
    }

    if (min)
        *min = rmin;
    if (max)
        *max = rmax;
}

/**
 * gwy_layer_basic_get_range_type:
 * @basic_layer: A basic data view layer.
 *
 * Gets the current color mapping mode.
 *
 * Returns: The current color mapping mode.
 *
 * Since: 2.7
 **/
GwyLayerBasicRangeType
gwy_layer_basic_get_range_type(GwyLayerBasic *basic_layer)
{
    g_return_val_if_fail(GWY_IS_LAYER_BASIC(basic_layer), 0);

    GwyLayerBasicPrivate *priv = basic_layer->priv;
    GwyContainer *data = GWY_DATA_VIEW_LAYER(basic_layer)->data;
    GwyLayerBasicRangeType range_type = priv->default_range_type;

    if (data && priv->range_type_key)
        gwy_container_gis_enum(data, priv->range_type_key, &range_type);

    return range_type;
}

static void
get_fixed_range(GwyLayerBasic *basic_layer,
                GwyContainer *container,
                GwyDataField *data_field,
                gdouble *rmin,
                gdouble *rmax)
{
    GQuark fixed_key = basic_layer->priv->fixed_key;
    const gchar *prefix;
    gchar *key;

    if (!fixed_key) {
        gwy_data_field_get_min_max(data_field, rmin, rmax);
        return;
    }

    prefix = g_quark_to_string(fixed_key);

    key = g_strconcat(prefix, "/min", NULL);
    if (!gwy_container_gis_double_by_name(container, key, rmin))
        *rmin = gwy_data_field_get_min(data_field);
    g_free(key);

    key = g_strconcat(prefix, "/max", NULL);
    if (!gwy_container_gis_double_by_name(container, key, rmax))
        *rmax = gwy_data_field_get_max(data_field);
    g_free(key);
}

/**
 * gwy_layer_basic_set_presentation_key:
 * @basic_layer: A basic data view layer.
 * @key: Container string key identifying the data field to actually display.
 *
 * Sets the container key of the data field to actually display by a basic layer.
 *
 * The data field set with gwy_pixmap_layer_get_data_key() is used to obtain values, it determines physical
 * dimensions, etc.  When a data field is set with this method, it is displayed instead of the actual data.
 **/
void
gwy_layer_basic_set_presentation_key(GwyLayerBasic *basic_layer,
                                     const gchar *key)
{
    g_return_if_fail(GWY_IS_LAYER_BASIC(basic_layer));

    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(basic_layer);
    GwyLayerBasicPrivate *priv = basic_layer->priv;
    GQuark quark = key ? g_quark_from_string(key) : 0;

    if (priv->show_key == quark)
        return;

    if (!layer->data) {
        priv->show_key = quark;
        g_object_notify_by_pspec(G_OBJECT(basic_layer), properties[PROP_PRESENTATION_KEY]);
        return;
    }

    gboolean presentation_switched = !!priv->show_field;
    g_clear_signal_handler(&priv->show_item_id, layer->data);
    show_field_disconnect(basic_layer);
    priv->show_key = quark;
    show_field_connect(basic_layer);
    container_connect(basic_layer, g_quark_to_string(quark),
                      &priv->show_item_id, G_CALLBACK(item_changed));
    presentation_switched ^= !!priv->show_field;

    gwy_pixmap_layer_queue_repaint(GWY_PIXMAP_LAYER(basic_layer));
    g_object_notify_by_pspec(G_OBJECT(basic_layer), properties[PROP_PRESENTATION_KEY]);
    if (presentation_switched)
        g_signal_emit(basic_layer, signals[SGNL_PRESENTATION_SWITCHED], 0);
    gwy_data_view_layer_updated(layer);
}

/**
 * gwy_layer_basic_get_presentation_key:
 * @basic_layer: A basic data view layer.
 *
 * Gets the key identifying data field this pixmap layer actually displays.
 *
 * See gwy_layer_basic_set_presentation_key() for details.
 *
 * Returns: The string key, or %NULL if it isn't set.
 **/
const gchar*
gwy_layer_basic_get_presentation_key(GwyLayerBasic *basic_layer)
{
    g_return_val_if_fail(GWY_IS_LAYER_BASIC(basic_layer), NULL);
    return g_quark_to_string(basic_layer->priv->show_key);
}

/**
 * gwy_layer_basic_get_has_presentation:
 * @basic_layer: A basic data view layer.
 *
 * Tests whether a basic layer displays a presentation instead of the data.
 *
 * Returns: %TRUE if the layer has a presentation, %FALSE it it displays the data.
 **/
gboolean
gwy_layer_basic_get_has_presentation(GwyLayerBasic *basic_layer)
{
    g_return_val_if_fail(GWY_IS_LAYER_BASIC(basic_layer), FALSE);
    return basic_layer->priv->show_field != NULL;
}

/**
 * gwy_layer_basic_reconnect_fixed:
 * @basic_layer: A basic data view layer.
 *
 * Connect to min, max container keys, or disconnect, depending on range type.
 **/
static void
reconnect_fixed(GwyLayerBasic *basic_layer)
{
    disconnect_fixed(basic_layer);

    if (gwy_layer_basic_get_range_type(basic_layer) == GWY_LAYER_BASIC_RANGE_FIXED)
        connect_fixed(basic_layer);
}

static void
connect_fixed(GwyLayerBasic *basic_layer)
{
    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(basic_layer);
    GwyLayerBasicPrivate *priv = basic_layer->priv;
    GQuark fixed_key = basic_layer->priv->fixed_key;
    const gchar *prefix;
    gchar *detailed_signal;

    if (!layer->data || !fixed_key)
        return;

    prefix = g_quark_to_string(fixed_key);

    detailed_signal = g_strconcat("item-changed::", prefix, "/min", NULL);
    priv->min_id = connect_swapped_after(layer->data, detailed_signal, min_max_changed, layer);
    g_free(detailed_signal);

    detailed_signal = g_strconcat("item-changed::", prefix, "/max", NULL);
    priv->max_id = connect_swapped_after(layer->data, detailed_signal, min_max_changed, layer);
    g_free(detailed_signal);
}

static void
disconnect_fixed(GwyLayerBasic *basic_layer)
{
    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(basic_layer);

    g_clear_signal_handler(&basic_layer->priv->min_id, layer->data);
    g_clear_signal_handler(&basic_layer->priv->max_id, layer->data);
}

/**
 * gwy_layer_basic_item_changed:
 * @layer: A basic data view layer.
 * @data: Container with the presentation data field this basic layer display.
 *
 * Reconnects signals to a new data field when it was replaced in the container.
 **/
static void
item_changed(GwyLayerBasic *basic_layer)
{
    GwyLayerBasicPrivate *priv = basic_layer->priv;
    gboolean presentation_switched = !!priv->show_field;

    show_field_disconnect(basic_layer);
    show_field_connect(basic_layer);
    presentation_switched ^= !!priv->show_field;
    gwy_pixmap_layer_queue_repaint(GWY_PIXMAP_LAYER(basic_layer));
    if (presentation_switched)
        g_signal_emit(basic_layer, signals[SGNL_PRESENTATION_SWITCHED], 0);
    gwy_data_view_layer_updated(GWY_DATA_VIEW_LAYER(basic_layer));
}

/**
 * gwy_layer_basic_show_field_connect:
 * @basic_layer: A basic layer.
 *
 * Eventually connects to new data field's "data-changed" signal.
 **/
static void
show_field_connect(GwyLayerBasic *basic_layer)
{
    GwyLayerBasicPrivate *priv = basic_layer->priv;

    g_return_if_fail(!priv->show_field);
    if (!priv->show_key)
        return;

    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(basic_layer);
    if (!gwy_container_gis_object(layer->data, priv->show_key, &priv->show_field))
        return;

    g_object_ref(priv->show_field);
    priv->show_id = g_signal_connect_swapped(priv->show_field, "data-changed", G_CALLBACK(show_changed), layer);
}

/**
 * gwy_layer_basic_show_field_disconnect:
 * @basic_layer: A basic layer.
 *
 * Disconnects from all data field's signals and drops reference to it.
 **/
static void
show_field_disconnect(GwyLayerBasic *basic_layer)
{
    GwyLayerBasicPrivate *priv = basic_layer->priv;

    g_clear_signal_handler(&priv->show_id, priv->show_field);
    g_clear_object(&priv->show_field);
}

static void
container_connect(GwyLayerBasic *basic_layer,
                  const gchar *data_key_string,
                  gulong *id,
                  GCallback callback)
{
    GwyDataViewLayer *layer;
    gchar *detailed_signal;

    layer = GWY_DATA_VIEW_LAYER(basic_layer);
    if (!data_key_string || !layer->data) {
        *id = 0;
        return;
    }
    detailed_signal = g_strconcat("item-changed::", data_key_string, NULL);
    *id = connect_swapped_after(layer->data, detailed_signal, callback, layer);
    g_free(detailed_signal);
}

static void
gradient_item_changed(GwyLayerBasic *basic_layer)
{
    gradient_disconnect(basic_layer);
    gradient_connect(basic_layer);
    gwy_pixmap_layer_queue_repaint(GWY_PIXMAP_LAYER(basic_layer));
    gwy_data_view_layer_updated(GWY_DATA_VIEW_LAYER(basic_layer));
}

static void
show_changed(GwyLayerBasic *basic_layer)
{
    gwy_pixmap_layer_queue_repaint(GWY_PIXMAP_LAYER(basic_layer));
    gwy_data_view_layer_updated(GWY_DATA_VIEW_LAYER(basic_layer));
}

static void
range_type_changed(GwyLayerBasic *basic_layer)
{
    reconnect_fixed(basic_layer);
    changed(GWY_PIXMAP_LAYER(basic_layer));
}

static void
min_max_changed(GwyLayerBasic *basic_layer)
{
    if (gwy_layer_basic_get_range_type(basic_layer) == GWY_LAYER_BASIC_RANGE_FIXED)
        changed(GWY_PIXMAP_LAYER(basic_layer));
}

static void
changed(GwyPixmapLayer *pixmap_layer)
{
    gwy_pixmap_layer_queue_repaint(pixmap_layer);
    gwy_data_view_layer_updated(GWY_DATA_VIEW_LAYER(pixmap_layer));
}

static void
set_default_range_type(GwyLayerBasic *basic_layer,
                       GwyLayerBasicRangeType range_type)
{
    GwyLayerBasicPrivate *priv = basic_layer->priv;
    GwyContainer *data;

    if (range_type == priv->default_range_type)
        return;

    priv->default_range_type = range_type;

    data = GWY_DATA_VIEW_LAYER(basic_layer)->data;
    if (data && !(priv->range_type_key && gwy_container_contains(data, priv->range_type_key)))
        range_type_changed(basic_layer);
    g_object_notify_by_pspec(G_OBJECT(basic_layer), properties[PROP_DEFAULT_RANGE_TYPE]);
}

/************************** Documentation ****************************/

/**
 * SECTION:gwylayer-basic
 * @title: GwyLayerBasic
 * @short_description: Display data colored with a gradient
 * @see_also: #GwyDataView -- data display widget,
 *            #GwyLayerMask -- display data as color mask,
 *            #GwyGradient -- color gradient representation
 *
 * #GwyLayerBasic is a data view layer that can paint data with a false color gradient.  It is not normally usable
 * outside a #GwyDataView.
 *
 * The layer takes all infromation from its parent #GwyDataView, therefore you do not set its properties directly, but
 * you change them in the corresponding #GwyContainer and/or use methods like gwy_layer_basic_set_gradient_key() to
 * make the layer read a different value from the container.
 *
 * Several types of mapping of data to colors are available, see #GwyLayerBasicRangeType for their description.
 **/

/**
 * GwyLayerBasicRangeType:
 * @GWY_LAYER_BASIC_RANGE_FULL: Color gradient is uniformly mapped to range
 *                              from data minimum to maximum.
 * @GWY_LAYER_BASIC_RANGE_FIXED: Color gradient is uniformly mapped to a fixed
 *                               range, independent on data.
 * @GWY_LAYER_BASIC_RANGE_AUTO: Color gradient is uniformly mapped to a range
 *                              inside which most of data points lie, that is
 *                              height distribution tails are cut off.
 * @GWY_LAYER_BASIC_RANGE_ADAPT: Color range is mapped nonuniformly,
 *                               see gwy_pixbuf_draw_data_field_adaptive().
 *
 * Types of color gradient mapping in #GwyLayerBasic.
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
