/*
 *  $Id: gwylayer-mask.c 27110 2025-01-02 11:02:46Z yeti-dn $
 *  Copyright (C) 2003-2024 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>

#include "libgwyddion/gwymacros.h"
#include "libdraw/gwypixfield.h"

#include "gwylayer-mask.h"

#define connect_swapped_after(obj, signal, cb, data) \
    g_signal_connect_object(obj, signal, G_CALLBACK(cb), data, G_CONNECT_SWAPPED | G_CONNECT_AFTER)

enum {
    PROP_0,
    PROP_COLOR_KEY,
    NUM_PROPERTIES
};

struct _GwyLayerMaskPrivate {
    GQuark color_key;
    gulong red_id;
    gulong green_id;
    gulong blue_id;
    gulong alpha_id;
    gulong handler_id;
};

static void       set_property    (GObject *object,
                                   guint prop_id,
                                   const GValue *value,
                                   GParamSpec *pspec);
static void       get_property    (GObject *object,
                                   guint prop_id,
                                   GValue *value,
                                   GParamSpec *pspec);
static GdkPixbuf* paint           (GwyPixmapLayer *layer);
static void       plugged         (GwyDataViewLayer *layer);
static void       unplugged       (GwyDataViewLayer *layer);
static void       connect_color   (GwyLayerMask *mask_layer);
static void       disconnect_color(GwyLayerMask *mask_layer);
static void       changed         (GwyPixmapLayer *pixmap_layer);

static GParamSpec *properties[NUM_PROPERTIES] = { NULL, };
static GObjectClass *parent_class = NULL;

G_DEFINE_TYPE_WITH_CODE(GwyLayerMask, gwy_layer_mask, GWY_TYPE_PIXMAP_LAYER,
                        G_ADD_PRIVATE(GwyLayerMask))

static void
gwy_layer_mask_class_init(GwyLayerMaskClass *klass)
{
    GwyDataViewLayerClass *layer_class = GWY_DATA_VIEW_LAYER_CLASS(klass);
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GwyPixmapLayerClass *pixmap_class = GWY_PIXMAP_LAYER_CLASS(klass);

    parent_class = gwy_layer_mask_parent_class;

    gobject_class->set_property = set_property;
    gobject_class->get_property = get_property;

    layer_class->plugged = plugged;
    layer_class->unplugged = unplugged;

    pixmap_class->paint = paint;

    /**
     * GwyLayerMask:color-key:
     *
     * The :color-key property is the container key used to identify mask color
     * in container.
     **/
    properties[PROP_COLOR_KEY] = g_param_spec_string("color-key",
                                                     "Color key",
                                                     "Key prefix identifying mask color in container",
                                                     NULL,
                                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    g_object_class_install_properties(gobject_class, NUM_PROPERTIES, properties);
}

static void
gwy_layer_mask_init(GwyLayerMask *layer)
{
    layer->priv = gwy_layer_mask_get_instance_private(layer);
}

static void
set_property(GObject *object,
             guint prop_id,
             const GValue *value,
             GParamSpec *pspec)
{
    GwyLayerMask *layer_mask = GWY_LAYER_MASK(object);

    switch (prop_id) {
        case PROP_COLOR_KEY:
        gwy_layer_mask_set_color_key(layer_mask, g_value_get_string(value));
        break;

        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
get_property(GObject *object,
             guint prop_id,
             GValue *value,
             GParamSpec *pspec)
{
    GwyLayerMaskPrivate *priv = GWY_LAYER_MASK(object)->priv;

    switch (prop_id) {
        case PROP_COLOR_KEY:
        g_value_set_static_string(value, g_quark_to_string(priv->color_key));
        break;

        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/**
 * gwy_layer_mask_new:
 *
 * Creates a new mask layer.
 *
 * By default, is uses a transparent color (thus not displaying anything).
 *
 * Returns: The newly created layer.
 **/
GwyPixmapLayer*
gwy_layer_mask_new(void)
{
    return g_object_new(GWY_TYPE_LAYER_MASK, NULL);
}

static GdkPixbuf*
paint(GwyPixmapLayer *layer)
{
    GwyLayerMask *mask_layer = GWY_LAYER_MASK(layer);
    GwyDataViewLayer *dv_layer = GWY_DATA_VIEW_LAYER(layer);
    GwyDataField *data_field = GWY_DATA_FIELD(layer->data_field);
    GwyContainer *data = dv_layer->data;
    GwyLayerMaskPrivate *priv = mask_layer->priv;
    GwyRGBA color = { 0, 0, 0, 0 };

    g_return_val_if_fail(data && data_field, NULL);
    if (priv->color_key)
        gwy_rgba_get_from_container(&color, dv_layer->data, g_quark_to_string(priv->color_key));
    gwy_pixmap_layer_make_pixbuf(layer, TRUE);
    gwy_pixbuf_draw_data_field_as_mask(layer->pixbuf, data_field, &color);

    return layer->pixbuf;
}

/**
 * gwy_layer_mask_get_color:
 * @mask_layer: A mask layer.
 *
 * Returns the color used by a mask layer.
 *
 * Returns: The color as #GwyRGBA.
 **/
GwyRGBA
gwy_layer_mask_get_color(GwyLayerMask *mask_layer)
{
    GwyRGBA color = { 0, 0, 0, 0 };

    g_return_val_if_fail(GWY_IS_LAYER_MASK(mask_layer), color);

    GwyLayerMaskPrivate *priv = mask_layer->priv;
    GwyContainer *data = GWY_DATA_VIEW_LAYER(mask_layer)->data;

    g_return_val_if_fail(data, color);

    if (priv->color_key)
        gwy_rgba_get_from_container(&color, GWY_DATA_VIEW_LAYER(mask_layer)->data, g_quark_to_string(priv->color_key));

    return color;
}

static void
plugged(GwyDataViewLayer *layer)
{
    GwyLayerMask *mask_layer = GWY_LAYER_MASK(layer);

    GWY_DATA_VIEW_LAYER_CLASS(parent_class)->plugged(layer);

    connect_color(mask_layer);
}

static void
unplugged(GwyDataViewLayer *layer)
{
    GwyPixmapLayer *pixmap_layer = GWY_PIXMAP_LAYER(layer);
    GwyLayerMask *mask_layer = GWY_LAYER_MASK(layer);

    disconnect_color(mask_layer);

    g_clear_object(&pixmap_layer->pixbuf);
    GWY_DATA_VIEW_LAYER_CLASS(parent_class)->unplugged(layer);
}

/**
 * gwy_layer_mask_set_color_key:
 * @mask_layer: A mask layer.
 * @prefix: Prefix of keys identifying mask color components, "/red", "/green", "/blue", and "/alpha" is appended to
 *          it to get the individual keys.
 *
 * Sets the container key of colour components of a mask layer.
 **/
void
gwy_layer_mask_set_color_key(GwyLayerMask *mask_layer,
                             const gchar *prefix)
{
    g_return_if_fail(GWY_IS_LAYER_MASK(mask_layer));

    GwyLayerMaskPrivate *priv = mask_layer->priv;
    GQuark quark = prefix ? g_quark_try_string(prefix) : 0;
    if (quark == priv->color_key)
        return;

    disconnect_color(mask_layer);
    priv->color_key = quark;
    connect_color(mask_layer);

    changed(GWY_PIXMAP_LAYER(mask_layer));
}

/**
 * gwy_layer_mask_get_color_key:
 * @mask_layer: A mask layer.
 *
 * Gets prefix identifying color components.
 *
 * Returns: The prefix, or %NULL if it isn't set.
 **/
const gchar*
gwy_layer_mask_get_color_key(GwyLayerMask *mask_layer)
{
    g_return_val_if_fail(GWY_IS_LAYER_MASK(mask_layer), NULL);
    return g_quark_to_string(mask_layer->priv->color_key);
}

static void
connect_color(GwyLayerMask *mask_layer)
{
    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(mask_layer);
    GwyLayerMaskPrivate *priv = mask_layer->priv;
    GwyContainer *data = layer->data;
    const gchar *prefix;
    gchar *detailed_signal;

    if (!data || !priv->color_key)
        return;

    prefix = g_quark_to_string(priv->color_key);

    detailed_signal = g_strconcat("item-changed::", prefix, "/red", NULL);
    priv->red_id = connect_swapped_after(data, detailed_signal, changed, layer);
    g_free(detailed_signal);

    detailed_signal = g_strconcat("item-changed::", prefix, "/green", NULL);
    priv->green_id = connect_swapped_after(data, detailed_signal, changed, layer);
    g_free(detailed_signal);

    detailed_signal = g_strconcat("item-changed::", prefix, "/blue", NULL);
    priv->blue_id = connect_swapped_after(data, detailed_signal, changed, layer);
    g_free(detailed_signal);

    detailed_signal = g_strconcat("item-changed::", prefix, "/alpha", NULL);
    priv->alpha_id = connect_swapped_after(data, detailed_signal, changed, layer);
    g_free(detailed_signal);
}

static void
disconnect_color(GwyLayerMask *mask_layer)
{
    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(mask_layer);
    GwyLayerMaskPrivate *priv = mask_layer->priv;
    GwyContainer *data = layer->data;

    g_clear_signal_handler(&priv->red_id, data);
    g_clear_signal_handler(&priv->green_id, data);
    g_clear_signal_handler(&priv->blue_id, data);
    g_clear_signal_handler(&priv->alpha_id, data);
}

static void
changed(GwyPixmapLayer *pixmap_layer)
{
    gwy_pixmap_layer_queue_repaint(pixmap_layer);
    gwy_data_view_layer_updated(GWY_DATA_VIEW_LAYER(pixmap_layer));
}

/************************** Documentation ****************************/

/**
 * SECTION:gwylayer-mask
 * @title: GwyLayerMask
 * @short_description: Display data as color mask
 * @see_also: #GwyDataView -- data display widget,
 *            #GwyLayerBasic -- display data colored with a gradient,
 *            #GwyRGBA -- color representation
 *
 * #GwyLayerMask is a data view layer that can paint data as a color mask with varying opacity.  It is not normally
 * usable outside a #GwyDataView.
 *
 * The layer takes all infromation from its parent #GwyDataView, therefore you do not set its properties directly, but
 * you change them in the corresponding #GwyContainer and/or use methods like gwy_layer_mask_set_color_key() to make
 * the layer read a different value from the container.
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
