/*
 *  $Id: gwypixmaplayer.c 27380 2025-02-27 17:00:56Z yeti-dn $
 *  Copyright (C) 2003-2024 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>
#include <libgwyddion/gwymacros.h>
#include <libprocess/datafield.h>
#include <libgwydgets/gwypixmaplayer.h>

#define BITS_PER_SAMPLE 8

#define connect_swapped_after(obj, signal, cb, data) \
    g_signal_connect_object(obj, signal, G_CALLBACK(cb), data, G_CONNECT_SWAPPED | G_CONNECT_AFTER)

enum {
    PROP_0,
    PROP_DATA_KEY,
    NUM_PROPERTIES
};

struct _GwyPixmapLayerPrivate {
    GQuark data_key;
    gulong item_changed_id;
    gulong data_changed_id;
    gulong handler_id;
    gboolean wants_repaint;
};

static void set_property         (GObject *object,
                                  guint prop_id,
                                  const GValue *value,
                                  GParamSpec *pspec);
static void get_property         (GObject *object,
                                  guint prop_id,
                                  GValue *value,
                                  GParamSpec *pspec);
static void dispose              (GObject *object);
static void plugged              (GwyDataViewLayer *layer);
static void unplugged            (GwyDataViewLayer *layer);
static void item_changed         (GwyPixmapLayer *pixmap_layer);
static void data_changed         (GwyPixmapLayer *pixmap_layer);
static void container_connect    (GwyPixmapLayer *pixmap_layer,
                                  const gchar *data_key_string);
static void data_field_connect   (GwyPixmapLayer *pixmap_layer);
static void data_field_disconnect(GwyPixmapLayer *pixmap_layer);

static GParamSpec *properties[NUM_PROPERTIES] = { NULL, };
static GwyDataViewLayerClass *parent_class = NULL;

G_DEFINE_ABSTRACT_TYPE_WITH_CODE(GwyPixmapLayer, gwy_pixmap_layer, GWY_TYPE_DATA_VIEW_LAYER,
                                 G_ADD_PRIVATE(GwyPixmapLayer))

static void
gwy_pixmap_layer_class_init(GwyPixmapLayerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GwyDataViewLayerClass *layer_class = GWY_DATA_VIEW_LAYER_CLASS(klass);

    parent_class = gwy_pixmap_layer_parent_class;

    gobject_class->set_property = set_property;
    gobject_class->get_property = get_property;
    gobject_class->dispose = dispose;

    layer_class->plugged = plugged;
    layer_class->unplugged = unplugged;

    /**
     * GwyPixmapLayer:data-key:
     *
     * The :data-key property is the container key used to identify
     * displayed #GwyDataField in container.
     */
    properties[PROP_DATA_KEY] = g_param_spec_string("data-key",
                                                    "Data key",
                                                    "Key identifying data field in container",
                                                    NULL,
                                                    G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    g_object_class_install_properties(gobject_class, NUM_PROPERTIES, properties);
}

static void
gwy_pixmap_layer_init(GwyPixmapLayer *layer)
{
    layer->priv = gwy_pixmap_layer_get_instance_private(layer);
}

static void
dispose(GObject *object)
{
    GwyPixmapLayer *layer = GWY_PIXMAP_LAYER(object);

    g_clear_object(&layer->data_field);
    g_clear_object(&layer->pixbuf);

    G_OBJECT_CLASS(parent_class)->dispose(object);
}

static void
set_property(GObject *object,
             guint prop_id,
             const GValue *value,
             GParamSpec *pspec)
{
    GwyPixmapLayer *pixmap_layer = GWY_PIXMAP_LAYER(object);

    switch (prop_id) {
        case PROP_DATA_KEY:
        gwy_pixmap_layer_set_data_key(pixmap_layer, g_value_get_string(value));
        break;

        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
get_property(GObject *object,
             guint prop_id,
             GValue *value,
             GParamSpec *pspec)
{
    GwyPixmapLayerPrivate *priv = GWY_PIXMAP_LAYER(object)->priv;

    switch (prop_id) {
        case PROP_DATA_KEY:
        g_value_set_static_string(value, g_quark_to_string(priv->data_key));
        break;

        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/**
 * gwy_pixmap_layer_paint:
 * @pixmap_layer: A pixmap data view layer.
 *
 * Returns a pixbuf with painted pixmap layer.
 *
 * This method does not enforce repaint.  If the layer doesn't think it needs to repaint the pixbuf, it simply returns
 * the current one.  To enforce update, emit "data-changed" signal on corresponding data field.
 *
 * Returns: The pixbuf.  It should not be modified or freed.  If the data field to draw is not present in the
 *          container, %NULL is returned.
 **/
GdkPixbuf*
gwy_pixmap_layer_paint(GwyPixmapLayer *pixmap_layer)
{
    GwyPixmapLayerClass *layer_class = GWY_PIXMAP_LAYER_GET_CLASS(pixmap_layer);
    GwyPixmapLayerPrivate *priv = pixmap_layer->priv;
    GdkPixbuf *pixbuf = NULL;

    g_return_val_if_fail(GWY_IS_PIXMAP_LAYER(pixmap_layer), NULL);
    g_return_val_if_fail(layer_class->paint, NULL);

    if (!pixmap_layer->data_field || !GWY_IS_DATA_FIELD(pixmap_layer->data_field)) {
        if (!priv->data_key)
            g_warning("No data field to paint. Set data key with gwy_pixmap_layer_set_data_key().");
        /* If the data key is set but there is no data field, fail silently and let GwyDataView deal with it. */
    }
    else {
        if (priv->wants_repaint)
            layer_class->paint(pixmap_layer);
        pixbuf = pixmap_layer->pixbuf;
    }
    priv->wants_repaint = FALSE;

    return pixbuf;
}

/**
 * gwy_pixmap_layer_wants_repaint:
 * @pixmap_layer: A pixmap data view layer.
 *
 * Checks whether a pixmap layer wants repaint.
 *
 * Returns: %TRUE if the the layer wants repaint itself, %FALSE otherwise.
 **/
gboolean
gwy_pixmap_layer_wants_repaint(GwyPixmapLayer *pixmap_layer)
{
    g_return_val_if_fail(GWY_IS_PIXMAP_LAYER(pixmap_layer), FALSE);
    return pixmap_layer->priv->wants_repaint;
}

/**
 * gwy_pixmap_layer_queue_repaint:
 * @pixmap_layer: A pixmap data view layer.
 *
 * Sets the want-repaint flag for a pixmap layer.
 **/
void
gwy_pixmap_layer_queue_repaint(GwyPixmapLayer *pixmap_layer)
{
    g_return_if_fail(GWY_IS_PIXMAP_LAYER(pixmap_layer));
    pixmap_layer->priv->wants_repaint = TRUE;
}

/**
 * gwy_pixmap_layer_data_field_connect:
 * @pixmap_layer: A pixmap layer.
 *
 * Eventually connects to new data field's "data-changed" signal.
 **/
static void
data_field_connect(GwyPixmapLayer *pixmap_layer)
{
    GwyPixmapLayerPrivate *priv = pixmap_layer->priv;
    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(pixmap_layer);

    g_return_if_fail(!pixmap_layer->data_field);
    if (!priv->data_key)
        return;

    if (!gwy_container_gis_object(layer->data, priv->data_key, &pixmap_layer->data_field))
        return;

    g_object_ref(pixmap_layer->data_field);
    priv->data_changed_id = g_signal_connect_swapped(pixmap_layer->data_field, "data-changed",
                                                     G_CALLBACK(data_changed), layer);
}

/**
 * data_field_disconnect:
 * @pixmap_layer: A pixmap layer.
 *
 * Disconnects from all data field's signals and drops reference to it.
 **/
static void
data_field_disconnect(GwyPixmapLayer *pixmap_layer)
{
    GwyPixmapLayerPrivate *priv = pixmap_layer->priv;

    g_clear_signal_handler(&priv->data_changed_id, pixmap_layer->data_field);
    g_clear_object(&pixmap_layer->data_field);
}

static void
plugged(GwyDataViewLayer *layer)
{
    GwyPixmapLayer *pixmap_layer = GWY_PIXMAP_LAYER(layer);
    GwyPixmapLayerPrivate *priv = pixmap_layer->priv;

    parent_class->plugged(layer);

    if (!priv->data_key)
        return;

    container_connect(pixmap_layer, g_quark_to_string(priv->data_key));
    data_field_connect(pixmap_layer);
    priv->wants_repaint = TRUE;
}

static void
unplugged(GwyDataViewLayer *layer)
{
    GwyPixmapLayer *pixmap_layer = GWY_PIXMAP_LAYER(layer);
    GwyPixmapLayerPrivate *priv = pixmap_layer->priv;

    priv->wants_repaint = FALSE;
    data_field_disconnect(pixmap_layer);
    g_clear_signal_handler(&priv->item_changed_id, layer->data);

    parent_class->unplugged(layer);
}

/**
 * item_changed:
 * @layer: A pixmap data view layer.
 * @data: Container with the data field this pixmap layer display.
 *
 * Reconnects signals to a new data field when it was replaced in the
 * container.
 **/
static void
item_changed(GwyPixmapLayer *pixmap_layer)
{
    GwyPixmapLayerPrivate *priv = pixmap_layer->priv;

    data_field_disconnect(pixmap_layer);
    data_field_connect(pixmap_layer);
    priv->wants_repaint = TRUE;
    gwy_data_view_layer_updated(GWY_DATA_VIEW_LAYER(pixmap_layer));
}

static void
data_changed(GwyPixmapLayer *pixmap_layer)
{
    GwyPixmapLayerPrivate *priv = pixmap_layer->priv;

    priv->wants_repaint = TRUE;
    gwy_data_view_layer_updated(GWY_DATA_VIEW_LAYER(pixmap_layer));
}

/**
 * gwy_pixmap_layer_set_data_key:
 * @pixmap_layer: A pixmap layer.
 * @key: Container string key identifying the data field to display.
 *
 * Sets the data field to display by a pixmap layer.
 **/
void
gwy_pixmap_layer_set_data_key(GwyPixmapLayer *pixmap_layer,
                              const gchar *key)
{
    g_return_if_fail(GWY_IS_PIXMAP_LAYER(pixmap_layer));

    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(pixmap_layer);
    GwyPixmapLayerPrivate *priv = pixmap_layer->priv;
    GQuark quark = key ? g_quark_from_string(key) : 0;

    if (priv->data_key == quark)
        return;

    if (!layer->data) {
        priv->data_key = quark;
        g_object_notify_by_pspec(G_OBJECT(pixmap_layer), properties[PROP_DATA_KEY]);
        return;
    }

    g_clear_signal_handler(&priv->item_changed_id, layer->data);
    data_field_disconnect(pixmap_layer);
    priv->data_key = quark;
    data_field_connect(pixmap_layer);
    container_connect(pixmap_layer, g_quark_to_string(quark));

    priv->wants_repaint = TRUE;
    g_object_notify_by_pspec(G_OBJECT(pixmap_layer), properties[PROP_DATA_KEY]);
    gwy_data_view_layer_updated(GWY_DATA_VIEW_LAYER(pixmap_layer));
}

static void
container_connect(GwyPixmapLayer *pixmap_layer,
                  const gchar *data_key_string)
{
    GwyPixmapLayerPrivate *priv = pixmap_layer->priv;
    GwyDataViewLayer *layer = GWY_DATA_VIEW_LAYER(pixmap_layer);
    gchar *detailed_signal;

    if (!data_key_string) {
        priv->item_changed_id = 0;
        return;
    }

    detailed_signal = g_strconcat("item-changed::", data_key_string, NULL);
    priv->item_changed_id = connect_swapped_after(layer->data, detailed_signal, item_changed, layer);
    g_free(detailed_signal);
}

/**
 * gwy_pixmap_layer_get_data_key:
 * @pixmap_layer: A pixmap layer.
 *
 * Gets the key identifying data field this pixmap layer displays.
 *
 * Returns: The string key, or %NULL if it isn't set.
 **/
const gchar*
gwy_pixmap_layer_get_data_key(GwyPixmapLayer *pixmap_layer)
{
    g_return_val_if_fail(GWY_IS_PIXMAP_LAYER(pixmap_layer), NULL);
    return g_quark_to_string(pixmap_layer->priv->data_key);
}

/**
 * gwy_pixmap_layer_make_pixbuf:
 * @pixmap_layer: A pixmap layer.
 * @has_alpha: Whether pixbuf should have alpha channel.
 *
 * Creates or resizes pixmap layer #GdkPixbuf to match its data field.
 *
 * This method is intended for pixmap layer implementation.
 **/
void
gwy_pixmap_layer_make_pixbuf(GwyPixmapLayer *pixmap_layer,
                             gboolean has_alpha)
{
    g_return_if_fail(GWY_IS_PIXMAP_LAYER(pixmap_layer));

    GwyDataField *data_field = GWY_DATA_FIELD(pixmap_layer->data_field);
    g_return_if_fail(data_field);

    gint dwidth = gwy_data_field_get_xres(data_field);
    gint dheight = gwy_data_field_get_yres(data_field);
    if (pixmap_layer->pixbuf) {
        gint pwidth = gdk_pixbuf_get_width(pixmap_layer->pixbuf);
        gint pheight = gdk_pixbuf_get_height(pixmap_layer->pixbuf);
        if (pwidth == dwidth && pheight == dheight)
            return;

        g_clear_object(&pixmap_layer->pixbuf);
    }

    pixmap_layer->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, has_alpha, BITS_PER_SAMPLE, dwidth, dheight);
}

/************************** Documentation ****************************/

/**
 * SECTION:gwypixmaplayer
 * @title: GwyPixmapLayer
 * @short_description: Base class for GwyDataView pixmap layers
 *
 * #GwyPixmapLayer is a base class for data field displaying #GwyDataViewLayer's.  It is a #GwyDataView component and
 * it is not normally usable outside of it.
 *
 * The layer takes the data field to display from its parent #GwyDataView. The key under which the data field is found
 * must be set with gwy_pixmap_layer_set_data_key().
 *
 * The other methods are only rarely needed outside #GwyDataView implementation.
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
