/*
 *  $Id: gwyruler.c 27380 2025-02-27 17:00:56Z yeti-dn $
 *  Copyright (C) 2003-2024 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>

#include "libgwyddion/gwymacros.h"
#include "libgwyddion/gwymath.h"

#include "gwyruler.h"
#include "gwyhruler.h"
#include "gwyvruler.h"
#include "gwydgettypes.h"
#include "widget-impl-utils.h"

enum { MINIMUM_INCR = 5 };

enum {
    PROP_0,
    PROP_LOWER,
    PROP_UPPER,
    PROP_POSITION,
    PROP_MAX_SIZE,
    PROP_UNITS_PLACEMENT,
    NUM_PROPERTIES,
};

typedef enum {
    SCALE_0 = 0,
    SCALE_1,
    SCALE_2,
    SCALE_2_5,
    SCALE_5,
    NUM_SCALES,
} ScaleBase;

struct _GwyRulerPrivate {
    GdkWindow *event_window;
    GwySIValueFormat *vformat;
    gint xsrc, ysrc;
    gdouble max_size;    /* The maximum size of the ruler */
    gdouble drawn_position;
    gboolean drawn_position_set;
    GwySIUnit *units;
    GwyUnitsPlacement units_placement;
};

struct _GwyRulerClassPrivate {
    gint dummy;
};

static void      draw_ticks         (GwyRuler *ruler,
                                     cairo_t *cr);
static void      realize            (GtkWidget *widget);
static void      unrealize          (GtkWidget *widget);
static void      map                (GtkWidget *widget);
static void      unmap              (GtkWidget *widget);
static void      size_allocate      (GtkWidget *widget,
                                     GdkRectangle *allocation);
static gint      draw               (GtkWidget *widget,
                                     cairo_t *cr);
static void      finalize           (GObject *object);
static void      set_property       (GObject *object,
                                     guint prop_id,
                                     const GValue *value,
                                     GParamSpec *pspec);
static void      get_property       (GObject *object,
                                     guint prop_id,
                                     GValue *value,
                                     GParamSpec *pspec);
static void      update_value_format(GwyRuler *ruler);
static gboolean  is_precision_ok    (const GwySIValueFormat *format,
                                     gdouble base,
                                     ScaleBase scale);
static ScaleBase next_scale         (ScaleBase scale,
                                     gdouble *base,
                                     gdouble measure,
                                     gint min_incr);

static const gdouble scale_steps[NUM_SCALES] = {
    0.0, 1.0, 2.0, 2.5, 5.0,
};

static GParamSpec *properties[NUM_PROPERTIES] = { NULL, };
static GtkWidgetClass *parent_class = NULL;

G_DEFINE_ABSTRACT_TYPE_WITH_CODE(GwyRuler, gwy_ruler, GTK_TYPE_WIDGET,
                                 G_ADD_PRIVATE(GwyRuler))

static void
gwy_ruler_class_init(GwyRulerClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(class);
    //GType type = G_TYPE_FROM_CLASS(klass);

    parent_class = gwy_ruler_parent_class;

    gobject_class->finalize = finalize;
    gobject_class->set_property = set_property;
    gobject_class->get_property = get_property;

    widget_class->realize = realize;
    widget_class->unrealize = unrealize;
    widget_class->map = map;
    widget_class->unmap = unmap;
    widget_class->size_allocate = size_allocate;
    widget_class->draw = draw;

    properties[PROP_LOWER] = g_param_spec_double("lower", "Lower limit",
                                                 "Lower limit (minimum value)",
                                                 -G_MAXDOUBLE, G_MAXDOUBLE, 0.0, GWY_GPARAM_RWE);

    properties[PROP_UPPER] = g_param_spec_double("upper", "Upper limit",
                                                 "Upper limit (maximum value)",
                                                 -G_MAXDOUBLE, G_MAXDOUBLE, 0.0, GWY_GPARAM_RWE);

    properties[PROP_POSITION] = g_param_spec_double("position", "Position",
                                                    "Value where a mark should be drawn",
                                                    -G_MAXDOUBLE, G_MAXDOUBLE, 0.0, GWY_GPARAM_RWE);

    properties[PROP_MAX_SIZE] = g_param_spec_double("max-size", "Maximum size",
                                                    "Maximum magnigude of values",
                                                    -G_MAXDOUBLE, G_MAXDOUBLE, 0.0, GWY_GPARAM_RWE);

    properties[PROP_UNITS_PLACEMENT] = g_param_spec_enum("units-placement", "Units placement",
                                                         "Placement of units",
                                                         GWY_TYPE_UNITS_PLACEMENT, GWY_UNITS_PLACEMENT_NONE,
                                                         GWY_GPARAM_RWE);

    g_object_class_install_properties(gobject_class, NUM_PROPERTIES, properties);
}

static void
gwy_ruler_init(GwyRuler *ruler)
{
    GwyRulerPrivate *priv;

    ruler->priv = priv = gwy_ruler_get_instance_private(ruler);
    priv->units = gwy_si_unit_new(NULL);
    ruler->pixelsize = 1;
    update_value_format(ruler);

    gtk_widget_set_has_window(GTK_WIDGET(ruler), FALSE);
}

static void
finalize(GObject *object)
{
    GwyRulerPrivate *priv = GWY_RULER(object)->priv;

    g_clear_object(&priv->units);
    gwy_si_unit_value_format_free(priv->vformat);
}

static void
set_property(GObject *object,
             guint prop_id,
             const GValue *value,
             GParamSpec *pspec)
{
    GwyRuler *ruler = GWY_RULER(object);
    GwyRulerPrivate *priv = ruler->priv;

    switch (prop_id) {
        case PROP_LOWER:
        gwy_ruler_set_range(ruler, g_value_get_double(value), ruler->upper, ruler->position, priv->max_size);
        break;

        case PROP_UPPER:
        gwy_ruler_set_range(ruler, ruler->lower, g_value_get_double(value), ruler->position, priv->max_size);
        break;

        case PROP_POSITION:
        gwy_ruler_set_range(ruler, ruler->lower, ruler->upper, g_value_get_double(value), priv->max_size);
        break;

        case PROP_MAX_SIZE:
        gwy_ruler_set_range(ruler, ruler->lower, ruler->upper, ruler->position, g_value_get_double(value));
        break;

        case PROP_UNITS_PLACEMENT:
        gwy_ruler_set_units_placement(ruler, (GwyUnitsPlacement)g_value_get_uint(value));
        break;

        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
get_property(GObject *object,
             guint prop_id,
             GValue *value,
             GParamSpec *pspec)
{
    GwyRuler *ruler = GWY_RULER(object);

    switch (prop_id) {
        case PROP_LOWER:
        g_value_set_double(value, ruler->lower);
        break;

        case PROP_UPPER:
        g_value_set_double(value, ruler->upper);
        break;

        case PROP_POSITION:
        g_value_set_double(value, ruler->position);
        break;

        case PROP_MAX_SIZE:
        g_value_set_double(value, ruler->priv->max_size);
        break;

        case PROP_UNITS_PLACEMENT:
        g_value_set_uint(value, (guint)ruler->priv->units_placement);
        break;

        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/**
 * gwy_ruler_set_range:
 * @ruler: A #GwyRuler
 * @lower: Lower limit of the ruler.
 * @upper: Upper limit of the ruler.
 * @position: Current position of the mark on the ruler.
 * @max_size: Maximum value used for calculating size of text labels.
 *
 * Sets range and current value of a ruler.
 **/
void
gwy_ruler_set_range(GwyRuler *ruler,
                    gdouble lower,
                    gdouble upper,
                    gdouble position,
                    gdouble max_size)
{
    g_return_if_fail(GWY_IS_RULER(ruler));

    GwyRulerPrivate *priv = ruler->priv;
    GObject *object = G_OBJECT(ruler);

    g_object_freeze_notify(object);
    if (ruler->lower != lower) {
        ruler->lower = lower;
        g_object_notify_by_pspec(object, properties[PROP_LOWER]);
    }
    if (ruler->upper != upper) {
        ruler->upper = upper;
        g_object_notify_by_pspec(object, properties[PROP_UPPER]);
    }
    if (ruler->position != position) {
        ruler->position = position;
        g_object_notify_by_pspec(object, properties[PROP_POSITION]);
    }
    if (priv->max_size != max_size) {
        priv->max_size = max_size;
        g_object_notify_by_pspec(object, properties[PROP_MAX_SIZE]);
    }
    g_object_thaw_notify(object);
    update_value_format(ruler);

    if (gtk_widget_is_drawable(GTK_WIDGET(ruler)))
        gtk_widget_queue_draw(GTK_WIDGET(ruler));
}

/**
 * gwy_ruler_get_range:
 * @ruler: A #GwyRuler
 * @lower: Location to store lower limit of the ruler, or %NULL
 * @upper: Location to store upper limit of the ruler, or %NULL
 * @position: Location to store the current position of the mark on the ruler, or %NULL
 * @max_size: Location to store the maximum size of the ruler used when calculating the space to leave for the text,
 *            or %NULL.
 *
 * Retrieves values indicating the range and current position of a #GwyRuler. See gwy_ruler_set_range().
 **/
void
gwy_ruler_get_range(GwyRuler *ruler,
                    gdouble *lower,
                    gdouble *upper,
                    gdouble *position,
                    gdouble *max_size)
{
    g_return_if_fail(GWY_IS_RULER(ruler));

    if (lower)
        *lower = ruler->lower;
    if (upper)
        *upper = ruler->upper;
    if (position)
        *position = ruler->position;
    if (max_size)
        *max_size = ruler->priv->max_size;
}

/**
 * gwy_ruler_set_units_placement:
 * @ruler: A #GwyRuler
 * @placement: Units placement specification.
 *
 * Sets whether and where units should be placed on the ruler.
 **/
void
gwy_ruler_set_units_placement(GwyRuler *ruler,
                              GwyUnitsPlacement placement)
{
    g_return_if_fail(GWY_IS_RULER(ruler));

    GwyRulerPrivate *priv = ruler->priv;
    placement = MIN(placement, GWY_UNITS_PLACEMENT_AT_ZERO);
    if (priv->units_placement == placement)
        return;

    priv->units_placement = placement;
    g_object_notify_by_pspec(G_OBJECT(ruler), properties[PROP_UNITS_PLACEMENT]);

    if (gtk_widget_is_drawable(GTK_WIDGET(ruler)))
        gtk_widget_queue_draw(GTK_WIDGET(ruler));
}

/**
 * gwy_ruler_get_units_placement:
 * @ruler: A #GwyRuler
 *
 * Gets current units placement of ruler @ruler.
 *
 * Returns: The units placement.
 **/
GwyUnitsPlacement
gwy_ruler_get_units_placement(GwyRuler *ruler)
{
    g_return_val_if_fail(GWY_IS_RULER(ruler), GWY_UNITS_PLACEMENT_NONE);
    return ruler->priv->units_placement;
}

static void
realize(GtkWidget *widget)
{
    GwyRuler *ruler = GWY_RULER(widget);
    GwyRulerPrivate *priv = ruler->priv;

    /* FIXME GTK3 widgets generally do not call parent's realize. Not sure why because it does a couple of things for
     * us like setting the widget realized. */
    GTK_WIDGET_CLASS(parent_class)->realize(widget);
    priv->event_window = gwy_create_widget_input_window(widget, GDK_POINTER_MOTION_MASK);
    ruler->layout = gtk_widget_create_pango_layout(widget, NULL);
}

static void
unrealize(GtkWidget *widget)
{
    GwyRuler *ruler = GWY_RULER(widget);
    GwyRulerPrivate *priv = ruler->priv;

    gwy_destroy_widget_input_window(widget, &priv->event_window);
    g_clear_object(&ruler->layout);
    GTK_WIDGET_CLASS(parent_class)->unrealize(widget);
}

static void
map(GtkWidget *widget)
{
    GwyRuler *ruler = GWY_RULER(widget);
    GwyRulerPrivate *priv = ruler->priv;

    parent_class->map(widget);

    if (priv->event_window)
        gdk_window_show(priv->event_window);
}

static void
unmap(GtkWidget *widget)
{
    GwyRuler *ruler = GWY_RULER(widget);
    GwyRulerPrivate *priv = ruler->priv;

    if (priv->event_window) {
        /* FIXME: Do we need to update the state in some other way? */
        gdk_window_hide(priv->event_window);
    }

    parent_class->unmap(widget);
}

static void
size_allocate(GtkWidget *widget, GdkRectangle *allocation)
{
    GwyRuler *ruler = GWY_RULER(widget);
    GwyRulerPrivate *priv = ruler->priv;

    parent_class->size_allocate(widget, allocation);

    if (priv->event_window) {
        gdk_window_move_resize(priv->event_window, allocation->x, allocation->y, allocation->width, allocation->height);
        update_value_format(ruler);
        priv->drawn_position_set = FALSE;
    }
}

static gboolean
draw(GtkWidget *widget, cairo_t *cr)
{
    GwyRuler *ruler = GWY_RULER(widget);
    GtkStyleContext *context = gwy_setup_widget_foreground_drawing(widget, cr);
    GwyRulerClass *klass = GWY_RULER_GET_CLASS(ruler);

    GdkRectangle allocation;
    gtk_widget_get_allocation(widget, &allocation);
    gtk_render_background(context, cr, 0, 0, allocation.width, allocation.height);

    draw_ticks(ruler, cr);

    GwyRulerPrivate *priv = ruler->priv;
    if (klass->draw_pos) {
        klass->draw_pos(ruler, cr);
        priv->drawn_position = ruler->position;
        priv->drawn_position_set = TRUE;
    }

    return FALSE;
}

/**
 * gwy_ruler_set_si_unit:
 * @ruler: A #GwyRuler.
 * @units: The base units this ruler should display.
 *
 * Sets the base units a ruler displays.
 *
 * Setting units to %NULL effectively disables them.
 **/
void
gwy_ruler_set_si_unit(GwyRuler *ruler,
                      GwySIUnit *units)
{
    g_return_if_fail(GWY_IS_RULER(ruler));

    GwyRulerPrivate *priv = ruler->priv;
    if (units) {
        g_return_if_fail(GWY_IS_SI_UNIT(units));
        if (gwy_si_unit_equal(priv->units, units))
            return;
        gwy_si_unit_assign(priv->units, units);
    }
    else {
        if (gwy_si_unit_equal_string(priv->units, NULL))
            return;
        gwy_si_unit_set_from_string(priv->units, NULL);
    }
    update_value_format(ruler);
    gtk_widget_queue_draw(GTK_WIDGET(ruler));
}

/**
 * gwy_ruler_get_si_unit:
 * @ruler: A #GwyRuler.
 *
 * Returns the base units a ruler uses.
 *
 * Returns: The units the rules uses.
 **/
GwySIUnit*
gwy_ruler_get_si_unit(GwyRuler *ruler)
{
    g_return_val_if_fail(GWY_IS_RULER(ruler), NULL);
    return ruler->priv->units;
}

/**
 * gwy_ruler_move_marker:
 * @ruler: A #GwyRuler.
 * @value: Value where to draw marker.
 *
 * Changes the position of a ruler marker.
 *
 * The position is also automatically updated when pointer is moved over the marker. This function is meant for
 * updating the ruler when pointer moves over other widgets.
 **/
void
gwy_ruler_move_marker(GwyRuler *ruler,
                      gdouble value)
{
    g_return_if_fail(GWY_IS_RULER(ruler));
    if (value == ruler->position)
        return;

    gboolean must_redraw = (CLAMP(value, ruler->lower, ruler->upper)
                            != CLAMP(ruler->position, ruler->lower, ruler->upper));
    ruler->position = value;
    if (must_redraw && gtk_widget_get_realized(GTK_WIDGET(ruler))) {
        GwyRulerPrivate *priv = ruler->priv;

        GWY_RULER_GET_CLASS(ruler)->queue_marker_redraw(ruler, value);
        if (priv->drawn_position_set)
            GWY_RULER_GET_CLASS(ruler)->queue_marker_redraw(ruler, priv->drawn_position);
    }
    g_object_notify_by_pspec(G_OBJECT(ruler), properties[PROP_POSITION]);
}

/**
 * gwy_ruler_coord_real_to_widget:
 * @ruler: A #GwyRuler.
 * @value: Real value on the ruler.
 *
 * Transforms real coordinate to position on the ruler widget.
 *
 * Returns: Position on widget corresponding to the value.
 **/
gdouble
gwy_ruler_coord_real_to_widget(GwyRuler *ruler,
                               gdouble value)
{
    g_return_val_if_fail(GWY_IS_RULER(ruler), 0.0);
    gdouble d = (ruler->upper - ruler->lower)/ruler->pixelsize;
    gdouble x = (value - ruler->lower)/d;
    return CLAMP(x, 0.0, ruler->pixelsize);
}

/**
 * gwy_ruler_coord_widget_to_real:
 * @ruler: A #GwyRuler.
 * @pos: Pixel position on the ruler, in widget coordinates.
 *
 * Transforms pixel position a ruler to real coordinate value.
 *
 * Returns: Ruler value corresponding to widget position.
 **/
gdouble
gwy_ruler_coord_widget_to_real(GwyRuler *ruler,
                               gdouble pos)
{
    g_return_val_if_fail(GWY_IS_RULER(ruler), 0.0);
    gdouble d = (ruler->upper - ruler->lower)/ruler->pixelsize;
    return ruler->lower + pos*d;
}

static void
update_value_format(GwyRuler *ruler)
{
    GwyRulerPrivate *priv = ruler->priv;
    gdouble max;

    max = priv->max_size;
    if (!max)
        max = MAX(fabs(ruler->lower), fabs(ruler->upper));
    if (!max)
        max = 1.2;

    priv->vformat = gwy_si_unit_get_format_with_resolution(priv->units, GWY_SI_UNIT_FORMAT_VFMARKUP, max, max/21,
                                                           priv->vformat);
}

static void
draw_ticks(GwyRuler *ruler, cairo_t *cr)
{
    enum {
        FINALLY_OK,
        FIRST_TRY,
        ADD_DIGITS,
        LESS_TICKS
    };

    struct {
        ScaleBase scale;
        gdouble base;
    } tick_info[4];

    gint labels = 0, scale_depth;
    gdouble step, first, bstep;
    GwyRulerPrivate *priv = ruler->priv;
    GwyRulerClass *klass = GWY_RULER_GET_CLASS(ruler);
    PangoRectangle rect;

    klass->prepare_sizes(ruler);

    gint min_tick_spacing = MINIMUM_INCR;
    gint min_label_spacing = ruler->hthickness + MINIMUM_INCR;

    GwySIValueFormat *vf = priv->vformat;
    PangoLayout *layout = ruler->layout;
    gdouble upper = ruler->upper;
    gdouble lower = ruler->lower;
    if (upper <= lower || ruler->pixelsize < 2 || ruler->pixelsize > 10000)
        return;
    gdouble max = priv->max_size;
    if (max == 0)
        max = MAX(fabs(lower), fabs(upper));

    gdouble range = upper - lower;
    gdouble measure = range/ruler->pixelsize;

    GString *unit_str = g_string_new(NULL);
    ScaleBase scale = SCALE_1;
    gint state = FIRST_TRY;
    gint ascent = 0;
    gdouble base; // = sizeof("Die, die GCC warning!");
    do {
        if (state != LESS_TICKS) {
            if (priv->units_placement == GWY_UNITS_PLACEMENT_AT_ZERO && *vf->units) {
                if (lower > 0)
                    g_string_printf(unit_str, "%.*f %s", vf->precision, lower/vf->magnitude, vf->units);
                else
                    g_string_printf(unit_str, "0 %s", vf->units);
            }
            else
                g_string_printf(unit_str, "%.*f", vf->precision, max/vf->magnitude);

            pango_layout_set_markup(layout, unit_str->str, unit_str->len);
            pango_layout_get_pixel_extents(layout, NULL, &rect);
            ascent = PANGO_ASCENT(rect);
            gint text_size = rect.width + 1;

            if (lower < 0) {
                g_string_printf(unit_str, "%.*f", vf->precision, lower/vf->magnitude);
                if (unit_str->str[0] == '-') {
                    g_string_erase(unit_str, 0, 1);
                    g_string_insert(unit_str, 0, "−");
                }
                pango_layout_set_markup(layout, unit_str->str, unit_str->len);
                pango_layout_get_pixel_extents(layout, NULL, &rect);
                if (text_size < rect.width + 1)
                    text_size = rect.width + 1;
            }

            /* fit as many labels as you can */
            labels = floor(ruler->pixelsize/(text_size + ruler->hthickness + min_label_spacing));
            labels = MAX(labels, 1);
            if (labels > 8)
                labels = 8 + (labels - 7)/2;

            step = range / labels;
            base = gwy_exp10(floor(log10(step) + 1e-12));
            step /= base;
            while (step <= 0.5) {
                base /= 10.0;
                step *= 10.0;
            }
            while (step > 5.0) {
                base *= 10.0;
                step /= 10.0;
            }
            if (step <= 1.0)
                scale = SCALE_1;
            else if (step <= 2.0)
                scale = SCALE_2;
            else if (step <= 2.5)
                scale = SCALE_2_5;
            else
                scale = SCALE_5;
        }

        step = scale_steps[scale];
        bstep = base*step;
        first = floor(lower / (bstep))*bstep;
        gwy_debug("%d first: %g, base: %g, step: %g, prec: %d, labels: %d",
                  state, first, base, step, vf->precision, labels);

        if (is_precision_ok(vf, base, scale)) {
            if (state == FIRST_TRY && vf->precision > 0)
                vf->precision--;
            else
                state = FINALLY_OK;
        }
        else {
            if (state == FIRST_TRY) {
                state = ADD_DIGITS;
                vf->precision++;
            }
            else {
                state = LESS_TICKS;
                base *= 10;
                scale = SCALE_1;
            }
        }
    } while (state != FINALLY_OK);

    /* draw labels */
    gboolean units_drawn = FALSE;
    for (gint i = 0; ; i++) {
        gint pos;
        gdouble val;

        val = i*bstep + first;
        pos = floor((val - lower)/measure);
        if (pos >= ruler->pixelsize)
            break;
        if (pos < 0)
            continue;
        if (!units_drawn
            && (upper < 0 || val >= 0)
            && priv->units_placement == GWY_UNITS_PLACEMENT_AT_ZERO
            && priv->units) {
            if (val == 0)
                g_string_printf(unit_str, "0 %s", vf->units);
            else
                g_string_printf(unit_str, "%.*f %s", vf->precision, val/vf->magnitude, vf->units);
            units_drawn = TRUE;
        }
        else
            g_string_printf(unit_str, "%.*f", vf->precision, val/vf->magnitude);

        if (unit_str->str[0] == '-') {
            g_string_erase(unit_str, 0, 1);
            g_string_insert(unit_str, 0, "−");
        }

        pango_layout_set_markup(ruler->layout, unit_str->str, unit_str->len);
        /* XXX: This is the best approximation of same positioning I'm able to do, but it's still wrong. */
        pango_layout_get_pixel_extents(ruler->layout, NULL, &rect);

        gint vpos = ruler->vthickness + ascent - PANGO_ASCENT(rect);
        klass->draw_layout(ruler, cr, pos, vpos);
    }

    g_string_free(unit_str, TRUE);

    /* draw tick marks, from smallest to largest */
    scale_depth = 0;
    while (scale && scale_depth < (gint)G_N_ELEMENTS(tick_info)) {
        tick_info[scale_depth].scale = scale;
        tick_info[scale_depth].base = base;
        scale = next_scale(scale, &base, measure, min_tick_spacing);
        scale_depth++;
    }
    scale_depth--;

    cairo_set_line_width(cr, 1.0);
    cairo_set_line_cap(cr, CAIRO_LINE_CAP_BUTT);
    while (scale_depth > -1) {
        gint tick_length = ruler->height/(scale_depth + 1) - 2;
        scale = tick_info[scale_depth].scale;
        base = tick_info[scale_depth].base;
        bstep = base*scale_steps[scale];
        first = floor(lower/bstep)*bstep;
        for (gint i = 0; ; i++) {
            gint pos;
            gdouble val;

            val = (i + 0.000001)*bstep + first;
            pos = floor((val - lower)/measure);
            if (pos >= ruler->pixelsize)
                break;
            if (pos < 0)
                continue;

            klass->draw_tick(ruler, cr, pos, tick_length);
        }
        scale_depth--;
    }
    cairo_stroke(cr);
}

static gboolean
is_precision_ok(const GwySIValueFormat *format,
                gdouble base,
                ScaleBase scale)
{
    gint m = GWY_ROUND(log10(base/format->magnitude));

    return format->precision + m >= (scale == SCALE_2_5);
}

static ScaleBase
next_scale(ScaleBase scale,
           gdouble *base,
           gdouble measure,
           gint min_incr)
{
    ScaleBase new_scale = SCALE_0;

    switch (scale) {
        case SCALE_1:
        *base /= 10.0;
        if ((gint)floor(*base*2.0/measure) > min_incr)
            new_scale = SCALE_5;
        else if ((gint)floor(*base*2.5/measure) > min_incr)
            new_scale = SCALE_2_5;
        else if ((gint)floor(*base*5.0/measure) > min_incr)
            new_scale = SCALE_5;
        break;

        case SCALE_2:
        if ((gint)floor(*base/measure) > min_incr)
            new_scale = SCALE_1;
        break;

        case SCALE_2_5:
        *base /= 10.0;
        if ((gint)floor(*base*5.0/measure) > min_incr)
            new_scale = SCALE_5;
        break;

        case SCALE_5:
        if ((gint)floor(*base/measure) > min_incr)
            new_scale = SCALE_1;
        else if ((gint)floor(*base*2.5/measure) > min_incr)
            new_scale = SCALE_2_5;
        break;

        default:
        g_assert_not_reached();
        break;
    }

    return new_scale;
}

/************************** Documentation ****************************/

/**
 * SECTION:gwyruler
 * @title: GwyRuler
 * @short_description: Base class for GwyHRuler and GwyVRuler
 *
 * #GwyRuler is a ruler similar to #GtkRuler, but it is more suited for a scientific application.  It is
 * scale-independent and thus has no arbitrary limits on the ranges or interpretation of displayed values.  It can
 * display units on the ruler (this can be controlled with gwy_ruler_set_units_placement()) and cooperates with
 * #GwySIUnit (see gwy_ruler_set_si_unit()).  It also better follows Gtk+ theme than #GtkRuler.
 **/

/**
 * GwyUnitsPlacement:
 * @GWY_UNITS_PLACEMENT_NONE: Units are omitted.
 * @GWY_UNITS_PLACEMENT_AT_ZERO: Units are placed to major tick at zero, or
 *                               to the leftmost position of zero is not
 *                               present.
 *
 * Units placement on a #GwyRuler.
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
