/*
 *  $Id: gwyvruler.c 27110 2025-01-02 11:02:46Z yeti-dn $
 *  Copyright (C) 2003-2024 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>

#include "libgwyddion/gwymacros.h"
#include "libgwyddion/gwymath.h"

#include "cairo-utils.h"
#include "gwyvruler.h"

#define LINE_WIDTH 1
#define RULER_WIDTH 20

struct _GwyVRulerPrivate {
    gint min_width;
    gint min_height;
};

static void     realize             (GtkWidget *widget);
static void     get_preferred_width (GtkWidget *widget,
                                     gint *minimum,
                                     gint *natural);
static void     get_preferred_height(GtkWidget *widget,
                                     gint *minimum,
                                     gint *natural);
static gboolean motion_notify       (GtkWidget *widget,
                                     GdkEventMotion *event);
static void     prepare_sizes       (GwyRuler *ruler);
static void     draw_layout         (GwyRuler *ruler,
                                     cairo_t *cr,
                                     gint hpos,
                                     gint vpos);
static void     draw_tick           (GwyRuler *ruler,
                                     cairo_t *cr,
                                     gdouble pos,
                                     gdouble length);
static void     draw_pos            (GwyRuler *ruler,
                                     cairo_t *cr);
static void     queue_marker_redraw (GwyRuler *ruler,
                                     gdouble value);

static GObjectClass *parent_class = NULL;

G_DEFINE_TYPE_WITH_CODE(GwyVRuler, gwy_vruler, GWY_TYPE_RULER,
                        G_ADD_PRIVATE(GwyVRuler))

static void
gwy_vruler_class_init(GwyVRulerClass *klass)
{
    GwyRulerClass *ruler_class = GWY_RULER_CLASS(klass);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);

    parent_class = gwy_vruler_parent_class;

    widget_class->motion_notify_event = motion_notify;
    widget_class->get_preferred_width = get_preferred_width;
    widget_class->get_preferred_height = get_preferred_height;
    widget_class->realize = realize;

    ruler_class->prepare_sizes = prepare_sizes;
    ruler_class->draw_layout = draw_layout;
    ruler_class->draw_tick = draw_tick;
    ruler_class->draw_pos = draw_pos;
    ruler_class->queue_marker_redraw = queue_marker_redraw;
}

static void
gwy_vruler_init(GwyVRuler *vruler)
{
    GwyVRulerPrivate *priv;

    vruler->priv = priv = gwy_vruler_get_instance_private(vruler);
    priv->min_height = 2*LINE_WIDTH + 1;
    priv->min_width = 2*LINE_WIDTH + RULER_WIDTH;
}

/**
 * gwy_vruler_new:
 *
 * Creates a new #GwyVRuler.
 *
 * Returns: The new ruler as a #GtkWidget.
 **/
GtkWidget*
gwy_vruler_new(void)
{
    return g_object_new(GWY_TYPE_VRULER, NULL);
}

static void
realize(GtkWidget *widget)
{
    if (GTK_WIDGET_CLASS(parent_class)->realize)
        (GTK_WIDGET_CLASS(parent_class)->realize)(widget);

    GwyRuler *ruler = GWY_RULER(widget);
    const PangoMatrix *cmatrix;
    PangoMatrix matrix = PANGO_MATRIX_INIT;
    PangoContext *context = pango_layout_get_context(ruler->layout);

    if ((cmatrix = pango_context_get_matrix(context)))
        matrix = *cmatrix;
    pango_matrix_rotate(&matrix, -90.0);
    pango_context_set_matrix(context, &matrix);
    pango_layout_context_changed(ruler->layout);
}

static void
get_preferred_width(GtkWidget *widget,
                    gint *minimum, gint *natural)
{
    GwyVRuler *vruler = GWY_VRULER(widget);

    *minimum = *natural = vruler->priv->min_width;
}

static void
get_preferred_height(GtkWidget *widget,
                     gint *minimum, gint *natural)
{
    GwyVRuler *vruler = GWY_VRULER(widget);

    *minimum = *natural = vruler->priv->min_height;
}

static gboolean
motion_notify(GtkWidget *widget, GdkEventMotion *event)
{
    GwyRuler *ruler = GWY_RULER(widget);
    gwy_ruler_move_marker(ruler, gwy_ruler_coord_widget_to_real(ruler, event->y));
    gdk_event_request_motions(event);

    return FALSE;
}

static void
prepare_sizes(GwyRuler *ruler)
{
    GdkRectangle allocation;
    gtk_widget_get_allocation(GTK_WIDGET(ruler), &allocation);

    /* TODO GTK3 Protected, hide from public API. */
    ruler->hthickness = LINE_WIDTH;
    ruler->vthickness = LINE_WIDTH;
    ruler->height = allocation.width;
    ruler->pixelsize = allocation.height;
}

static void
draw_layout(GwyRuler *ruler, cairo_t *cr,
            gint hpos, gint vpos)
{
    GtkWidget *widget = GTK_WIDGET(ruler);
    GtkStyleContext *context = gtk_widget_get_style_context(widget);

    /* FIXME: The (h|v)positions are swapped here becuase we had some wonky coordinate convention. */
    gtk_render_layout(context, cr, ruler->height - vpos, hpos + 3, ruler->layout);
}

static void
draw_tick(GwyRuler *ruler, cairo_t *cr,
          gdouble pos, gdouble length)
{
    gwy_cairo_line(cr, ruler->height, pos + 0.5, ruler->height - length, pos + 0.5);
    /* Let the caller to cairo_stroke(). */
}

static void
draw_pos(GwyRuler *ruler, cairo_t *cr)
{
    gdouble w = 0.5*ruler->height;
    gdouble y = gwy_ruler_coord_real_to_widget(ruler, ruler->position);

    gwy_cairo_triangle_right(cr, ruler->height - 0.5*w, y, 0.5*w);
    cairo_fill(cr);
}

static void
queue_marker_redraw(GwyRuler *ruler, gdouble value)
{
    gint w = (gint)ceil(0.5*ruler->height) + 1;
    gint y = GWY_ROUND(gwy_ruler_coord_real_to_widget(ruler, value));
    gint x0 = MAX(ruler->height - w, 0), x1 = ruler->height;
    gint y0 = CLAMP(y - w/2 - 1, 0, ruler->pixelsize), y1 = CLAMP(y + w/2 + 1, 0, ruler->pixelsize);
    GdkRectangle allocation;
    gtk_widget_get_allocation(GTK_WIDGET(ruler), &allocation);
    gtk_widget_queue_draw_area(GTK_WIDGET(ruler), allocation.x + x0, allocation.y + y0, x1 - x0, y1 - y0);
}

/************************** Documentation ****************************/

/**
 * SECTION:gwyvruler
 * @title: GwyVRuler
 * @short_description: Vertical ruler, similar to GtkRuler
 * @see_also: #GwyHRuler -- horizontal ruler
 *
 * Please see #GwyRuler for differences from #GtkVRuler.
 *
 * A specific #GwyVRuler feature is that it's drawn vertically including text, like a rotated horizontal ruler.
 **/

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
