/*
 *  $Id: fft_filter_2d.c 27504 2025-03-09 18:58:44Z yeti-dn $
 *  Copyright (C) 2005 Christopher Anderson, Molecular Imaging Corp.
 *  E-mail: Chris Anderson (sidewinder.asu@gmail.com)
 *  Copyright (C) 2011-2025 David Necas (Yeti).
 *  E-mail: yeti@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <gtk/gtk.h>
#include <libgwyddion/gwymacros.h>
#include <libgwyddion/gwymath.h>
#include <libprocess/arithmetic.h>
#include <libprocess/elliptic.h>
#include <libprocess/inttrans.h>
#include <libprocess/stats.h>
#include <libgwydgets/gwystock.h>
#include <libgwymodule/gwymodule-process.h>
#include <app/gwymoduleutils.h>
#include <app/gwyapp.h>
#include "preview.h"
#include "app/sanity.h"

#define RUN_MODES (GWY_RUN_INTERACTIVE)

enum {
    PARAM_ZOOM,
    PARAM_EDIT_MODE,
    PARAM_DISPLAY,
    PARAM_OUTPUT,
    PARAM_SNAP,

    BUTTON_UNDO,
    BUTTON_CLEAR,
    BUTTON_FILL,
};

enum {
    RESPONSE_UNDO = GWY_RESPONSE_USER,
    RESPONSE_FILL,
};

typedef enum {
    UPDATE_NOTHING  = 0,
    UPDATE_ZOOMED   = 1 << 0,
    UPDATE_FILTERED = 1 << 1,
} UpdateWhat;

typedef enum {
    FFT_ELLIPSE_ADD   = 0,
    FFT_RECTANGLE_ADD = 1,
    FFT_ELLIPSE_SUB   = 2,
    FFT_RECTANGLE_SUB = 3,
} EditType;

typedef enum {
    PREVIEW_FFT      = 0,
    PREVIEW_IMAGE    = 1,
    PREVIEW_FILTERED = 2,
    PREVIEW_DIFF     = 3,
} PreviewType;

typedef enum {
    OUTPUT_FFT   = 0,
    OUTPUT_IMAGE = 1,
    OUTPUT_DIFF  = 2,
} OutputType;

typedef struct {
    GwyParams *params;
    GwyDataField *field;
    GwyDataField *mask;
    GwyDataField *result;
    GwyDataField *diff;
    GwyDataField *filtered_modulus;
} ModuleArgs;

typedef struct {
    ModuleArgs *args;
    GwyDataField *modulus;
    GwyContainer    *data;

    GtkWidget *dialog;
    GwyParamTable *table;

    gulong          rect_signal;
    gulong          ellipse_signal;

    GtkWidget       *view;
    GwyPixmapLayer  *view_layer;
    GwyPixmapLayer  *mask_layer;
    UpdateWhat update;
} ModuleGUI;

typedef void (*FieldFillFunc)(GwyDataField *field,
                              gint col, gint row,
                              gint width, gint height,
                              gdouble value);

static gboolean         module_register        (void);
static void             module_main            (GwyContainer *data,
                                                GwyRunType run);
static void             execute                (ModuleArgs *args);
static GwyDialogOutcome run_gui                (ModuleArgs *args,
                                                GwyContainer *data,
                                                gint id);
static void             dialog_response        (ModuleGUI *gui,
                                                gint response);
static void             preview                (gpointer user_data);
static void             param_changed          (ModuleGUI *gui,
                                                gint id);
static void             update_sensitivity     (ModuleGUI *gui);
static void             selection_finished     (GwySelection *selection,
                                                ModuleGUI *gui);
static void             update_snap            (ModuleGUI *gui);
static void             switch_layer           (ModuleGUI *gui);
static void             calculate_zoomed_fields(ModuleGUI *gui);
static GwyDataField*    cut_field_to_zoom      (GwyDataField *field,
                                                gint zoom);
static void             calculate_modulus      (GwyDataField *field,
                                                GwyDataField *fftmod);
static gint             add_output_field       (GwyContainer *data,
                                                GwyDataField *field,
                                                gint oldid,
                                                const gchar *title);

static GwyModuleInfo module_info = {
    GWY_MODULE_ABI_VERSION,
    &module_register,
    N_("2D FFT Filtering"),
    "Chris Anderson <sidewinder.asu@gmail.com>",
    "2.0",
    "Chris Anderson, Molecular Imaging Corp.",
    "2005",
};

GWY_MODULE_QUERY2(module_info, fft_filter_2d)

static gboolean
module_register(void)
{
    gwy_process_func_register("fft_filter_2d",
                              (GwyProcessFunc)module_main,
                              N_("/_Correct Data/_2D FFT Filtering..."),
                              GWY_ICON_FFT_FILTER_2D,
                              RUN_MODES,
                              GWY_MENU_FLAG_DATA,
                              N_("Two-dimensional FFT filtering"));

    return TRUE;
}

static GwyParamDef*
define_module_params(void)
{
    static const GwyEnum displays[] = {
        { N_("_FFT mask editor"),  PREVIEW_FFT,      },
        { N_("Original _image"),   PREVIEW_IMAGE,    },
        { N_("Fi_ltered image"),   PREVIEW_FILTERED, },
        { N_("Image _difference"), PREVIEW_DIFF,     },
    };
    static const GwyEnum outputs[] = {
        { N_("Filtered i_mage"),       1u << OUTPUT_IMAGE, },
        { N_("Ima_ge difference"),     1u << OUTPUT_DIFF,  },
        { N_("Filtered FFT mo_dulus"), 1u << OUTPUT_FFT,   },
    };
    static const GwyEnum modes[] = {
        { N_("Add an ellipse to the FFT mask"),         FFT_ELLIPSE_ADD, },
        { N_("Add a rectangle to the FFT mask"),        FFT_RECTANGLE_ADD,    },
        { N_("Subtract an ellipse from the FFT mask"),  FFT_ELLIPSE_SUB, },
        { N_("Subtract a rectangle from the FFT mask"), FFT_RECTANGLE_SUB,    },
    };
    static GwyEnum zooms[5];
    static GwyParamDef *paramdef = NULL;

    if (paramdef)
        return paramdef;

    for (guint i = 0; i < G_N_ELEMENTS(zooms); i++) {
        zooms[i].value = 1u << i;
        zooms[i].name = g_strdup_printf("%u×", zooms[i].value);
    }

    paramdef = gwy_param_def_new();
    gwy_param_def_set_function_name(paramdef, gwy_process_func_current());
    gwy_param_def_add_gwyenum(paramdef, PARAM_ZOOM, "zoom", _("Zoom"), zooms, G_N_ELEMENTS(zooms), 1);
    gwy_param_def_add_gwyenum(paramdef, PARAM_EDIT_MODE, "edit_mode", _("Edit mode"),
                              modes, G_N_ELEMENTS(modes), FFT_ELLIPSE_ADD);
    gwy_param_def_add_gwyenum(paramdef, PARAM_DISPLAY, NULL, _("Display"),
                              displays, G_N_ELEMENTS(displays), PREVIEW_FFT);
    gwy_param_def_add_gwyflags(paramdef, PARAM_OUTPUT, "output", _("Output"),
                               outputs, G_N_ELEMENTS(outputs), 1 << OUTPUT_IMAGE);
    gwy_param_def_add_boolean(paramdef, PARAM_SNAP, "snap", _("_Snap to origin"), TRUE);
    return paramdef;
}

static void
module_main(GwyContainer *data, GwyRunType run)
{
    g_return_if_fail(run & RUN_MODES);

    ModuleArgs args;
    gwy_clear1(args);

    gint id;
    gwy_app_data_browser_get_current(GWY_APP_DATA_FIELD, &args.field,
                                     GWY_APP_DATA_FIELD_ID, &id,
                                     0);

    g_return_if_fail(GWY_IS_DATA_FIELD(args.field));
    args.result = gwy_data_field_new_alike(args.field, FALSE);
    args.diff = gwy_data_field_new_alike(args.field, FALSE);
    args.filtered_modulus = gwy_data_field_new_alike(args.field, FALSE);
    args.mask = gwy_data_field_new_alike(args.field, TRUE);
    gwy_si_unit_set_from_string(gwy_data_field_get_si_unit_z(args.mask), NULL);
    args.params = gwy_params_new_from_settings(define_module_params());

    GwyDialogOutcome outcome = run_gui(&args, data, id);
    gwy_params_save_to_settings(args.params);
    if (outcome == GWY_DIALOG_CANCEL)
        goto end;
    if (outcome != GWY_DIALOG_HAVE_RESULT)
        execute(&args);

    OutputType outputs = gwy_params_get_flags(args.params, PARAM_OUTPUT);
    if (outputs & OUTPUT_IMAGE) {
        gint newid = add_output_field(data, args.result, id, _("Filtered Data"));
        gwy_app_sync_data_items(data, data, id, newid, FALSE,
                                GWY_DATA_ITEM_RANGE_TYPE,
                                GWY_DATA_ITEM_RANGE,
                                0);
    }
    if (outputs & OUTPUT_DIFF) {
        gint newid = add_output_field(data, args.diff, id, _("Difference"));
        gwy_app_sync_data_items(data, data, id, newid, FALSE,
                                GWY_DATA_ITEM_RANGE_TYPE,
                                0);
    }
    /* FIXME: Show the mask? */
    if (outputs & OUTPUT_FFT) {
        add_output_field(data, args.filtered_modulus, id, _("Filtered Modulus"));
    }

end:
    g_object_unref(args.params);
    g_clear_object(&args.result);
    g_clear_object(&args.diff);
    g_clear_object(&args.mask);
    g_clear_object(&args.filtered_modulus);
}

static GwyDialogOutcome
run_gui(ModuleArgs *args, GwyContainer *data, gint id)
{
    static const GwyEnum mode_icons[] = {
        { GWY_ICON_MASK_CIRCLE_INCLUSIVE, FFT_ELLIPSE_ADD,   },
        { GWY_ICON_MASK_RECT_INCLUSIVE,   FFT_RECTANGLE_ADD, },
        { GWY_ICON_MASK_CIRCLE_EXCLUSIVE, FFT_ELLIPSE_SUB,   },
        { GWY_ICON_MASK_RECT_EXCLUSIVE,   FFT_RECTANGLE_SUB, },
    };
    static const GwyRGBA mask_color = { 0.56, 0.39, 0.07, 0.5 };

    ModuleGUI gui;
    gwy_clear1(gui);
    gui.args = args;
    gui.update = UPDATE_FILTERED;

    gui.modulus = gwy_data_field_new_alike(args->field, FALSE);
    calculate_modulus(args->field, gui.modulus);
    gui.data = gwy_container_new();
    gwy_container_set_object(gui.data, gwy_app_get_data_key_for_id(PREVIEW_IMAGE), args->field);
    gwy_container_set_object(gui.data, gwy_app_get_data_key_for_id(PREVIEW_FILTERED), args->result);
    gwy_container_set_object(gui.data, gwy_app_get_data_key_for_id(PREVIEW_DIFF), args->diff);
    calculate_zoomed_fields(&gui);

    gwy_container_set_const_string(gui.data, gwy_app_get_data_palette_key_for_id(PREVIEW_FFT), "DFit");
    gwy_container_set_enum(gui.data, gwy_app_get_data_range_type_key_for_id(PREVIEW_FFT), GWY_LAYER_BASIC_RANGE_AUTO);
    gwy_app_sync_data_items(data, gui.data, id, PREVIEW_FFT, FALSE,
                            GWY_DATA_ITEM_REAL_SQUARE,
                            0);
    gwy_rgba_store_to_container(&mask_color, gui.data, "/0/mask");
    for (gint i = PREVIEW_IMAGE; i <= PREVIEW_DIFF; i++) {
        gwy_app_sync_data_items(data, gui.data, id, i, FALSE,
                                GWY_DATA_ITEM_GRADIENT,
                                GWY_DATA_ITEM_RANGE_TYPE,
                                GWY_DATA_ITEM_REAL_SQUARE,
                                0);
    }
    gwy_app_sync_data_items(data, gui.data, id, PREVIEW_IMAGE, FALSE, GWY_DATA_ITEM_RANGE, 0);

    gui.dialog = gwy_dialog_new(_("2D FFT Filtering"));
    GwyDialog *dialog = GWY_DIALOG(gui.dialog);
    gwy_dialog_add_buttons(dialog, GWY_RESPONSE_RESET, GTK_RESPONSE_CANCEL, GTK_RESPONSE_OK, 0);

    gui.view = gwy_create_preview(gui.data, 0, PREVIEW_SIZE, TRUE);
    GtkWidget *hbox = gwy_create_dialog_preview_hbox(GTK_DIALOG(dialog), GWY_DATA_VIEW(gui.view), FALSE);

    GwyParamTable *table = gui.table = gwy_param_table_new(args->params);
    gwy_param_table_append_header(table, -1, _("Edit Filter"));
    gwy_param_table_append_radio_buttons(table, PARAM_EDIT_MODE, mode_icons);
    gwy_param_table_append_checkbox(table, PARAM_SNAP);
    gwy_param_table_append_button(table, BUTTON_UNDO, -1, RESPONSE_UNDO, _("_Undo"));
    gwy_param_table_button_set_icon(table, BUTTON_UNDO, GWY_ICON_GTK_UNDO);
    gwy_param_table_append_button(table, BUTTON_CLEAR, BUTTON_UNDO, GWY_RESPONSE_CLEAR, _("_Clear"));
    gwy_param_table_button_set_icon(table, BUTTON_CLEAR, GWY_ICON_GTK_CLEAR);
    gwy_param_table_append_button(table, BUTTON_FILL, BUTTON_CLEAR, RESPONSE_FILL, _("_Fill"));
    gwy_param_table_button_set_icon(table, BUTTON_FILL, GWY_ICON_MASK);
    gwy_param_table_append_header(table, -1, _("Preview"));
    gwy_param_table_append_radio(table, PARAM_DISPLAY);
    gwy_param_table_append_radio_row(table, PARAM_ZOOM);
    gwy_param_table_append_header(table, -1, _("Output"));
    gwy_param_table_append_checkboxes(table, PARAM_OUTPUT);
    /* FIXME: There used to be a bunch of tooltips. We need param table support.
    gtk_widget_set_tooltip_text(button, _("Undo the last change to the filter mask"));
    gtk_widget_set_tooltip_text(button, _("Clear the entire filter mask"));
    gtk_widget_set_tooltip_text(button, _("Fill the entire filter mask"));
    gtk_widget_set_tooltip_text(button, _("Force shapes to center around the origin"));
    */

    gtk_box_pack_start(GTK_BOX(hbox), gwy_param_table_widget(table), FALSE, FALSE, 0);
    gwy_dialog_add_param_table(dialog, table);
    gwy_dialog_set_preview_func(dialog, GWY_PREVIEW_IMMEDIATE, preview, &gui, NULL);
    g_signal_connect_swapped(table, "param-changed", G_CALLBACK(param_changed), &gui);
    g_signal_connect_swapped(dialog, "response", G_CALLBACK(dialog_response), &gui);

    GwyDialogOutcome outcome = gwy_dialog_run(dialog);

    g_object_unref(gui.data);
    g_object_unref(gui.modulus);

    return outcome;
}

static void
param_changed(ModuleGUI *gui, gint id)
{
    g_printerr("param_changed(%d)\n", id);
    if (id < 0 || id == PARAM_DISPLAY)
        update_sensitivity(gui);
    if (id < 0 || id == PARAM_DISPLAY || id == PARAM_EDIT_MODE)
        switch_layer(gui);
    if (id < 0 || id == PARAM_ZOOM)
        gui->update |= UPDATE_ZOOMED;
    if (id < 0 || id == PARAM_SNAP)
        update_snap(gui);

    if (id < 0 || id == PARAM_DISPLAY || id == PARAM_ZOOM)
        gwy_dialog_invalidate(GWY_DIALOG(gui->dialog));

    /* TODO: Disable OK button when there is no output. */
}

static void
preview(gpointer user_data)
{
    ModuleGUI *gui = (ModuleGUI*)user_data;
    ModuleArgs *args = gui->args;
    PreviewType display = gwy_params_get_enum(args->params, PARAM_DISPLAY);

    if (display == PREVIEW_FFT && (gui->update & UPDATE_ZOOMED)) {
        calculate_zoomed_fields(gui);
        gwy_set_data_preview_size(GWY_DATA_VIEW(gui->view), PREVIEW_SIZE);
        gui->update &= ~UPDATE_ZOOMED;
    }

    if ((display == PREVIEW_FILTERED || display == PREVIEW_DIFF) && (gui->update & UPDATE_FILTERED)) {
        execute(args);
        gwy_data_field_data_changed(args->result);
        gwy_data_field_data_changed(args->diff);
        gui->update &= ~UPDATE_FILTERED;
    }
}

static void
update_sensitivity(ModuleGUI *gui)
{
    PreviewType display = gwy_params_get_enum(gui->args->params, PARAM_DISPLAY);
    GwyParamTable *table = gui->table;
    gboolean is_editor = (display == PREVIEW_FFT);
    gboolean has_undo = gwy_undo_container_has_undo(gui->data);

    gwy_param_table_set_sensitive(table, PARAM_ZOOM, is_editor);
    gwy_param_table_set_sensitive(table, PARAM_EDIT_MODE, is_editor);
    gwy_param_table_set_sensitive(table, PARAM_SNAP, is_editor);
    gwy_param_table_set_sensitive(table, BUTTON_UNDO, is_editor && has_undo);
    gwy_param_table_set_sensitive(table, BUTTON_CLEAR, is_editor);
    gwy_param_table_set_sensitive(table, BUTTON_FILL, is_editor);
}

static void
switch_layer(ModuleGUI *gui)
{
    GwyParams *params = gui->args->params;
    PreviewType display = gwy_params_get_enum(params, PARAM_DISPLAY);
    EditType mode = gwy_params_get_enum(params, PARAM_EDIT_MODE);
    GwyDataView *dataview = GWY_DATA_VIEW(gui->view);
    GwyPixmapLayer *player = GWY_PIXMAP_LAYER(gwy_data_view_get_base_layer(dataview));
    GwyLayerBasic *blayer = GWY_LAYER_BASIC(player);

    g_printerr("switch layer to %u\n", display);
    gwy_pixmap_layer_set_data_key(player, g_quark_to_string(gwy_app_get_data_key_for_id(display)));
    gwy_layer_basic_set_gradient_key(blayer, g_quark_to_string(gwy_app_get_data_palette_key_for_id(display)));
    gwy_layer_basic_set_min_max_key(blayer, g_quark_to_string(gwy_app_get_data_base_key_for_id(display)));
    gwy_layer_basic_set_range_type_key(blayer, g_quark_to_string(gwy_app_get_data_range_type_key_for_id(display)));
    gwy_set_data_preview_size(dataview, PREVIEW_SIZE);

    if (display != PREVIEW_FFT) {
        gwy_data_view_set_top_layer(dataview, NULL);
        gwy_data_view_set_alpha_layer(dataview, NULL);
        return;
    }

    GwyDataViewLayer *mask_layer = gwy_data_view_get_alpha_layer(dataview);
    if (!mask_layer) {
        player = gwy_layer_mask_new();
        gwy_pixmap_layer_set_data_key(player, g_quark_to_string(gwy_app_get_mask_key_for_id(PREVIEW_FFT)));
        gwy_layer_mask_set_color_key(GWY_LAYER_MASK(player),
                                     g_quark_to_string(gwy_app_get_mask_key_for_id(PREVIEW_FFT)));
        gwy_data_view_set_alpha_layer(dataview, GWY_DATA_VIEW_LAYER(player));
    }

    GwyDataViewLayer *vec_layer = gwy_data_view_get_top_layer(dataview);
    GType layer_type = 0;
    const gchar *selname = "";
    if (mode == FFT_RECTANGLE_ADD || mode == FFT_RECTANGLE_SUB) {
        layer_type = GWY_TYPE_LAYER_RECTANGLE;
        selname = "/0/select/rectangle";
    }
    else if (mode == FFT_ELLIPSE_ADD || mode == FFT_ELLIPSE_SUB) {
        layer_type = GWY_TYPE_LAYER_ELLIPSE;
        selname = "/0/select/ellipse";
    }
    else {
        g_return_if_reached();
    }

    if (!vec_layer || !g_type_is_a(G_OBJECT_TYPE(vec_layer), layer_type)) {
        vec_layer = g_object_new(layer_type, NULL);
        gwy_data_view_set_top_layer(dataview, vec_layer);
        gwy_vector_layer_set_selection_key(GWY_VECTOR_LAYER(vec_layer), selname);
        update_snap(gui);

        GwySelection *selection = gwy_vector_layer_ensure_selection(GWY_VECTOR_LAYER(vec_layer));
        g_signal_connect(selection, "finished", G_CALLBACK(selection_finished), gui);
    }
}

static void
update_snap(ModuleGUI *gui)
{
    GwyDataViewLayer *vec_layer = gwy_data_view_get_top_layer(GWY_DATA_VIEW(gui->view));
    if (vec_layer) {
        gboolean snap = gwy_params_get_boolean(gui->args->params, PARAM_SNAP);
        g_object_set(vec_layer,
                     "snap-to-center", snap,
                     "draw-reflection", !snap,
                     NULL);
    }
}

static void
selection_finished(GwySelection *selection, ModuleGUI *gui)
{
    gdouble sel[4];
    if (!gwy_selection_get_object(selection, 0, sel))
        return;

    ModuleArgs *args = gui->args;
    GwyDataField *modulus = gui->modulus, *mask = args->mask;
    GwyDataField *zoomed = gwy_container_get_object(gui->data, gwy_app_get_data_key_for_id(PREVIEW_FFT));
    EditType mode = gwy_params_get_enum(args->params, PARAM_EDIT_MODE);
    gboolean snap = gwy_params_get_boolean(args->params, PARAM_SNAP);

    gint width = gwy_data_field_get_xres(modulus);
    gint height = gwy_data_field_get_yres(modulus);
    gint zwidth = gwy_data_field_get_xres(zoomed);
    gint zheight = gwy_data_field_get_yres(zoomed);

    gint isel[4] = {
        gwy_data_field_rtoj(zoomed, sel[0]) + (width - zwidth)/2,
        gwy_data_field_rtoi(zoomed, sel[1]) + (height - zheight)/2,
        gwy_data_field_rtoj(zoomed, sel[2]) + (width - zwidth)/2,
        gwy_data_field_rtoi(zoomed, sel[3]) + (height - zheight)/2,
    };
    if (!snap) {
        isel[2]++;
        isel[3]++;
    }

    /* Because of the offset (see below), must make sure selection is not along left or top edge. */
    isel[0] = MAX(isel[0], 1);
    isel[1] = MAX(isel[1], 1);

    /* For the mirrored selection to look "correct" as far as the FFT goes, it must be shifted one pixel to the right,
     * and one pixel down for even-sized images.
     * XXX: It looks a bit weird when the selection and mask are displayed together. */
    gint mirror[4];
    mirror[0] = (-isel[2] + width) + (1 - width % 2);
    mirror[1] = (-isel[3] + height) + (1 - height % 2);
    mirror[2] = (-isel[0] + width) + (1 - width % 2);
    mirror[3] = (-isel[1] + height) + (1 - height % 2);

    g_printerr("area [%d..%d] × [%d..%d]\n", isel[0], isel[2], isel[1], isel[3]);
    g_printerr("mirror [%d..%d] × [%d..%d]\n", mirror[0], mirror[2], mirror[1], mirror[3]);

    /* Change coordinates to sizes. */
    isel[2] -= isel[0];
    isel[3] -= isel[1];
    mirror[2] -= mirror[0];
    mirror[3] -= mirror[1];

    /* Choose the right operation and shape. */
    FieldFillFunc fill_func = gwy_data_field_area_fill;
    if (mode == FFT_ELLIPSE_ADD || mode == FFT_ELLIPSE_SUB)
        fill_func = (FieldFillFunc)gwy_data_field_elliptic_area_fill;

    gdouble value = 1.0;
    if (mode == FFT_RECTANGLE_SUB || mode == FFT_ELLIPSE_SUB)
        value = 0.0;

    /* Apply change to mask and queue recomputation. */
    gwy_app_undo_qcheckpoint(gui->data, gwy_app_get_mask_key_for_id(0), NULL);
    fill_func(mask, isel[0], isel[1], isel[2], isel[3], value);
    fill_func(mask, mirror[0], mirror[1], mirror[2], mirror[3], value);
    gwy_selection_clear(selection);

    gui->update |= UPDATE_ZOOMED | UPDATE_FILTERED;
    update_sensitivity(gui);
    gwy_dialog_invalidate(GWY_DIALOG(gui->dialog));
}

static void
dialog_response(ModuleGUI *gui, gint response)
{
    ModuleArgs *args = gui->args;

    if (response == RESPONSE_UNDO) {
        if (gwy_undo_container_has_undo(gui->data)) {
            gwy_undo_undo_container(gui->data);
            gui->update |= UPDATE_ZOOMED | UPDATE_FILTERED;
            update_sensitivity(gui);
            gwy_dialog_invalidate(GWY_DIALOG(gui->dialog));
        }
    }
    else if (response == GWY_RESPONSE_CLEAR || response == RESPONSE_FILL) {
        gwy_app_undo_qcheckpoint(gui->data, gwy_app_get_mask_key_for_id(0), NULL);
        /* Compute the results trivially here. No need to queue FFT. */
        if (response == GWY_RESPONSE_CLEAR) {
            gwy_data_field_clear(args->mask);
            gwy_data_field_assign(args->result, args->field);
            gwy_data_field_clear(args->diff);
        }
        else {
            gwy_data_field_fill(args->mask, 1.0);
            gwy_data_field_clear(args->result);
            gwy_data_field_assign(args->diff, args->field);
        }
        gui->update |= UPDATE_ZOOMED;
        update_sensitivity(gui);
        gwy_dialog_invalidate(GWY_DIALOG(gui->dialog));
    }
}

static void
execute(ModuleArgs *args)
{
    g_printerr("calculating RESULT fields\n");
    /* Run the forward FFT. */
    GwyDataField *r_in = args->result;
    GwyDataField *r_out = gwy_data_field_new_alike(r_in, FALSE);
    GwyDataField *i_out = gwy_data_field_new_alike(r_in, FALSE);
    gwy_data_field_assign(r_in, args->field);

    gwy_data_field_fft_2d_raw(r_in, NULL, r_out, i_out, GWY_TRANSFORM_DIRECTION_FORWARD);

    /* Remember the filtered modulus in case someone wants it as an output. */
    gwy_data_field_hypot_of_fields(args->filtered_modulus, r_out, i_out);
    gwy_data_field_fft_postprocess(args->filtered_modulus, TRUE);
    gwy_data_field_multiply_fields(args->filtered_modulus, args->filtered_modulus, args->mask);

    /* Apply mask to the fft */
    gwy_data_field_fft_2d_center(r_out);
    gwy_data_field_fft_2d_center(i_out);

    gwy_data_field_multiply_fields(r_out, r_out, args->mask);
    gwy_data_field_multiply_fields(i_out, i_out, args->mask);

    /* Run the inverse FFT */
    GwyDataField *i_in = gwy_data_field_new_alike(r_in, TRUE);
    gwy_data_field_fft_2d_decenter(r_out);
    gwy_data_field_fft_2d_decenter(i_out);
    gwy_data_field_fft_2d_raw(r_out, i_out, r_in, i_in, GWY_TRANSFORM_DIRECTION_BACKWARD);

    gwy_data_field_subtract_fields(args->diff, args->field, args->result);

    /* Finalize */
    g_object_unref(i_out);
    g_object_unref(r_out);
    g_object_unref(i_in);
}

static void
calculate_zoomed_fields(ModuleGUI *gui)
{
    g_printerr("calculating ZOOMED fields\n");
    ModuleArgs *args = gui->args;
    guint zoom = gwy_params_get_enum(args->params, PARAM_ZOOM);
    GwyDataField *zoomed;

    zoomed = cut_field_to_zoom(gui->modulus, zoom);
    gwy_data_field_data_changed(zoomed);
    gwy_container_pass_object(gui->data, gwy_app_get_data_key_for_id(PREVIEW_FFT), zoomed);

    zoomed = cut_field_to_zoom(args->mask, zoom);
    gwy_data_field_data_changed(zoomed);
    gwy_container_pass_object(gui->data, gwy_app_get_mask_key_for_id(PREVIEW_FFT), zoomed);
}

static GwyDataField*
cut_field_to_zoom(GwyDataField *field, gint zoom)
{
    GwyDataField *zoomed;
    guint xres, yres, width, height;

    xres = gwy_data_field_get_xres(field);
    yres = gwy_data_field_get_yres(field);
    width = (xres/zoom) | 1;
    height = (yres/zoom) | 1;
    if (width < 17)
        width = MAX(width, MIN(17, xres));
    if (height < 17)
        height = MAX(height, MIN(17, yres));

    if (width >= xres && height >= yres)
        return g_object_ref(field);

    zoomed = gwy_data_field_area_extract(field, (xres - width + 1)/2, (yres - height + 1)/2, width, height);
    gwy_data_field_set_xoffset(zoomed, -0.5*gwy_data_field_get_xreal(zoomed));
    gwy_data_field_set_yoffset(zoomed, -0.5*gwy_data_field_get_yreal(zoomed));

    return zoomed;
}

/* This is the modulus we show in GUI (with windowing, etc.) */
static void
calculate_modulus(GwyDataField *field, GwyDataField *fftmod)
{
    gwy_data_field_psdf_2d(field, fftmod, GWY_WINDOWING_LANCZOS, 1);
    gint n = gwy_data_field_get_xres(fftmod)*gwy_data_field_get_yres(fftmod);
    gdouble *d = gwy_data_field_get_data(fftmod);
    for (gint i = 0; i < n; i++)
        d[i] = sqrt(fmax(d[i], 0.0));
}

static gint
add_output_field(GwyContainer *data, GwyDataField *field,
                 gint oldid, const gchar *title)
{
    gint newid = gwy_app_data_browser_add_data_field(field, data, TRUE);
    gwy_app_sync_data_items(data, data, oldid, newid, FALSE,
                            GWY_DATA_ITEM_GRADIENT,
                            GWY_DATA_ITEM_MASK_COLOR,
                            GWY_DATA_ITEM_RANGE,
                            GWY_DATA_ITEM_RANGE_TYPE,
                            GWY_DATA_ITEM_REAL_SQUARE,
                            0);
    gwy_app_set_data_field_title(data, newid, title);
    gwy_app_channel_log_add_proc(data, oldid, newid);

    return newid;
}

/* vim: set cin columns=120 tw=118 et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
